//==================================================================================================
// NUON Boot Loader
// by DragonShadow Industries
//
// Allows burning of multiple NUON executable files to one disc.
//
// See the accompanying README.TXT for full program description and usage.
//
//   Copyright (C) 2004 Scott Cartier
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//==================================================================================================

//==================================================================================================
// Please widen your window to allow 100 characters per line
//==================================================================================================
// 4567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
//==================================================================================================

//==================================================================================================
// Includes
//==================================================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <nuon/mml2d.h>
#include <nuon/mutil.h>
#include <nuon/msprintf.h>
#include <nuon/dma.h>
#include <nuon/termemu.h>
#include <nuon/cdn50x.h>
#include <nuon/nise.h>
#include <jpeg/jpeglib.h>
#include <nuon/intr.h>
#include "bootload.h"

//==================================================================================================
// Defines
//==================================================================================================

// comment these out for release
//#define USE_LOCAL_HOST          // Un-comment for use on dev systems
//#define DEBUG                   // Debug console messages
//#define TRACE                   // Print function enter/exit
//#define CONSOLE_PRINT           // Print console messages to stdout
//#define APPLIST_PRINT           // Print the applist parsing to stdout

#ifdef OLD_TGA
#include "tga/tga.h"
#endif

// options
#define NO_INSTR2               // disables holding of B button to get 2nd page of instructions
#define NO_SCROLL_DESC          // disables the c-up and c-down buttons for scrolling descriptions

// Init background selection (select one)
#define drawInit drawInitNormal
//#define drawInit drawInitGamesAndDemos
//#define drawInit drawInitYarozeClassics

//#define USEINITBG               // set when using an image during loading

#ifdef USE_LOCAL_HOST
#define DSI_DEBUG               // enables the screenshot routine when on a dev system
#endif

#define UNUSED              0   // for unused function parameters
#define DC                  0   // for don't care function parameters

#define FALSE               0   // duh
#define TRUE                1

#define MPE0                0
#define MPE1                1
#define MPE2                2

#define RET_NORMAL          0   // function return codes
#define RET_ERROR           1
#define RET_DONE           -1

#define SCREEN_W           720  // display dimensions
#define SCREEN_H           480

#define NUM_BUFFERS         2   // double buffer the display

#define DMAFLAGS_RD        (DMA_PIXEL_READ  | ((SCREEN_W/8)<<16) | (4<<4) | (1<<11))

#define ELEMENT_SPACE      10   // default element spacing

#define MAX_USER_FONTS      4   // maximum number of user-defined fonts
#define MAX_EMBEDDED_FONTS  1

#define CONSOLE_LINES_MAX  30   // max number of console lines
#define CONSOLE_LINE_SIZE  50   // max number of console chars per line

#define MIN_FONT_SIZE      15   // font size limits
#define MAX_FONT_SIZE      70

#define MIN_FONT_INDEX  -MAX_EMBEDDED_FONTS   // font index limits
#define MAX_FONT_INDEX   MAX_USER_FONTS

#define MIN_SPACING       -15   // element spacing limits
#define MAX_SPACING        50

#define MIN_BORDER          0   // screen border limits
#define MAX_BORDER        100

#define IMG_INTRO_W       720   // size of intro image
#define IMG_INTRO_H       480

#define BUTTON_DELAY       31   // the number of frames to wait before repeating a held down button

#define REOPEN_ATTEMPTS    10   // number of times to try to read the data file before giving up

// "BOOTLOAD"(8) + file size(4) + header size(4) + checksum(4) + num files(4)
#define DATA_FILE_HEADER_SIZE  24

#define MIN_ROTATE_IMG_DELAY 1
#define MAX_ROTATE_IMG_DELAY 60

#define MENU_MAX_ENTRIES_MIN 0
#define MENU_MAX_ENTRIES_MAX 100

// Defines to abstract away the file routines since they differ depending on whether we're using
// the file server with the dev system or the CD library with a real player.
#ifdef USE_LOCAL_HOST
#define myFileType FILE
#define myfclose(fp) fclose(fp)
#define myfseek(fp,off,whence) fseek(fp, off, whence)
#define myftell(fp) ftell(fp)
#define myfread(buf,numbytes,fp) fread(buf, 1, numbytes, fp)
#define HOST_PATH (bufDataFile ? "\0" : "/host/")
#else
#define myFileType void *
#define myfclose(fp) iso_close(cdContextPtr, fp, &cdError);
#define myfseek(fp,off,whence) iso_lseek(cdContextPtr, fp, off, whence, &cdError)
#define myftell(fp) iso_lseek(cdContextPtr, fp, 0, SEEK_CUR, &cdError)
#define myfread(buf,numbytes,fp) iso_read(cdContextPtr, fp, buf, numbytes, &cdError);
#define HOST_PATH "\0"
#endif

// Combine the IR remote, and controllers 1 & 2
#define JOY_BUTTONS  (Buttons(_Controller[0]) | Buttons(_Controller[1]) | Buttons(_Controller[2]))

//==================================================================================================
// Binaries (see data.s)
//==================================================================================================
extern char Wait[];            // assembly to clear the caches and wait for a comm bus packet
extern char WaitEnd[];
extern char coff_c_start[];    // assembly to load the coff file from sdram to system ram
extern char coff_c_size[];
#ifdef USEINITBG
extern char LoadingImg[];      // image to display while initializing
#endif

// Embedded fonts
extern char FontEpilog[];

// Embedded images
//extern char ImgWait[];
//extern char ImgButtons[];

// Assembly TGA decoder
#ifndef OLD_TGA
extern char tgac_start[], tgac_size[];
extern char tgad_start[], tgad_size[];
#endif

// Embedded data file so I can use the debugger
//extern char DataFile[], DataFileEnd[];
#define DataFile NULL

//==================================================================================================
// Structures
//==================================================================================================

// App flags
enum {
  kFlagAppValid       = (1<<0),   // the app is valid (has an executable)
  kFlagSectionHeader  = (1<<4),   // set if this entry specifies the start of a section only
};

enum {
  kImgStatusInit,
  kImgStatusSubmitted,
  kImgStatusDrawn,
  kImgStatusError,
};

enum {
  kTextFlagNoDraw = (1<<0),   // do not actually draw anything (just return the bounding height)
  kTextFlagCenter = (1<<1),   // center the text in the provided area
  kTextFlagRight  = (1<<2),   // right justify the text in the provided area
  kTextFlagLeft   = (1<<3),   // left justify the text in the provided area
};

typedef struct {
  unsigned char *filename;    // text file name of this image (pointer into file buffer)
  mmlDisplayPixmap *pixmap;   // MML pixmap used to draw this image
  int w, h;                   // actual dimensions of this image (pixmap rounds up)
  volatile int status;        // 0 = not drawn, 1 = submitted, 2 = drawn, 3+ = error of some sort
  struct sAppImgGroupTag *parent;  // pointer back to group structure
} sAppImg;

struct sAppImgGroupTag {
  sAppImg *aAppImg;           // image array
  int numImg;                 // number of images in this group
  int curImg;                 // index of image being displayed (or index into 'order' array)
  int speed;                  // how often to switch to a new image
  int *order;                 // array of image indices indicating the image order
  int w, h;                   // screen dimensions and position...
  int x, y;                   // ...only used with absolute positioning
  int draw;                   // flag indicating if this image needs drawing
  int time;                   // used for image rotation
};

typedef struct sAppImgGroupTag sAppImgGroup;

typedef struct {
  char *dirName;              // Directory name
  long flags;                 // valid and other bits (see above)
  char *txtTitle;             // Text to use in menu
  char *txtDesc;              // Descriptive string to appear below images
  sAppImgGroup *aImgGroups;   // App images
  int numImgGroups;           // Size of the image group array
  sAppImg imgIntro; // Image shown when app is selected
} sApp;

// Structures to make screen layout a bit easier to track
typedef struct {
  long x;
  long y;
} sCoord;

typedef struct {
  long x, y;
  long x2, y2;
  long w, h;
} sRect;

//==================================================================================================
// Globals
//==================================================================================================

// Display stuff
mmlSysResources sysRes;   // system resources
mmlGC gc;                 // graphics context
VidDisplay vd;            // video display
VidChannel vcMain;        // main channel
mmlDisplayPixmap frameBuffer[NUM_BUFFERS];   // frame buffers
int dispBuffer;
int drawBuffer;           // index into frame buffer array for the current draw buffer
int fps;                  // current frame rate

// Background
#ifdef OLD_TGA
mmlDisplayPixmap *loadingImg = NULL;  // pixmap to use when initializing boot loader
mmlDisplayPixmap *bgImg = NULL;       // pixmap to use as a background image
#define LOADING_IMG_PTR loadingImg
#define BG_IMG_PTR bgImg
#else
sAppImg loadingImg;       // pixmap to use when initializing boot loader
sAppImg bgImg;            // pixmap to use as a background image
#define LOADING_IMG_PTR loadingImg.pixmap
#define BG_IMG_PTR bgImg.pixmap
#endif

// Sounds
void *wavMove   = NULL;   // sound when user navigates the menu or description
void *wavSelect = NULL;   // sound when user selects a game, starts a game, or toggles fast mode

// Console output
char aConsole[CONSOLE_LINES_MAX][CONSOLE_LINE_SIZE+1];
int consoleStart = 0;
int forbidAppend = FALSE;

// Menu output
int curApp  = -1;   // The app that's the user has currently highlighted
int topApp  = 0;    // The app at the start of the menu
int botApp  = 0;    // The app at the end of the menu
int menuY   = 0x20; // Y component for selected menu item
int topDesc = 0;    // Top line of the description to start drawing

// A temporary string for general use
#define TMP_STR_SIZE 100
char tmpStr[TMP_STR_SIZE];

// File names
char *fileDat       = "bootload.dat";   // data file
char *fileAppList   = "applist.txt";    // app names & parameters
char *fileAppExe1   = "cd_app.cof";     // executable
char *fileAppExe2   = "NUON.CD";        // alternate executable
char *fileAppText   = "app.txt";        // app title & description
char *fileImgIntro  = "intro.tga";      // intro image when starting app
char *fileImgBG     = "bg.tga";         // background image
char *fileWavMove   = "move.wav";       // movement sound
char *fileWavSelect = "select.wav";     // select sound
char *fileAppImgA   = "img1.tga";       // app images (for compatability)
char *fileAppImgB   = "img1a.tga";      // app images (new expanded filenames)

typedef enum {
  kFileTypeExe,
  kFileTypeText,
  kFileTypeTGA,
  kFileTypeJPG,
  kFileTypeWAV,
  kFileTypeTTF,
} eFileType;

//
// Colors
//
enum {
  kColorBg,               // Background color
  kColorTitle,            // Title text color
  kColorInstr,            // Title instructions color
  kColorConsoleNormal,    // Color of console text for past messages
  kColorConsoleHighlight, // Color of console text for current message
  kColorConsoleError,     // Color of console text for errors
  kColorMenuNormal,       // Menu color for normal entries
  kColorMenuHighlight,    // Menu color for currently highlighted entry
  kColorMenuError,        // Menu color when an app has a problem
  kColorSection,          // Menu color for section headers
  kColorArrow1,           // Up/down arrows in menu indicating more apps in the list
  kColorArrow2,           // ...
  kColorDesc,             // Color for app description
  kColorLoading,          // Color for loading message
  kColorAries3,           // Color for Aries3 mode indicator
  kColorMax               // Used in array declarations
};

typedef struct {
  char r, g, b;
} sColor;

// Constant colors used for the inialization screen
sColor myBlue   = {0x20, 0x20, 0x30};
sColor myYellow = {0x80, 0x80, 0x20};
sColor myWhite  = {0x80, 0x80, 0x80};

// For special case loading screen
sColor colorInit1  = {0xFF, 0xFF, 0xFF};
sColor colorInit2  = {0xFF, 0xFF, 0xFF};
sColor colorInit3  = { 200,  200,  255};

sColor aColorDefaults[kColorMax] = {
  { 0x20, 0x20, 0x30 },  // Background (used when no image present)
  { 0x80, 0x80, 0x20 },  // Title
  { 0x80, 0x80, 0x80 },  // Instructions
  { 0x80, 0x80, 0x80 },  // Console message - normal
  { 0xC0, 0xC0, 0xC0 },  // Console message - current
  { 0x80, 0x40, 0x40 },  // Console message - error
  { 0x80, 0x80, 0x80 },  // Menu entry - normal
  { 0x80, 0xC0, 0x80 },  // Menu entry - highlighted
  { 0xC0, 0x80, 0x80 },  // Menu entry - error
  { 0xFF, 0xFF, 0xFF },  // Menu entry - section header
  { 0xFF, 0xFF, 0xFF },  // Menu arrow color
  { 0x00, 0x00, 0x00 },  // ...
  { 0xC0, 0xC0, 0x80 },  // Description text
  { 0x80, 0x80, 0x20 },  // Loading...
  { 0x80, 0x60, 0x60 },  // Aries3 mode indicator
};
mmlColor aColors[kColorMax];

char *aParamColor[] = {
  "ColorBg",
  "ColorTextTitle",
  "ColorTextInstr",
  "ColorTextConsoleNormal",
  "ColorTextConsoleHighlight",
  "ColorTextConsoleError",
  "ColorTextMenuNormal",
  "ColorTextMenuHighlight",
  "ColorTextMenuError",
  "ColorTextSection",
  "ColorTextArrow1",
  "ColorTextArrow2",
  "ColorDesc",
  "ColorTextLoading",
  "ColorTextAries3",
};

//
// Fonts
//
mmlFontContext fc;
mmlFont aFont[1+MAX_USER_FONTS+MAX_EMBEDDED_FONTS];  // array of fonts
char* aUserFontNames[MAX_USER_FONTS];

typedef struct {
  void *ptr;
  int size;
  char *name;
} sEmbeddedFont;

sEmbeddedFont aEmbeddedFonts[MAX_EMBEDDED_FONTS] = {
  { FontEpilog, 30136, "Epilog" },
};

char *strParamUserFont = "UserFont";

// Symbols to point into the font array
enum {
  kStrFontIndexTitle,
  kStrFontIndexInstr,
  kStrFontIndexMenu,
  kStrFontIndexConsole,
  kStrFontIndexDesc,
  kStrFontIndexLoading,
  kStrFontIndexAries3,
  kStrFontIndexSection,
  kStrFontIndexArrow,
  kStrFontIndexMax,
};

// Font index array with defaults
int aFontIndices[kStrFontIndexMax] = {
  0,  // Title
  0,  // Instructions
  0,  // Menu
  0,  // Console
  0,  // Description text
  0,  // Loading...
  0,  // Aries3 mode indicator
  0,  // Menu section header
  0,  // Arrow
};

// Parameters that alter font sizes
char *aParamFontIndices[] = {
  "TextFontIndexTitle",
  "TextFontIndexInstr",
  "TextFontIndexMenu",
  "TextFontIndexConsole",
  "TextFontIndexDesc",
  "TextFontIndexLoading",
  "TextFontIndexAries3",
  "TextFontIndexSection",
  "TextFontIndexArrow",
};

// Font size array with defaults
int aFontSizes[kStrFontIndexMax] = {
  50,  // Title
  30,  // Instructions
  40,  // Menu
  20,  // Console
  20,  // Description text
  50,  // Loading...
  30,  // Aries3 mode indicator
  30,  // Menu section header
  20,  // Arrow
};

// Parameters that alter font sizes
char *aParamFontSize[] = {
  "TextFontSizeTitle",
  "TextFontSizeInstr",
  "TextFontSizeMenu",
  "TextFontSizeConsole",
  "TextFontSizeDesc",
  "TextFontSizeLoading",
  "TextFontSizeAries3",
  "TextFontSizeSection",
  "TextFontSizeArrow",
};

//
// Strings
//

// Constant strings
char *strInit1 = "NUON Boot Loader";
char *strInit2 = "Initializing...";
char  strInit3[50];
char *strAbout   = NULL;
char *strVersion = " v4.0 ";  // last changed 2005/11/19
char *strAuthor  = "by DragonShadow Industries";

int fontIndexInit1 = 5;  // Use first embedded font
int fontIndexInit2 = 0;  // 0=system font, 1-4=user fonts, 5+=embedded fonts
int fontIndexInit3 = 0;

// Symbols to point into the string array
enum {
  kStrTitle,         // title text
  kStrInstr1Normal,  // instructions below title (normal game pad)
  kStrInstr1Remote,  // instructions below title (IR remote)
  kStrInstr2Normal,  // instructions when B button held (Aries 2)
  kStrInstr2Aries3,  // instructions when B button held (Aries 3)
  kStrLoading,       // text to show when app is loading
  kStrAries3,        // tells when the app is in fast mode
  kStrArrowPrev,     // indicator that there are more apps after the current one
  kStrArrowNext,     // indicator that there are more apps before the current one
  kStrMax
};

// Default strings.  Will be copied into the strings array at init.
char *aStringDefaults[kStrMax] = {
  "NUON Boot Loader",
  "Up/Dn selects, Lt/Rt skips section, A starts, B for more help",
  "Use direction buttons to choose, ENTER to start",
  "C-up/C-dn: scroll description, L+R: program info",
  "C-up/C-dn: scroll description, C-rt: fast mode, L+R: program info",
  "Loading...",
  "Fast Mode",
  "^",
  "v",
};

// Strings array
char *aStrings[kStrMax] = {
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
};

// Parameters that alter strings
char *aParamString[kStrMax] = {
  "TextTitle",
  "TextInstr1",
  "TextInstr1Remote",
  "TextInstr2",
  "TextInstr2Aries3",
  "TextLoading",
  "TextAries3",
  "TextArrowPrev",
  "TextArrowNext",
};

// Text horizontal alignment
enum {
  kHAlignLeft,
  kHAlignCenter,
  kHAlignRight,
};

char *textAlignLeft   = "Left";
char *textAlignCenter = "Center";
char *textAlignRight  = "Right";

enum {
  kTextHAlignTitle,
  kTextHAlignInstr,
  kTextHAlignMenu,
  kTextHAlignConsole,
  kTextHAlignDesc,
  kTextHAlignLoading,
  kTextHAlignAries3,
  kTextHAlignSection,
  kTextHAlignArrow,
  kTextHAlignMax
};

int aTextHAlign[kTextHAlignMax] = {
  kHAlignCenter,  // Title
  kHAlignCenter,  // Instructions
  kHAlignLeft,    // Menu
  kHAlignLeft,    // Console
  kHAlignLeft,    // Description
  kHAlignCenter,  // Loading
  kHAlignLeft,    // Aries3 indicator
  kHAlignLeft,    // Section
  kHAlignLeft,    // Arrows
};

char *aParamHAlign[kTextHAlignMax] = {
  "TextHAlignTitle",
  "TextHAlignInstr",
  "TextHAlignMenu",
  "TextHAlignConsole",
  "TextHAlignDesc",
  "TextHAlignLoading",
  "TextHAlignAries3",
  "TextHAlignSection",
  "TextHAlignArrow",
};

//
// Spacing
//

// Symbols to index into the spacing array
enum {
  kSpaceBelowTitle,
  kSpaceBelowInstr,
  kSpaceBelowImg1,
  kSpaceBelowImg2,
  kSpaceBelowMenuEntry,
  kSpaceMenuIndent,
  kSpaceMax,
};

// Spacing array
int aSpacing[kSpaceMax] = {
  -10, // Space between title and instructions
  0,   // Space between instructions and content
  5,   // Space between image 1 and image 2
  5,   // Space between image 1 and description
  0,   // Space between menu entries
  20,  // Space indenting for menu entries
};

// Parameters that alter the spacing of elements
char *aParamSpacing[] = {
  "SpaceBelowTitle",
  "SpaceBelowInstr",
  "SpaceBelowImg1",
  "SpaceBelowImg2",
  "SpaceBelowMenuEntry",
  "SpaceMenuIndent",
};

// Screen borders
enum {
  kBorderTop,
  kBorderBottom,
  kBorderRight,
  kBorderLeft,
  kBorderMax
};

int aBorders[kBorderMax] = {
  30,  // Top
  30,  // Bottom
  50,  // Right
  50,  // Left
};

char *aParamBorder[] = {
  "ScreenBorderTop",
  "ScreenBorderBottom",
  "ScreenBorderRight",
  "ScreenBorderLeft",
};

//
// Absolute positioning
//

char *strParamAbsolutePos = "AbsolutePositioning";
int absolutePositioning = FALSE;

enum {
  kPosTitle,
  kPosInstr,
  kPosDesc,
  kPosMenu,
  kPosAries3,
  kPosArrowPrev,
  kPosArrowNext,
  kPosConsole,
  kPosImg1,
  kPosImg2,
  kPosImg3,
  kPosImg4,
  kPosImg5,
  kPosImg6,
  kPosImg7,
  kPosImg8,
  kPosImg9,
  kPosMax
};

sRect aPos[kPosMax] = {
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
  { -1, -1, -1, -1, -1, -1 },  // arrow prev (position computed automatically unless specified)
  { -1, -1, -1, -1, -1, -1 },  // arrow next (position computed automatically unless specified)
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
  { 50, 30, SCREEN_W - 50, SCREEN_H - 30, SCREEN_W - 100, SCREEN_H - 60 },
};

char *aParamPos[] = {
  "PosTitle",
  "PosInstr",
  "PosDesc",
  "PosMenu",
  "PosAries3",
  "PosArrowPrev",
  "PosArrowNext",
  "PosConsole",
  "PosImg1",
  "PosImg2",
  "PosImg3",
  "PosImg4",
  "PosImg5",
  "PosImg6",
  "PosImg7",
  "PosImg8",
  "PosImg9",
};

//
// Menu parameters
//

// Menu orientation
char *strParamMenuHorizontal = "MenuHorizontal";

enum {
  kMenuVertical,
  kMenuHorizontal
};

int menuOrientation = kMenuVertical;

// Menu max entries to display
char *strParamMenuMaxEntries = "MenuMaxEntries";
int menuMaxEntries = -1;  // default = no maximum

// Menu highlight disable
char *strParamMenuNoHighlight = "MenuNoHighlight";
int menuNoHighlight = FALSE;

// Menu wrapping
char *strParamMenuWrap = "MenuWrap";
int menuWrap = FALSE;

//
// Image parameters
//

char *rotateImgNever       = "Never";
char *rotateImgOnHighlight = "OnHighlight";

#define kRotateImgNever       0
#define kRotateImgOnHighlight -1

enum {
  kRotateImg1,
  kRotateImg2,
  kRotateImg3,
  kRotateImg4,
  kRotateImg5,
  kRotateImg6,
  kRotateImg7,
  kRotateImg8,
  kRotateImg9,
  kRotateImgMax
};

int aRotateImg[kRotateImgMax] = {
  kRotateImgNever,  // default to static images
  kRotateImgNever,
  kRotateImgNever,
  kRotateImgNever,
  kRotateImgNever,
  kRotateImgNever,
  kRotateImgNever,
  kRotateImgNever,
  kRotateImgNever,
};

char *aParamRotateImg[] = {
  "RotateImg1",
  "RotateImg2",
  "RotateImg3",
  "RotateImg4",
  "RotateImg5",
  "RotateImg6",
  "RotateImg7",
  "RotateImg8",
  "RotateImg9",
};

//
// Misc parameters
//

char *strParamShowConsole = "ShowConsole";

char *strSection = "+Section";

char *strNoTitle = "???";

// -------------------------------------------------------------------------------------------------

// Embedded images
typedef struct {
  void *ptr;              // pointer to raw image data
  mmlDisplayPixmap *img;  // image data
  int w, h;               // image dimensions
  int r, c;               // number of rows and columns (for images that contain multiple frames)
} sEmbeddedImg;

enum {
  kEmbeddedImgWait,
  kEmbeddedImgMax,
};

sEmbeddedImg aEmbeddedImg[kEmbeddedImgMax] = {
  //  { ImgWait, },
};

// -------------------------------------------------------------------------------------------------

// App info
int  numApps;
sApp *aAppInfo;  // Array of app info

// Screenshot
void *screenshotBuffer = NULL;   // pointer to raster order screen shot buffer
char *localScratch;              // pointer to local RAM usable by the application
int localSize;                   // size of local RAM

// For CD library
void *cdContextPtr;              // pointer passed to all CD lib calls
int  cdError = 0;                // error flag passed to, and modified by, all CD lib calls
struct stat fileStat;            // file stat structure (not used)

//
// Global state
//

sCoord coordTitle;  // element locations
sCoord coordInstr;
sCoord coordImg1;
sCoord coordImg2;
sCoord coordDesc;

sRect  rectLeft  = {-1, -1, -1, -1};   // location and dimensions of the left side of the display
sRect  rectRight = {-1, -1, -1, -1};   // same for right (must be symmetric)

// Controllers
enum {
  kControlRemote,
  kControlJoy1,
  kControlJoy2,
};

// Aries 3 fast mode
int ariesVer;
int fastMode = -1;
int controlType = kControlJoy1;

// The current program state
enum {
  kStateInit,
  kStateConsoleAndMenu,
  kStateConsoleAndStuff,
  kStateMenuAndStuff,
  kStateIntro,
  kStateLoading,
};

int flags = 0;
int state = kStateInit;
int nextState = kStateInit;
//int state = kStateConsoleAndMenu;
//int state = kStateConsoleAndStuff;
int error = FALSE;

// Bit fields to indicate what part of the display needs redrawing.
// Each flag uses 4 bits so space them out accordingly.
enum {
  kRedrawMenu      = (1<<0),  // redraws the full menu
  kRedrawHighlight = (1<<4),  // subset of kRedrawMenu - only redraws the highlighted entry
  kRedrawAppStuff  = (1<<8),  // redraws the app images and description
  kRedrawTitle     = (1<<12), // redraws the title and instructions
  kRedrawConsole   = (1<<16), // redraws the console
  kRedrawInit3     = (1<<20), // redraws the third initialization line
  kRedrawImages    = (1<<24), // redraws the app images
  kRedrawAll       = kRedrawMenu | kRedrawHighlight | kRedrawAppStuff | kRedrawTitle | kRedrawConsole | kRedrawInit3 | kRedrawImages,
};
long redraw = kRedrawAll;

// Handy macro to check if a field needs redrawing.  For every field, setting bit 0 means causes
// it to be redrawn the next two frames.  Bits 1 & 2 indicate this.
#define checkRedraw(x) ( (((redraw>>1) | (redraw>>2)) & (x)) == (x) )


//
// Misc state
//

// Comm Bus interrupt magic number
void *cbMagicNum = NULL;

// Image info
long g_imgW, g_imgH;     // filled in when an image is constructed (bad way to do it, but oh well)
sAppImg *g_img;          // pointer to the image structure that's being loaded
//TGA *g_tga;
//JPG *g_jpg;

// General file info
int curFileSize;
int fieldsAccum = 0;       // cumulative number of fields for pad delay
                         // gets reset every BUTTON_DELAY frames
int cursorFields = 0;    // similar to above, but is used for color cycling the cursor
int blinkCursor = TRUE;  // can disable cursor color cycling to indicate non-responsiveness
char *instrStr = NULL;   // the string to print beneath the title

// DAT files
myFileType *fpDataFile = NULL;  // file pointer to data file
char *bufDataFile = NULL;       // pointer to data file buffer
int numDataFiles = 0;           // number of tiles contained in data file
int headerSize = 0;             // size of data file header (all but coffs)

// Sound stuff
PCMPOS Pan = {0, 0, 0, 0};

//==================================================================================================
// Local prototypes
//==================================================================================================

// Init
void    initDisplay();
void    initFonts();
void    initUserFonts();
void    initConsole();
void    initColors();
void    initStrings();
void    initCDLib();
void    initLoadingBG();
void    initSystem();

// Util
int     VideoSync(void);
#ifndef OLD_TGA
int     TGADoneCommHandler(int,int,int,int,int,int,int,int);
#endif

// Rendering
void    updateScreen();
void    drawInit();
void    drawBackground(m2dRect *rect, int noImages);
void    drawTitle();
void    drawConsole(sRect *area);
void    drawMenu(sRect *area);
void    drawImages(m2dRect *rect);
int     checkOverlap(sRect *rect1, sRect *rect2);
void    drawAppStuff(sRect *area, int appNum);
void    parseText(char *str, sRect *rectIn, int fontIndex);
void    drawText(char *str, sRect *rectIn, int fontIndex,unsigned long color,int topLine,int align);
void    drawIntro();
void    drawLoading();
m2dRect drawStringCentered(int x, int y, char *str);
int     showIntroImage(int appNum);

// File reading
int     getAppList();
int     getAppData();
void    getUserInput();
void    getAppImages(int appNum);
void    loadApp(int appNum);
void    loadAppFromFP(myFileType *fp, int fileSize, int headerSize);
//void    load(void *ptr);
void*   checkAndReadImg(char *dir, char *filename);
int     checkImg(char *dir, char *filename);
void*   checkAndReadFile(char *dir, char *filename, eFileType fileType);

// Console operations
void    insertConsoleString(char *str);
void    appendConsoleString(char *str);

// Data file functions
void    checkDataFile();
long    getChecksum(char *buffer, int bytes);
void*   getDataFilePtr(char *filename, int *fileSize);

// Misc utilities
int           getFileSize   (myFileType *fp);
char*         myReadFile    (myFileType *fp, int *fileSize);
myFileType*   myfopen       (char *filename);
int           mygets        (char *dest, int maxDest, char **src, int *maxSrc);
void*         decompressJPEG(void *jpegdata, int jpeg_size, long *imgW, long *imgH);
unsigned long getFreeMem    (int memType);
int           PlayWAV       (void *wavaddr, int loop, PCMPOS *Pan, int volume, int echo);
void          freeAppImages ();
int           imagesReady   (int appNum);
void          redrawImages  (int thisApp, int nextApp);

//==================================================================================================
// Functions
//==================================================================================================

int main()
{
  int appDataDone = FALSE;

  nextState = (Buttons(_Controller[1]) & JOY_NUON) ? kStateConsoleAndMenu : kStateInit;
  state = nextState;

#ifndef OLD_TGA
  _MPELoad(MPE2, (void *)0x20300000, tgac_start, (long) tgac_size);
  _MPELoad(MPE2, (void *)0x20100000, tgad_start, (long) tgad_size);
  _MPERun (MPE2, (void *)0x20300000);

  cbMagicNum = _IntSetVector(kIntrCommRecv, TGADoneCommHandler);
#endif
  
  // Initialize stuff
  localScratch = _MemLocalScratch(&localSize);
  initDisplay();
  initFonts();
  initColors();
  initConsole();   // Don't insert anything into the console before this
  initCDLib();
  initLoadingBG();
  initStrings();
  initSystem();
#ifndef NO_AUDIO
  AUDIOInit();
#endif
  
  // Open and verify bootload.dat
  checkDataFile();
  if (!bufDataFile) {
    // On error just hang
    insertConsoleString("**ERR** Could not find bootload.dat");
    while (1) updateScreen();
  }

  // Parse applist.txt
  if (getAppList()) {
    // On error just hang
    insertConsoleString("HANGING PROGRAM");
    while (1) updateScreen();
  }

  // Make sure each app has an executable and parse the description text to get the title
  while (!appDataDone) {
    appDataDone = getAppData();
    updateScreen();
  }

  // Done initializing stuff.  Switch to the normal console display.
  insertConsoleString("Finished reading titles & descriptions");
  if (state == kStateInit) {
    // Redraw everyting in case the background changed
    redraw |= kRedrawAll;
    nextState = kStateMenuAndStuff;
    if (LOADING_IMG_PTR) {
      mmlReleasePixmaps((mmlPixmap *) LOADING_IMG_PTR, &sysRes, 1);
#ifndef OLD_TGA
      loadingImg.status = kImgStatusInit;
#endif
      LOADING_IMG_PTR = NULL;
    }
  }

  // Infinite loop of getting user input and updating the screen
  insertConsoleString("Entering main loop");
  while (1) {
    blinkCursor = FALSE;
    getAppImages(curApp);        // Construct the app images if the user moved the highlight
    getUserInput();

    redraw |= kRedrawHighlight;  // The one thing that always gets redrawn
    blinkCursor = TRUE;
    updateScreen();
  }

  return (0);
}

//==================================================================================================
// initSystem()
//
// Detects what chip is being used.  It reads this info from MPE1 since it is not being used.
// Reading from a running MPE can crash it.
//
// Inputs:  none
// Outpust: none
// Globals modified:
//   ariesVer - the chip revision
//
//==================================================================================================

void initSystem()
{
#ifdef TRACE
  printf("%s() called\n", __func__);
#endif
  
  ariesVer = (_MPEReadRegister(1, (void*) 0x20500ff0) >> 24) & 0xFF;
  switch (ariesVer) {
    case 1:  insertConsoleString("Aries 1 detected");  break;
    case 3:  insertConsoleString("Aries 2 detected");  break;
    case 4:  insertConsoleString("Aries 3 detected");  break;
    default:
      insertConsoleString("Unknown Aries version: ");
      msprintf(tmpStr, "%d", ariesVer);
      appendConsoleString(tmpStr);
      break;
  }
}

//==================================================================================================
// initDisplay()
//
// Allocates frame buffers and sets up the display.  Clears both buffers.
//
// Inputs:  none
// Outputs: none
// Globals modified:
//   sysRes        - system resources
//   gc            - graphics context
//   vd            - video display
//   frameBuffer[] - frame buffer pointers
//   vcMain        - main channel video structure
//
//==================================================================================================

void initDisplay()
{
  int i;
  m2dRect rect;
  mmlColor color;

  mmlPowerUpGraphics(&sysRes);
  mmlInitGC(&gc, &sysRes);

  _VidQueryConfig(&vd);
  
  mmlInitDisplayPixmaps(&frameBuffer[0], &sysRes, SCREEN_W, SCREEN_H, e888Alpha, 1, NULL);
  mmlInitDisplayPixmaps(&frameBuffer[1], &sysRes, SCREEN_W, SCREEN_H, e888Alpha, 1, NULL);
  drawBuffer = 0;
  
  vcMain.dmaflags    = DMAFLAGS_RD;
  vcMain.base        = frameBuffer[0].memP;
  vcMain.dest_xoff   = -1;
  vcMain.dest_yoff   = -1;
  vcMain.dest_width  = vd.dispwidth;
  vcMain.dest_height = vd.dispheight;
  vcMain.src_xoff    = 0;
  vcMain.src_yoff    = 0;
  vcMain.src_width   = SCREEN_W;
  vcMain.src_height  = SCREEN_H;
  vcMain.clut_select = 0;
  vcMain.alpha       = 0;
  vcMain.vfilter     = VID_VFILTER_2TAP;
  vcMain.hfilter     = VID_HFILTER_4TAP;
  for(i = 0; i < 5; i++) vcMain.reserved[i] = 0;

  _VidConfig(&vd, &vcMain, NULL, NULL);

  rect.leftTop.x = 0;
  rect.leftTop.y = 0;
  rect.rightBot.x = SCREEN_W - 1;
  rect.rightBot.y = SCREEN_H - 1;
#ifdef USEINITBG
  color = mmlColorFromRGB(0, 0, 0);
#else
  color = mmlColorFromRGB(myBlue.r, myBlue.g, myBlue.b);
#endif
  m2dFillColor(&gc, &frameBuffer[drawBuffer], NULL, color);
  VideoSync();
  m2dFillColor(&gc, &frameBuffer[drawBuffer], NULL, color);

#ifndef OLD_TGA
  loadingImg.status = kImgStatusInit;
#endif
  LOADING_IMG_PTR = NULL;

#ifndef OLD_TGA
  bgImg.status = kImgStatusInit;
#endif
  BG_IMG_PTR = NULL;
}

//==================================================================================================
// initFonts()
//
// Sets up the font context and instantiates the system font.
//
// Inputs:  none
// Outputs: none
// Globals modified:
//   fc      - font context
//   aFont[] - array of installed fonts
//
//==================================================================================================

void initFonts()
{
  int i;

  mmlInitFontContext(&gc, &sysRes, &fc, 16384);  

  // Font 0 is always the system font
  aFont[0] = mmlAddFont(fc, "sysFont", eT2K, SysFont, 40000);

  for (i = 0; i < MAX_EMBEDDED_FONTS; i++) {
    aFont[1+MAX_USER_FONTS+i] = mmlAddFont(fc, aEmbeddedFonts[i].name, eTrueType, 
					   aEmbeddedFonts[i].ptr, aEmbeddedFonts[i].size);
  }
}

//==================================================================================================
// initUserFont()
//
// Inits the user fonts, if any.
//
// Inputs:  none
// Outputs: none
// Globals modified:
//   aFont[] - array of installed fonts
//
//==================================================================================================

void initUserFont(int i, char *ttf, int size)
{
  char *fontName = "fontX";

  fontName[4] = '1' + i;
  aFont[i+1] = mmlAddFont(fc, fontName, eTrueType, ttf, size);
}

//==================================================================================================
// initConsole()
//
// Clears the console strings and sets up the X/W elements of the right and left areas.
//
// Inputs:  none
// Outputs: none
// Globals modified:
//   rectLeft   - defines the left area (normally the menu)
//   rectRight  - defines the right area (normally the images and description)
//   aConsole[] - array of console strings
//
//==================================================================================================

void initConsole()
{
  int i;

  // y & h elements will be filled in later
  rectLeft.x = aBorders[kBorderLeft];
  rectLeft.w = ((SCREEN_W - aBorders[kBorderLeft] - aBorders[kBorderRight]) >> 1) - ELEMENT_SPACE;

  rectRight.x = rectLeft.x + rectLeft.w + (ELEMENT_SPACE << 1);
  rectRight.w = rectLeft.w;

  for (i = 0; i < CONSOLE_LINES_MAX; i++) aConsole[i][0] = 0;

  insertConsoleString("Console initialized");
}

//==================================================================================================
// initColors()
//
// Converts the sColor array in RGB to a mmlColor in YCrCb.
//
// Inputs:  none
// Outputs: none
// Globals modified:
//   aColors[] - array of colors
//
//==================================================================================================

void initColors()
{
  int i;
  for (i = 0; i < kColorMax; i++)
    aColors[i] = mmlColorFromRGB(aColorDefaults[i].r, aColorDefaults[i].g, aColorDefaults[i].b);
}

//==================================================================================================
// initStrings()
//
// Allocates space fo the strings and sets them to their defaults.  Constructs the About string
// from the author, version, and current date.
//
// Inputs:  none
// Outputs: none
// Globals modified:
//   aStrings[] - array of strings
//   strAbout   - the About string
//   instrStr   - pointer to the instruction string
//
//==================================================================================================

void initStrings()
{
  int i;

#ifdef TRACE
  printf("%s() called\n", __func__);
#endif
  
  // Clear them all to NULL first
  for (i = 0; i < kStrMax; i++) {
    aStrings[i] = NULL;
  }

  // Now allocate space for the default strings and copy them in.
  for (i = 0; i < kStrMax; i++) {
    if (aStringDefaults[i]) {
      aStrings[i] = (char *) malloc (sizeof(char) * (strlen(aStringDefaults[i]) + 1));
      if (!aStrings[i]) {
	insertConsoleString("**ERR** Could not allocate string space");
      } else {
	aStrings[i]++;  // allow one character for the parse flag
	strcpy(aStrings[i], aStringDefaults[i]);
	aStrings[i][-1] = 0;  // flag that indicates this string needs to be parsed
      }
    }
  }

  // Special case the About string since it is not changeable
  strAbout = (char *) malloc(sizeof(char) * (strlen(strVersion) + strlen(strAuthor) +
					     strlen(__DATE__) + 2));
  strAbout++;  // allow one character for the parse flag
  strcpy(strAbout, strAuthor);
  strcat(strAbout, strVersion);
  strcat(strAbout, __DATE__);
  strAbout[-1] = 0;  // flag that indicates this string needs to be parsed

  instrStr = (controlType == kControlRemote)
    ? aStrings[kStrInstr1Remote] : aStrings[kStrInstr1Normal];
}

//==================================================================================================
// initCDLib()
//
// Initializes the CD library.  Does nothing on a dev system due to the #ifdef's.
//
// Inputs:  none
// Outputs: none
// Globals modified:
//   cdError - is passed into every CD library function and gets set to non-zero on error
//
//==================================================================================================

void initCDLib()
{
#ifdef TRACE
  printf("%s\n", __func__);
#endif

  cdError = 0;
  
#ifndef USE_LOCAL_HOST
  cdError = CDInit();
  if (cdError) {
    insertConsoleString("**ERR** CDInit failed!");
    while (1) updateScreen();
  }
  
  cdContextPtr = iso_init(0, &cdError);
  if (cdError) {
    insertConsoleString("**ERR** iso_init failed!");
    while (1) updateScreen();
  }
#endif
}

//==================================================================================================
// initLoadingBG()
//
// Gets the background image for the initialization screen if necessary.
//
// Inputs:  none
// Outputs: none
// Globals modified:
//   loadingImg - pointer to the display pixmap for the background loading image
//
//==================================================================================================

void initLoadingBG()
{
#ifdef USEINITBG
  long dmaFlags;
  mmlStatus status;
#ifndef OLD_TGA
  long sec, prevSec;
#endif
#endif
  
#ifndef OLD_TGA
  loadingImg.status = kImgStatusInit;
#endif
  LOADING_IMG_PTR = NULL;

#ifdef USEINITBG
#ifdef TRACE
  printf("%s() begin\n", __func__);
#endif

  // Get the image size
  //  GetTGASize(LoadingImg, &imgW, &imgH);
  g_imgW = (unsigned char) LoadingImg[12] | ((unsigned char) LoadingImg[13] << 8);
  g_imgH = (unsigned char) LoadingImg[14] | ((unsigned char) LoadingImg[15] << 8);

  if ((g_imgW != SCREEN_W) || (g_imgH != SCREEN_H)) {
    msprintf(tmpStr, "**ERR** Background loading image not %dx%d", SCREEN_W, SCREEN_H);
    insertConsoleString(tmpStr);
    return;
  }

  LOADING_IMG_PTR = (mmlDisplayPixmap *) malloc(sizeof(mmlDisplayPixmap));
  if (!LOADING_IMG_PTR) {
    insertConsoleString("**ERR** Could not allocate pixmap for loading image");
    return;
  }

  status = mmlInitDisplayPixmaps(LOADING_IMG_PTR, &sysRes, SCREEN_W, SCREEN_H, e888Alpha, 1,NULL);
  if (status != eOK) {
    free(LOADING_IMG_PTR);
    LOADING_IMG_PTR = NULL;
    insertConsoleString("**ERR** mmlInitDisplayPixmaps failed for loading image");
    return;
  }
      
  // Got the image space we need, now decode the TGA into that memory.
  dmaFlags = DMA_PIXEL_WRITE | ((SCREEN_W/8)<<16) | (4<<4) | (1<<11);

#ifdef OLD_TGA
  if (!ReadTGA(LoadingImg, dmaFlags, loadingImg->memP, 0, 0, NULL, 0)) {
    // Error while decoding TGA
    mmlReleasePixmaps((mmlPixmap *) loadingImg, &sysRes, 1);
    free(loadingImg);
    loadingImg = NULL;
    insertConsoleString("**ERR** Error decoding TGA for loading image");
    return;
  }
#else
  loadingImg.status = kImgStatusSubmitted;

  _CommSendDirect((long) LoadingImg, (long) loadingImg.pixmap->memP, dmaFlags, (int) &loadingImg,
		  MPE2, 0);

  // Wait for helper MPE to finish before proceeding
  _TimeElapsed(&prevSec, NULL);
  while (loadingImg.status == kImgStatusSubmitted) {
    // Time out after a bit
    _TimeElapsed(&sec, NULL);
    if ((sec - prevSec) > 5) {
      insertConsoleString("**ERR** Loading image render timeout");
      break;
    }
  }

  // If it didn't work then free the memory and just use a background color instead
  if (loadingImg.status != kImgStatusDrawn) {
    insertConsoleString("**ERR** Loading image render failed");
    mmlReleasePixmaps((mmlPixmap *) loadingImg.pixmap->memP, &sysRes, 1);
    free(loadingImg.pixmap);
    loadingImg.pixmap = NULL;
  }
#endif

  insertConsoleString("Got loading background image");
  
#ifdef TRACE
  printf("%s() end\n", __func__);
#endif

#endif
}

//==================================================================================================
// screenshotCheck()
//
// Gets called every video sync.  If the user presses the L, R, and NUON buttons simultaneously
// the current frame is copied to a location in system memory in raster order.  The address and
// dimensions of the buffer are printed so the user can download it.
//
// Is controlled by an #ifdef so it does not function on a real player.  ** DEV SYSTEM ONLY **
//
// Inputs:  none
// Outputs: none
// Globals modified:
//   screenshotBuffer - pointer to the screenshotBuffer
//
//==================================================================================================

void screenshotCheck()
{
#ifdef DSI_DEBUG
  
  int x, y;
  char *ptr;
  static int stillDown = 0;
  static int printed = 0;
  int size;
  
  if ((_Controller[1].buttons & JOY_L) &&
      (_Controller[1].buttons & JOY_R) &&
      (_Controller[1].buttons & JOY_NUON)) {
    if (!printed) {
      size = SCREEN_W * SCREEN_H * 4;
      screenshotBuffer = malloc(size);
      if (!screenshotBuffer) {
	printf("\nERROR: Could not allocate %d bytes for screenshot buffer",
	       SCREEN_W * SCREEN_H * 4);
      }
      printf("\nBuffer address=%08X  Size=%d (%d scalars)  W,H=%d,%d",
	     (int) screenshotBuffer, size, size>>2, SCREEN_W, SCREEN_H);
      fflush(stdout);
      printed = 1;
    }

    if (!stillDown) {
      stillDown = 1;
      ptr = screenshotBuffer;
      for (y = 0; y < SCREEN_H; y++)
	for (x = 0; x < SCREEN_W; x += 8) {
	  _DMABiLinear(DMAFLAGS_RD, frameBuffer[drawBuffer].memP, (8<<16)|(x<<0), (1<<16)|(y<<0),
		       localScratch);
	  _DMALinear((8<<16), ptr, localScratch);
	  ptr += 8*4;
	}
    }
  } else {
    stillDown = 0;
  }
  
#endif
}

//==================================================================================================
// VideoSync()
//
// Swaps frame buffers and waits until a new field starts before returning.  Keeps track of the
// frame rate (fps).
//
// Inputs:  none
// Outputs:
//   int - number of frames that passed since last sync
// Globals modified:
//   fps         - frame rate over the last second
//   fieldsAccum   - field count for use by button delay
//   cursorFields - field count to change the highlight color
//
//==================================================================================================

int VideoSync(void)
{
  static int frames;
  static int lastfield;
  static int lastsecond;
  int ret;
  int curfield;

  // First check if we should capture a screenshot
  screenshotCheck();

  // we want to view the buffer we just finished drawing
  //  _VidChangeBase(VID_CHANNEL_MAIN, vcMain.dmaflags, &frameBuffer[drawBuffer]);
  vcMain.base = frameBuffer[drawBuffer].memP;
  _VidConfig(&vd, &vcMain, NULL, NULL);
  
  // find out how many fields have been drawn
  curfield = _VidSync(-1);

  // remember that we've drawn a frame
  frames++;

  // have more than 60 fields (1 second) passed?
  if ((curfield - lastsecond) >= 60) {
    fps = frames;
    frames = 0;
    lastsecond = curfield;
  }

  // draw to a new buffer
  dispBuffer = drawBuffer;
  drawBuffer++;
  if (drawBuffer >= NUM_BUFFERS) drawBuffer = 0;

#if NUM_BUFFERS == 2
  // wait until a new field starts (so it's no longer showing the field we are going to draw on)
  curfield = _VidSync(0);
#endif

  ret = curfield - lastfield;
  while (ret == 0) {
    curfield = _VidSync(-1);
    ret = curfield - lastfield;
  }
  lastfield = curfield;

  fieldsAccum += ret;
  if (blinkCursor && (ret < 5)) cursorFields += ret;

  return ret;
}

//==================================================================================================
// updateScreen()
//
// Paints the screen.  There are four states the program may be in.  The normal two are the
// initialization screen and the menu/image gui.  The other two states display the console.  This
// may be entered in one of several ways:
//   o The user can cycle through the last three states with the NUON button
//   o If the applist.txt file included the parameter +ShowConsole then it will default to shown
//   o When the program detects an error it will automatically switch to the console
//
// Each draw*() function will only redraw its portion if the appropriate bit in the "redraw" flags
// is set.
//
// Inputs:  none
// Outputs: none
// Globals modified:
//   redraw - shifts the redraw flags so each element is drawn for exactly two fields
//   state  - the current program state
//
//==================================================================================================

void updateScreen()
{
  static int noRecurse = FALSE;
  
#ifdef TRACE
  //  printf("%s begin\n", __func__);
#endif

  // Some of the functions here can end up calling "updateScreen()" by
  // inserting messages in the console.  Prevent silly recursion errors
  // by disallowing this to happen.
  if (noRecurse) {
#ifdef DSI_DEBUG
    printf("\n**ERROR** Recursed in updateScreen()!");  fflush(stdout);
#endif
    
  } else {
    noRecurse = TRUE;

    switch (state) {
      case kStateInit:
	// Draw the initialization screen
	drawInit();
	break;
      case kStateMenuAndStuff:
	// Draw the menu on the left and images/description on the right (TYPICAL STATE)
	drawBackground(NULL, FALSE);
	drawTitle     ();
	drawMenu      (&rectLeft);
	drawAppStuff  (&rectRight, curApp);
	break;
      case kStateConsoleAndMenu:
	// Draw the console on the left and menu on the right
	drawBackground(NULL, FALSE);
	drawTitle     ();
	drawConsole   (&rectLeft);
	drawMenu      (&rectRight);
	drawAppStuff  (NULL, curApp);
	break;
      case kStateConsoleAndStuff:
	// Draw the console on the left and images/description on the right
	drawBackground(NULL, FALSE);
	drawTitle     ();
	drawConsole   (&rectLeft);
	drawMenu      (NULL);
	drawAppStuff  (&rectRight, curApp);
	break;
      case kStateIntro:
	drawIntro     ();
	break;
      case kStateLoading:
	drawBackground(NULL, FALSE);
	drawLoading   ();
	break;
    }
    redraw = (redraw & (kRedrawAll | (kRedrawAll<<1))) << 1;
    state = nextState;

    VideoSync();

    noRecurse = FALSE;
  }
  
#ifdef TRACE
  //  printf("%s end\n", __func__);
#endif
}

//==================================================================================================
// drawInit()
//
// Draws the initialization screen.  The third line is changeable so the program can update the
// user on what it is doing.
//
// Inputs:  none
// Outputs: none
// Globals modified: none
//
//==================================================================================================

void drawInitNormal()
{
  mmlColor color;
  m2dRect rectInit1;
  m2dRect rectInit2;
  m2dRect rectInit3;
  m2dRect clearRect;
  static int prevY = -1;

  if (checkRedraw(kRedrawAll)) {
    // Get the extents of the title string (set font size first)
    mmlSetTextProperties(fc, aFont[fontIndexInit1], 50, DC, DC, eBlend, 0, UNUSED);
    rectInit1.leftTop.x = 0;
    rectInit1.leftTop.y = 0;
    rectInit1.rightBot.x = SCREEN_W - 1;
    rectInit1.rightBot.y = SCREEN_H - 1;
    mmlGetTextBox(fc, strInit1, 0, strlen(strInit1)-1, &rectInit1);

    // Get the extents of the text (set font size first)
    mmlSetTextProperties(fc, aFont[fontIndexInit2], 30, DC, DC, eBlend, 0, UNUSED);
    rectInit2.leftTop.x = 0;
    rectInit2.leftTop.y = 0;
    rectInit2.rightBot.x = SCREEN_W - 1;
    rectInit2.rightBot.y = SCREEN_H - 1;
    mmlGetTextBox(fc, strInit2, 0, strlen(strInit2)-1, &rectInit2);


    // Clear the screen
    color = mmlColorFromRGB(myBlue.r, myBlue.g, myBlue.b);
    m2dFillColor(&gc, &frameBuffer[drawBuffer], NULL, color);


    // Set the color to display the program title
    color = mmlColorFromRGB(myYellow.r, myYellow.g, myYellow.b);
    mmlSetTextProperties(fc, aFont[fontIndexInit1], 50, color, DC, eBlend, 0, UNUSED);

    // Draw the title string
    rectInit1.leftTop.x += (SCREEN_W - (rectInit1.rightBot.x - rectInit1.leftTop.x)) >> 1;
    rectInit1.leftTop.y += (SCREEN_H >> 1) - (rectInit1.rightBot.y - rectInit1.leftTop.y);
    rectInit1.rightBot.x = SCREEN_W - 1;
    rectInit1.rightBot.y = SCREEN_H - 1;
    mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], strInit1, strlen(strInit1), &rectInit1);
    
    // Set the color for "Initializing..."
    color = mmlColorFromRGB(myWhite.r, myWhite.g, myWhite.b);
    mmlSetTextProperties(fc, aFont[fontIndexInit2], 30, color, DC, eBlend, 0, UNUSED);

    // Draw "Initializing..."
    rectInit2.leftTop.x += (SCREEN_W - (rectInit2.rightBot.x - rectInit2.leftTop.x)) >> 1;
    rectInit2.leftTop.y += SCREEN_H >> 1;
    rectInit2.rightBot.x = SCREEN_W - 1;
    rectInit2.rightBot.y = SCREEN_H - 1;
    mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], strInit2, strlen(strInit2), &rectInit2);
    
    prevY = rectInit2.rightBot.y;
  }

  if (checkRedraw(kRedrawInit3)) {
    // Set the color to display the action description
    color = mmlColorFromRGB(myWhite.r, myWhite.g, myWhite.b);
    mmlSetTextProperties(fc, aFont[fontIndexInit3], 30, color, DC, eBlend, 0, UNUSED);

    // Draw the text that indicates what is being initialized
    if (strInit3[0] || (prevY == -1)) {
      rectInit3.leftTop.x = 0;
      rectInit3.leftTop.y = 0;
      rectInit3.rightBot.x = SCREEN_W - 1;
      rectInit3.rightBot.y = SCREEN_H - 1;
      mmlGetTextBox(fc, strInit3, 0, strlen(strInit3)-1, &rectInit3);

      rectInit3.leftTop.x += (SCREEN_W - (rectInit3.rightBot.x - rectInit3.leftTop.x)) >> 1;
      rectInit3.leftTop.y  = prevY;
      rectInit3.rightBot.x = SCREEN_W - 1;
      rectInit3.rightBot.y = SCREEN_H - 1;

      clearRect.leftTop.x  = 0;
      clearRect.leftTop.y  = rectInit3.leftTop.y;
      clearRect.rightBot.x = SCREEN_W - 1;
      clearRect.rightBot.y  = rectInit3.rightBot.y;
      color = mmlColorFromRGB(myBlue.r, myBlue.g, myBlue.b);
      m2dFillColor(&gc, &frameBuffer[drawBuffer], &clearRect, color);

      mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], strInit3, strlen(strInit3), &rectInit3);
    }
  }
}

//==================================================================================================

void drawInitGamesAndDemos()
{
  mmlColor color;
  m2dRect rectInit1;
  m2dRect rectInit2;
  m2dRect rectInit3;
  m2dRect clearRect;
  m2dPoint point;
  static int prevY = -1;

  if (checkRedraw(kRedrawAll)) {
    // Get the extents of the title string (set font size first)
    mmlSetTextProperties(fc, aFont[fontIndexInit1], 50, DC, DC, eBlend, 0, UNUSED);
    rectInit1.leftTop.x = 0;
    rectInit1.leftTop.y = 0;
    rectInit1.rightBot.x = SCREEN_W - 1;
    rectInit1.rightBot.y = SCREEN_H - 1;
    mmlGetTextBox(fc, strInit1, 0, strlen(strInit1)-1, &rectInit1);

    // Get the extents of the text (set font size first)
    mmlSetTextProperties(fc, aFont[fontIndexInit2], 40, DC, DC, eBlend, 0, UNUSED);
    rectInit2.leftTop.x = 0;
    rectInit2.leftTop.y = 0;
    rectInit2.rightBot.x = SCREEN_W - 1;
    rectInit2.rightBot.y = SCREEN_H - 1;
    mmlGetTextBox(fc, strInit2, 0, strlen(strInit2)-1, &rectInit2);

    // Redraw the background
    clearRect.leftTop.x = 0;
    clearRect.leftTop.y = 0;
    clearRect.rightBot.x = SCREEN_W - 1;
    clearRect.rightBot.y = SCREEN_H - 1;
    point.x = clearRect.leftTop.x;
    point.y = clearRect.leftTop.y;
    m2dCopyRectDis(&gc, LOADING_IMG_PTR, &frameBuffer[drawBuffer], &clearRect, point);
      
    // Set the color for "Initializing..."
    color = mmlColorFromRGB(colorInit2.r, colorInit2.g, colorInit2.b);
    mmlSetTextProperties(fc, aFont[fontIndexInit2], 40, color,DC,eBlend,0,UNUSED);

    // Draw "Initializing..."
    rectInit2.leftTop.x += (SCREEN_W - (rectInit2.rightBot.x - rectInit2.leftTop.x)) >> 1;
    rectInit2.leftTop.y = 380;
    rectInit2.rightBot.x = SCREEN_W - 1;
    rectInit2.rightBot.y = SCREEN_H - 1;
    mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], strInit2, strlen(strInit2), &rectInit2);
    
    prevY = rectInit2.rightBot.y;
  }

  if (checkRedraw(kRedrawInit3)) {
    // Set the color to display the action description
    color = mmlColorFromRGB(colorInit3.r, colorInit3.g, colorInit3.b);
    mmlSetTextProperties(fc, aFont[fontIndexInit3], 40, color, DC, eBlend, 0, UNUSED);

    // Draw the text that indicates what is being initialized
    if (strInit3[0] || (prevY == -1)) {
      rectInit3.leftTop.x = 0;
      rectInit3.leftTop.y = 0;
      rectInit3.rightBot.x = SCREEN_W - 1;
      rectInit3.rightBot.y = SCREEN_H - 1;
      mmlGetTextBox(fc, strInit3, 0, strlen(strInit3)-1, &rectInit3);

      rectInit3.leftTop.x += (SCREEN_W - (rectInit3.rightBot.x - rectInit3.leftTop.x)) >> 1;
      rectInit3.leftTop.y  = prevY;
      rectInit3.rightBot.x = SCREEN_W - 1;
      rectInit3.rightBot.y = SCREEN_H - 1;

      clearRect.leftTop.x  = 0;
      clearRect.leftTop.y  = rectInit3.leftTop.y;
      clearRect.rightBot.x = SCREEN_W - 1;
      clearRect.rightBot.y  = rectInit3.rightBot.y;
      color = mmlColorFromRGB(0, 0, 0);
      m2dFillColor(&gc, &frameBuffer[drawBuffer], &clearRect, color);

      mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], strInit3, strlen(strInit3), &rectInit3);
    }
  }
}

//==================================================================================================

void drawInitYarozeClassics()
{
  mmlColor color;
  m2dRect rectInit1;
  m2dRect rectInit2;
  m2dRect rectInit3;
  m2dRect clearRect;
  m2dPoint point;

  if (checkRedraw(kRedrawAll)) {
    // Get the extents of the title string (set font size first)
    mmlSetTextProperties(fc, aFont[fontIndexInit1], 50, DC, DC, eBlend, 0, UNUSED);
    rectInit1.leftTop.x = 0;
    rectInit1.leftTop.y = 0;
    rectInit1.rightBot.x = SCREEN_W - 1;
    rectInit1.rightBot.y = SCREEN_H - 1;
    mmlGetTextBox(fc, strInit1, 0, strlen(strInit1)-1, &rectInit1);

    // Get the extents of the text (set font size first)
    mmlSetTextProperties(fc, aFont[fontIndexInit2], 40, DC, DC, eBlend, 0, UNUSED);
    rectInit2.leftTop.x = 0;
    rectInit2.leftTop.y = 0;
    rectInit2.rightBot.x = SCREEN_W - 1;
    rectInit2.rightBot.y = SCREEN_H - 1;
    mmlGetTextBox(fc, strInit2, 0, strlen(strInit2)-1, &rectInit2);

    // Redraw the background
    clearRect.leftTop.x = 0;
    clearRect.leftTop.y = 0;
    clearRect.rightBot.x = SCREEN_W - 1;
    clearRect.rightBot.y = SCREEN_H - 1;
    point.x = clearRect.leftTop.x;
    point.y = clearRect.leftTop.y;
    m2dCopyRectDis(&gc, LOADING_IMG_PTR, &frameBuffer[drawBuffer], &clearRect, point);
      
    // Set the color for "Initializing..."
    color = mmlColorFromRGB(colorInit2.r, colorInit2.g, colorInit2.b);
    mmlSetTextProperties(fc, aFont[fontIndexInit2], 40, color, DC, eBlend, 0, UNUSED);

    // Draw "Initializing..."
    rectInit2.leftTop.x = 400;
    rectInit2.leftTop.y = 350;
    rectInit2.rightBot.x = SCREEN_W - 1;
    rectInit2.rightBot.y = SCREEN_H - 1;
    mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], strInit2, strlen(strInit2), &rectInit2);
  }

  if (checkRedraw(kRedrawInit3)) {
    // Set the color to display the action description
    color = mmlColorFromRGB(colorInit3.r, colorInit3.g, colorInit3.b);
    mmlSetTextProperties(fc, aFont[fontIndexInit3], 40, color, DC, eBlend, 0, UNUSED);

    // Draw the text that indicates what is being initialized
    if (strInit3[0]) {

      clearRect.leftTop.x = 460;
      clearRect.leftTop.y = 390;
      clearRect.rightBot.x = SCREEN_W - 1;
      clearRect.rightBot.y = SCREEN_H - 1;
      point.x = clearRect.leftTop.x;
      point.y = clearRect.leftTop.y;
      m2dCopyRectDis(&gc, LOADING_IMG_PTR, &frameBuffer[drawBuffer], &clearRect, point);

      if ((strcmp(strInit3, "/Invs/") == 0) || (strcmp(strInit3, "Invs") == 0)) {
	msprintf(strInit3, "Invs");
	rectInit3.leftTop.x = 500;
	rectInit3.leftTop.y = 200;
      } else if ((strcmp(strInit3, "/Katapila/") == 0) ||
		 (strcmp(strInit3, "Katapila") == 0)) {
	msprintf(strInit3, "Katapila");
	rectInit3.leftTop.x = 170;
	rectInit3.leftTop.y = 340;
      } else {
	rectInit3.leftTop.x = 460;
	rectInit3.leftTop.y = 390;
      }
      rectInit3.rightBot.x = SCREEN_W - 1;
      rectInit3.rightBot.y = SCREEN_H - 1;
      mmlGetTextBox(fc, strInit3, 0, strlen(strInit3)-1, &rectInit3);

      mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], strInit3, strlen(strInit3), &rectInit3);
    }
  }
}

//==================================================================================================
// drawBackground()
//
// Repaints the background over the portion of the screen described by the input rectangle.  If the
// rectangle pointer is NULL then the entire screen will be repainted.
//
// Optimization: If the full background is repainted (via kRedrawAll) then do not repaint any
// subsequent rectangles since it's already been done.
//
// Inputs:
//   rectIn - the area in which to repaint the background
//   noImages - if set to TRUE, will not redraw images after clearing background
// Outputs: none
// Globals modified: none
//
//==================================================================================================

void drawBackground(m2dRect *rectIn, int noImages)
{
  m2dPoint point;
  m2dRect thisRect;
  m2dRect *rect;

#ifdef TRACE
  if (rectIn) {
    printf("%s() begin ({%d,%d,%d,%d}, %d)\n", __func__, rectIn->leftTop.x, rectIn->leftTop.y,
	   rectIn->rightBot.x, rectIn->rightBot.y, noImages);
  } else {
    printf("%s() begin (NULL, %d)\n", __func__, noImages);
  }
#endif

  // Only redraw this rectangle under one of two conditions:
  //   o the rectangle is NULL, indicating a full repaint
  //   o the rectangle is non-NULL *and* the full screen has not been repainted
  if (!checkRedraw(kRedrawAll) && !rectIn) {
    // Still want to draw images
    if (!noImages) {
      drawImages(rectIn);
    }
#ifdef TRACE
  printf("%s() end\n", __func__);
#endif
    return;
  }

  // If the rectangle pointer passed in is NULL then repaint the entire background
  thisRect.leftTop.x = 0;
  thisRect.leftTop.y = 0;
  thisRect.rightBot.x = SCREEN_W - 1;
  thisRect.rightBot.y = SCREEN_H - 1;
  rect = rectIn ? rectIn : &thisRect;

  // Choose whether to draw an image or a constant color
  if (BG_IMG_PTR) {
    point.x = rect->leftTop.x;
    point.y = rect->leftTop.y;
    m2dCopyRectDis(&gc, BG_IMG_PTR, &frameBuffer[drawBuffer], rect, point);
    
  } else {
    m2dFillColor(&gc, &frameBuffer[drawBuffer], rect, aColors[kColorBg]);
  }

  // Now draw any images on top of it
  if (!noImages) {
    drawImages(rectIn);
  }

#ifdef TRACE
  printf("%s() end\n", __func__);
#endif
}

//==================================================================================================
// drawTitle()
//
// Draws the title text and instructions.  Lots of coordinates for things are set up here.  This is 
// likely only called a few times since the title is static.
//
// Inputs:  none
// Outputs: none
// Globals modified:
//   coordTitle - center coordinates of the title text
//   coordInstr - center coordinates of the instruction text
//   rectLeft   - initializes the Y/H values of the menu regions
//   rectRight  - ditto
//
//==================================================================================================

void drawTitle()
{
  m2dRect rect;
  static int maxY = -1;
  
  // Don't redraw unless necessary
  if (!checkRedraw(kRedrawTitle)) return;

#ifdef TRACE
  printf("%s() begin\n", __func__);
#endif

  // No instructions chosen yet
  if (!instrStr) {
#ifdef TRACE
    printf("%s() end\n", __func__);
#endif
    redraw |= kRedrawTitle;
    return;
  }

  // When using absolute positioning, just draw the text wherever we are told
  if (absolutePositioning) {
    drawText(aStrings[kStrTitle], &aPos[kPosTitle], kStrFontIndexTitle, aColors[kColorTitle], 0,
	     aTextHAlign[kTextHAlignTitle]);
    drawText(instrStr, &aPos[kPosInstr], kStrFontIndexInstr, aColors[kColorInstr], 0, 
	     aTextHAlign[kTextHAlignInstr]);
    if (fastMode == 1) {
      drawText(aStrings[kStrAries3], &aPos[kPosAries3], kStrFontIndexAries3, aColors[kColorAries3],
	       0, aTextHAlign[kTextHAlignAries3]);
    } else if (fastMode == 0) {
      rect.leftTop.x = aPos[kPosAries3].x;
      rect.leftTop.y = aPos[kPosAries3].y;
      rect.rightBot.x = rect.leftTop.x + aPos[kPosAries3].w - 1;
      rect.rightBot.y = rect.leftTop.y + aPos[kPosAries3].h - 1;
      drawBackground(&rect, FALSE);
    }

  } else {
    // Need to go through this function once before we know exactly where to clear
    if (maxY != -1) {
      rect.leftTop.x = 0;
      rect.leftTop.y = 0;
      rect.rightBot.x = SCREEN_W - 1;
      rect.rightBot.y = maxY;
      drawBackground(&rect, FALSE);
    }
  
    coordTitle.x =
      aBorders[kBorderRight] + ((SCREEN_W - aBorders[kBorderRight] - aBorders[kBorderLeft]) >> 1);
    coordTitle.y = aBorders[kBorderTop];
    
    // Draw program title
    if (aStrings[kStrTitle]) {
      mmlSetTextProperties(fc, aFont[aFontIndices[kStrFontIndexTitle]], aFontSizes[kStrFontIndexTitle],
			   //    mmlSetTextProperties(fc, aFont[0], aFontSizes[kStrFontIndexTitle],
			   aColors[kColorTitle], DC, eBlend, 0, UNUSED);
      rect = drawStringCentered(coordTitle.x, coordTitle.y, aStrings[kStrTitle]);
    }

    coordInstr.x = coordTitle.x;
    coordInstr.y = rect.rightBot.y + aSpacing[kSpaceBelowTitle];

    if (instrStr) {
      mmlSetTextProperties(fc, aFont[aFontIndices[kStrFontIndexInstr]], aFontSizes[kStrFontIndexInstr],
			   aColors[kColorInstr], DC, eBlend, 0, UNUSED);
      rect = drawStringCentered(coordInstr.x, coordInstr.y, instrStr);
    } else {
#ifdef DSI_DEBUG
      printf("\n str NULL");
#endif
    }

    maxY = rect.rightBot.y + 2 + aSpacing[kSpaceBelowInstr] - 1;

    rectLeft.y  = maxY + 1;
    rectLeft.h  = SCREEN_H - aBorders[kBorderBottom] - rectLeft.y;
    rectRight.y = rectLeft.y;
    rectRight.h = rectLeft.h;

    // Draw an indicator that we're in fast mode (Aries3 only)
    if (fastMode == 1) {
      rect.leftTop.x = aBorders[kBorderLeft];
      rect.leftTop.y = aBorders[kBorderTop];
      rect.rightBot.x = SCREEN_W - 1;
      rect.rightBot.y = SCREEN_H - 1;
      if (aStrings[kStrAries3]) {
	mmlSetTextProperties(fc, aFont[aFontIndices[kStrFontIndexAries3]], aFontSizes[kStrFontIndexAries3],
			     aColors[kColorAries3], DC, eBlend, 0, UNUSED);
	mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], aStrings[kStrAries3],
			  strlen(aStrings[kStrAries3]), &rect);
      }
    }
  }
  
#ifdef TRACE
  printf("%s() end\n", __func__);
#endif
}

//==================================================================================================
// drawConsole()
//
// Draws the console text strings.  The most recent string is displayed at the bottom, with the
// rest rising up the screen.  The current one is shown in a different color than the others.
// Errors are also highlighted.
//
// Inputs:
//   area - a rectangle defining the area into which to draw the console
// Outputs: none
// Globals modified: none
//
//==================================================================================================

void drawConsole(sRect *areaIn)
{
  int consoleEnd;
  int line;
  m2dRect rect, drawRect;
  mmlColor color;
  sRect *area;

  // Only redraw if the console changed
  if (!checkRedraw(kRedrawConsole)) {
    return;
  }
  
  // Only redraw if the console is active
  if ((state != kStateConsoleAndMenu) && (state != kStateConsoleAndStuff)) {
    return;
  }

#ifdef TRACE
  printf("%s({%ld,%ld,%ld,%ld}) begin\n", __func__, areaIn->x, areaIn->y, areaIn->w, areaIn->h);
#endif

  area = absolutePositioning ? &aPos[kPosConsole] : areaIn;

  /*
  printf("%d %ld,%ld %ld,%ld\n",
	 absolutePositioning,
	 aPos[kPosConsole].x, aPos[kPosConsole].y,
	 aPos[kPosConsole].w, aPos[kPosConsole].h);
  */

  // Redraw the background
  rect.leftTop.x = area->x;
  rect.leftTop.y = area->y;
  rect.rightBot.x = area->x + area->w - 1;
  rect.rightBot.y = area->y + area->h - 1;
  drawBackground(&rect, FALSE);
  
  // Set up initial clipping region
  rect.leftTop.x = 0;
  rect.leftTop.y = 0;
  drawRect.leftTop.x = area->x;

  // First line is highlighted
  color = aColors[kColorConsoleHighlight];

  // Get the font height
  mmlSetTextProperties(fc, aFont[aFontIndices[kStrFontIndexConsole]],
		       aFontSizes[kStrFontIndexConsole], color, DC, eBlend, 0, UNUSED);
  
  // Loop through all the console strings.  Stop drawing when either we've run out of strings or
  // when the previous string didn't fit on the screen.
  consoleEnd = consoleStart + 1;
  if (consoleEnd >= CONSOLE_LINES_MAX) {
    consoleEnd = 0;
  }
  for (line = consoleStart; line != consoleEnd; ) {  // increment handled in loop body

    if (strlen(aConsole[line]) > 0) {
      // Get the extents for this string
      rect.rightBot.x = area->w;
      rect.rightBot.y = area->h;
      mmlGetTextBox(fc, aConsole[line], 0, strlen(aConsole[line])-1, &rect);
    
      // Make sure this line will fit on the screen
      if (rect.rightBot.y < area->h) {
	// Reverse the order of the strings
	drawRect.leftTop.y = area->y + area->h - rect.rightBot.y;
	drawRect.rightBot.x = area->x + area->w - 1;
	drawRect.rightBot.y = area->y + area->h - 1;

	if (!strncmp(aConsole[line], "**ERR**", strlen("**ERR**"))) {
	  color = aColors[kColorConsoleError];
	}

	// Set the color for this line
	mmlSetTextProperties(fc, aFont[aFontIndices[kStrFontIndexConsole]],
			     aFontSizes[kStrFontIndexConsole], color, DC, eBlend, 0, UNUSED);

	mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], aConsole[line],
			  strlen(aConsole[line]), &drawRect);
	rect.leftTop.y = rect.rightBot.y;

	color = aColors[kColorConsoleNormal];
	
	// Go to next line to print
	line--;
	if (line < 0) line = CONSOLE_LINES_MAX - 1;
      } else {
	line = consoleEnd;  // force the loop to fall through
      }
      
    } else {
      line = consoleEnd;  // force the loop to fall through
    }
  }
  
#ifdef TRACE
  printf("%s() end\n", __func__);
#endif
}

//==================================================================================================
// drawMenuLine()
//
// Draws a single menu entry.
//
// Inputs:
//   appNum = the app to draw
//   rect = the area in which to draw
//   flags = modify the drawing (left, right, or center justification)
// Output:
//   Boolean indicating if this entry was drawn or not.
// Globals modified: none
//==================================================================================================

int drawMenuLine(int appNum, m2dRect *rectIn, int flags)
{
  char *titlePtr, *dirName = NULL;
  m2dRect rect, drawRect;
  char y, cr, cb;
  mmlColor color;
  int fontSize;
  int fontIndex;
  int debug = 0;

#ifdef TRACE
  //  printf("%s() begin\n", __func__);
#endif

  // Copy the input rect
  rect = *rectIn;

  // 
  // Get the string to display for this app
  //
  titlePtr = NULL;
  if (aAppInfo[appNum].txtTitle) {

    // Grab the app title from the app.txt file
    titlePtr = aAppInfo[appNum].txtTitle;
    
  } else if (aAppInfo[appNum].dirName) {

    // Display the directory name if the user didn't give a title
    if (aAppInfo[appNum].flags & kFlagSectionHeader) {
      titlePtr = aAppInfo[appNum].dirName;
    } else {
      // Remove leading slash
      dirName = (char *) malloc (sizeof(char) * (strlen(aAppInfo[appNum].dirName) - 1));
      strncpy(dirName, &aAppInfo[appNum].dirName[1], strlen(aAppInfo[appNum].dirName) - 2);
      dirName[strlen(aAppInfo[appNum].dirName) - 2] = 0;
      titlePtr = dirName;
    }
    
  } else {
    titlePtr = strNoTitle;  // shouldn't get here
  }

  //
  // Get the font attributes
  //
  if (aAppInfo[appNum].flags & kFlagSectionHeader) {
    color = aColors[kColorSection];
    fontSize  = aFontSizes[kStrFontIndexSection];
    fontIndex = aFontIndices[kStrFontIndexSection];
    
  } else {
    fontSize  = aFontSizes[kStrFontIndexMenu];
    fontIndex = aFontIndices[kStrFontIndexMenu];
    
    // Get the appropriate color for this entry
    if (aAppInfo[appNum].flags & kFlagAppValid) {
      color = aColors[kColorMenuNormal];
      if (appNum == curApp) {
	color = aColors[kColorMenuHighlight];
      }
    } else {
      color = aColors[kColorMenuError];
    }
    
    // Highlight the one currently selected
    if ((appNum == curApp) && !menuNoHighlight) {
      mmlGetYCCComponents(color, (uint8 *) &y, (uint8 *) &cr, (uint8 *) &cb);
      
      cursorFields &= 0x3F;
      menuY = (cursorFields & 0x20)
	? (0x40 + ((cursorFields & 0x1F)<<2))
	: (0xC0 - ((cursorFields & 0x1F)<<2));
      
      y = menuY;
      
      color = mmlColorFromYCC(y, cr, cb);
    }
  }
  
  mmlSetTextProperties(fc, aFont[fontIndex], fontSize, color, DC, eBlend, 0, UNUSED);

  rect.rightBot.x = SCREEN_W - 1;
  rect.rightBot.y = SCREEN_H - 1;
  if (debug) printf("%s %08X\n", titlePtr, flags);
  if (debug) printf("rectIn: %d,%d %d,%d\n", rectIn->leftTop.x, rectIn->leftTop.y, rectIn->rightBot.x, rectIn->rightBot.y);
  if (debug) printf("rect: %d,%d %d,%d\n", rect.leftTop.x, rect.leftTop.y, rect.rightBot.x, rect.rightBot.y);
  mmlGetTextBox(fc, titlePtr, 0, strlen(titlePtr)-1, &rect);
  if (debug) printf("rect: %d,%d %d,%d\n", rect.leftTop.x, rect.leftTop.y, rect.rightBot.x, rect.rightBot.y);
  
  // Make sure this line will fit in the provided area
  if ((rect.rightBot.y > rectIn->rightBot.y) || (rect.rightBot.x > rectIn->rightBot.x)) {
#ifdef TRACE
    //    printf("%s() end (early 1)\n", __func__);
#endif
    return FALSE;
  }

  // Align the text appropriately
  drawRect = rect;  // assume left/top justification for now
  if (flags & kTextFlagCenter) {
    if (menuOrientation == kMenuVertical) {
      drawRect.leftTop.y = (rectIn->rightBot.y + rectIn->leftTop.y) >> 1;
      drawRect.leftTop.y -= (rect.rightBot.y - rect.leftTop.y + 1) >> 1;
      drawRect.rightBot.y = drawRect.leftTop.y + (rect.rightBot.y - rect.leftTop.y);
    } else {
      drawRect.leftTop.x = (rectIn->rightBot.x + rectIn->leftTop.x) >> 1;
      drawRect.leftTop.x -= (rect.rightBot.x - rect.leftTop.x + 1) >> 1;
      drawRect.rightBot.x = drawRect.leftTop.x + (rect.rightBot.x - rect.leftTop.x);
    }

  } else if (flags & kTextFlagRight) {
    if (menuOrientation == kMenuVertical) {
      drawRect.leftTop.y  = rectIn->rightBot.y - (rect.rightBot.y - rect.leftTop.y);
      drawRect.rightBot.y = rectIn->rightBot.y;
    } else {
      drawRect.leftTop.x  = rectIn->rightBot.x - (rect.rightBot.x - rect.leftTop.x);
      drawRect.rightBot.x = rectIn->rightBot.x;
    }
  }
  if (debug) printf("drawRect: %d,%d %d,%d\n", drawRect.leftTop.x, drawRect.leftTop.y, drawRect.rightBot.x, drawRect.rightBot.y);
  
  // Make sure this line will fit in the provided area (post-alignment)
  if ((drawRect.leftTop.x  < rectIn->leftTop.x ) || (drawRect.leftTop.y  < rectIn->leftTop.y ) ||
      (drawRect.rightBot.x > rectIn->rightBot.x) || (drawRect.rightBot.y > rectIn->rightBot.y)) {
#ifdef TRACE
    //    printf("%s() end (early 2)\n", __func__);
#endif
    return FALSE;
  }

  // Return the actual rectangle used
  *rectIn = drawRect;

  // Redraw this line if either the full menu should be redrawn or the highlight should be 
  // redrawn and this is the highlighted entry
  if ((checkRedraw(kRedrawMenu) || (checkRedraw(kRedrawHighlight) && (appNum == curApp)))) {
    if (!checkRedraw(kRedrawMenu)) {
      drawBackground(&drawRect, FALSE);
    }
      
    // Indent menu entries when not in horizontal mode.  Do not indent section headers.
    if (!(aAppInfo[appNum].flags & kFlagSectionHeader) && (menuOrientation != kMenuHorizontal)) {
      drawRect.leftTop.x  += aSpacing[kSpaceMenuIndent];
      drawRect.rightBot.x += aSpacing[kSpaceMenuIndent];
    }
      
    // Draw the text
    mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], titlePtr, strlen(titlePtr), &drawRect);

  } else {
    mmlGetTextBox(fc, titlePtr, 0, strlen(titlePtr)-1, &drawRect);
  }
    
  if (dirName) {
    free(dirName);
    dirName = NULL;
  }

#ifdef TRACE
  //  printf("%s() end\n", __func__);
#endif
  return TRUE;
}

//==================================================================================================
// drawMenu()
//
// Draws the menu portion of the GUI.  Each menu entry corresponds to an app.  The name displayed
// is taken from the app.txt file associated with the app.  If not available, the directory name,
// taken from applist.txt and without the slashes, will be used instead.
//
// The currently selected menu entry will be highlighted by cycling its Y component.  This makes
// it pulse.  If an app does not have an executable it is shown in a different color.
//
// Inputs:
//   areaIn - pointer to a rectangle defining the region in which to draw
// Outputs: none
// Globals modified:
//   menuY - the Y component of the highlight color
//
//==================================================================================================

void drawMenu(sRect *areaIn)
{
  static int vidSync, prevVidSync = -1, prevFrameCount = 0, frameCount = 0;
  int thisApp;
  sRect *area;
  m2dRect rectPrev, rectNext, rectArrow;
  mmlColor color;
  int appsDrawn;

  // This code is used to blink the previous/next arrows in the menu
  vidSync = _VidSync(-1);
  if (prevVidSync == -1) {
    prevVidSync = vidSync;
  }
  frameCount += vidSync - prevVidSync;
  prevVidSync = vidSync;
  
  if ((prevFrameCount & 0x10) != (frameCount & 0x10)) {
    frameCount &= 0x10;  // round down so this color lasts at least 16 frames
  }

  prevFrameCount = frameCount;
  
  // Only redraw if necessary (removed so that the arrows don't flicker)
  //  if (!checkRedraw(kRedrawMenu) && !checkRedraw(kRedrawHighlight)) return;

  // Don't redraw the highlight if we'll just redraw the whole menu next time
  if (checkRedraw(kRedrawHighlight) && (redraw & kRedrawMenu)) {
    return;
  }

  // If the menu isn't displayed then don't bother drawing
  if ((state != kStateConsoleAndMenu) && (state != kStateMenuAndStuff)) {
    return;
  }

  if (!aAppInfo) {
    return;
  }

#ifdef TRACE
  printf("%s() begin\n", __func__);
#endif

  area = areaIn;
  if (!areaIn) {
    area = &rectLeft;
  }

  if (absolutePositioning) {
    area = &aPos[kPosMenu];
  }

  // Set up initial clipping regions
  rectNext.leftTop.x = area->x;
  rectNext.leftTop.y = area->y;
  rectNext.rightBot.x = area->x + area->w - 1;
  rectNext.rightBot.y = area->y + area->h - 1;

  // Redraw the whole background if we're drawing the entire menu
  if (checkRedraw(kRedrawMenu)) {
    drawBackground(&rectNext, FALSE);
  }

  if ((aTextHAlign[kTextHAlignMenu] == kHAlignCenter) && (menuOrientation == kMenuHorizontal)) {
    // Menu is centered - draw the active menu entry first since it 
    // determines where everything else is placed

    if (drawMenuLine(curApp, &rectNext, kTextFlagCenter)) {
      if (menuOrientation == kMenuVertical) {
	rectPrev.rightBot.x = area->x + area->w - 1;
	rectPrev.rightBot.y = rectNext.leftTop.y - 1 - aSpacing[kSpaceBelowMenuEntry];
	rectNext.leftTop.y = rectNext.rightBot.y + 1 + aSpacing[kSpaceBelowMenuEntry];
	rectNext.rightBot.x = area->x + area->w - 1;
	rectNext.rightBot.y = area->y + area->h - 1;
      } else {
	rectPrev.rightBot.x = rectNext.leftTop.x - 1 - aSpacing[kSpaceBelowMenuEntry];
	rectPrev.rightBot.y = area->y + area->h - 1;
	rectNext.leftTop.x = rectNext.rightBot.x + 1 + aSpacing[kSpaceBelowMenuEntry];
	rectNext.rightBot.x = area->x + area->w - 1;
	rectNext.rightBot.y = area->y + area->h - 1;
      }

      // Draw entries to prior to the current app
      for (thisApp = curApp - 1, appsDrawn = 0; 1; thisApp--, appsDrawn++) {

	// Check for wrap around of menu entries
	if (thisApp < 0) {
	  if (menuWrap) {
	    thisApp += numApps;
	  } else {
	    break;
	  }
	}

	// Skip once we've wrapped all the way around
	if (thisApp == curApp) {
	  break;
	}
    
	// Quit if we hit the limit on number of menu entries
	if ((menuMaxEntries > 0) && (appsDrawn >= ((menuMaxEntries - 1) >> 1))) {
	  break;
	}

	// Only draw section headers in vertical mode (for now)
	if ((aAppInfo[thisApp].flags & kFlagSectionHeader) && (menuOrientation == kMenuHorizontal)){
	  continue;
	}

	// Reset the top left corner
	rectPrev.leftTop.x = area->x;
	rectPrev.leftTop.y = area->y;

	// Draw the line, quitting if it doesn't fit in the area
	if (!drawMenuLine(thisApp, &rectPrev, kTextFlagRight)) {
	  break;
	}

	// Update the area for the next line
	if (menuOrientation == kMenuVertical) {
	  rectPrev.rightBot.x = area->x + area->w - 1;
	  rectPrev.rightBot.y = rectPrev.leftTop.y - 1 - aSpacing[kSpaceBelowMenuEntry];
	} else {
	  rectPrev.rightBot.x = rectPrev.leftTop.x - 1 - aSpacing[kSpaceBelowMenuEntry];
	  rectPrev.rightBot.y = area->y + area->h - 1;
	}
      }

      // Draw entries to following the current app
      for (thisApp = curApp + 1, appsDrawn = 0; 1; thisApp++, appsDrawn++) {

	// Check for wrap around of menu entries
	if (thisApp >= numApps) {
	  if (menuWrap) {
	    thisApp -= numApps;
	  } else {
	    break;
	  }
	}
    
	// Skip once we've wrapped all the way around
	if (thisApp == curApp) {
	  break;
	}

	// Quit if we hit the limit on number of menu entries
	if ((menuMaxEntries > 0) && (appsDrawn >= (menuMaxEntries >> 1))) {
	  break;
	}

	// Reset the bottom right corner
	rectNext.rightBot.x = area->x + area->w - 1;
	rectNext.rightBot.y = area->y + area->h - 1;

	// Only draw section headers in vertical mode (for now)
	if ((aAppInfo[thisApp].flags & kFlagSectionHeader) && (menuOrientation == kMenuHorizontal)){
	  continue;
	}

	// Draw the line, quitting if it doesn't fit in the area
	if (!drawMenuLine(thisApp, &rectNext, kTextFlagLeft)) {
	  break;
	}
	
	// Update the area for the next line
	if (menuOrientation == kMenuVertical) {
	  rectNext.leftTop.y = rectNext.rightBot.y + 1 + aSpacing[kSpaceBelowMenuEntry];
	} else {
	  rectNext.leftTop.x = rectNext.rightBot.x + 1 + aSpacing[kSpaceBelowMenuEntry];
	}

	botApp = thisApp;
      }
    }

    // Draw the "previous" arrow
    color = (frameCount & 0x10) ? aColors[kColorArrow1] : aColors[kColorArrow2];
    mmlSetTextProperties(fc, aFont[aFontIndices[kStrFontIndexArrow]],
			 aFontSizes[kStrFontIndexArrow], color, DC, eBlend, 0, UNUSED);
    rectArrow.leftTop.x  = 0;
    rectArrow.leftTop.y  = 0;
    rectArrow.rightBot.x = SCREEN_W;
    rectArrow.rightBot.y = SCREEN_H;
    mmlGetTextBox(fc, aStrings[kStrArrowPrev], 0, 0, &rectArrow);
    
    if (aPos[kPosArrowPrev].x == -1) {
      if (menuOrientation == kMenuVertical) {
	rectArrow.leftTop.x  = rectPrev.leftTop.x;
	rectArrow.leftTop.y  = rectPrev.rightBot.y - rectArrow.rightBot.y;
      } else {
	rectArrow.leftTop.x  = rectPrev.rightBot.x - rectArrow.rightBot.x;
	rectArrow.leftTop.y  = area->y;
      }
    } else {
      rectArrow.leftTop.x = aPos[kPosArrowPrev].x;
      rectArrow.leftTop.y = aPos[kPosArrowPrev].y;
    }
    rectArrow.rightBot.x += rectArrow.leftTop.x;
    rectArrow.rightBot.y += rectArrow.leftTop.y;
    
    if ((curApp > 0) || menuWrap) {
      drawBackground(&rectArrow, FALSE);
      mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], aStrings[kStrArrowPrev], 1, &rectArrow);
    } else {
      if (checkRedraw(kRedrawMenu)) {
	drawBackground(&rectArrow, FALSE);
      }
    }

  } else {

    // Menu is not centered.  Start drawing from the top app.  This infinite loop will break out
    // when either the end of the app list is reached or, if wrapping is enabled, the area is
    // filled.

    // Draw the "previous" arrow
    color = (frameCount & 0x10) ? aColors[kColorArrow1] : aColors[kColorArrow2];
    mmlSetTextProperties(fc, aFont[aFontIndices[kStrFontIndexArrow]],
			 aFontSizes[kStrFontIndexArrow], color, DC, eBlend, 0, UNUSED);
    rectArrow.leftTop.x  = 0;
    rectArrow.leftTop.y  = 0;
    rectArrow.rightBot.x = SCREEN_W;
    rectArrow.rightBot.y = SCREEN_H;
    mmlGetTextBox(fc, aStrings[kStrArrowPrev], 0, 0, &rectArrow);
    
    if (aPos[kPosArrowPrev].x == -1) {
      if (menuOrientation == kMenuVertical) {
	// Center the arrow on vertical menus
	rectArrow.leftTop.x = area->x + (area->w>>1) - (rectArrow.rightBot.x>>1);
	rectArrow.leftTop.y = area->y;
      } else {
	rectArrow.leftTop.x = area->x;
	rectArrow.leftTop.y = area->y;
      }
    } else {
      rectArrow.leftTop.x = aPos[kPosArrowPrev].x;
      rectArrow.leftTop.y = aPos[kPosArrowPrev].y;
    }
    rectArrow.rightBot.x += rectArrow.leftTop.x;
    rectArrow.rightBot.y += rectArrow.leftTop.y;
    
    if ((topApp > 0) || menuWrap) {
      drawBackground(&rectArrow, FALSE);
      mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], aStrings[kStrArrowPrev], 1, &rectArrow);
    } else {
      if (checkRedraw(kRedrawMenu)) {
	drawBackground(&rectArrow, FALSE);
      }
    }

    // Update the drawing region
    if (menuOrientation == kMenuVertical) {
      rectNext.leftTop.y = rectArrow.rightBot.y + 1 + aSpacing[kSpaceBelowMenuEntry];
    } else {
      rectNext.leftTop.x = rectArrow.rightBot.x + 1 + aSpacing[kSpaceBelowMenuEntry];
    }

    // Draw the menu entries
    for (thisApp = topApp, appsDrawn = 0; 1; thisApp++, appsDrawn++) {

      // Check for wrap around of menu entries
      if (thisApp >= numApps) {
	if (menuWrap) {
	  thisApp -= numApps;
	} else {
	  break;
	}
      }
    
      // Skip once we've wrapped all the way around
      if ((thisApp == topApp) && (appsDrawn > 0)) {
	break;
      }

      // Quit if we hit the limit on number of menu entries
      if ((menuMaxEntries > 0) && (appsDrawn >= menuMaxEntries)) {
	break;
      }

      // Reset the bottom right corner
      rectNext.rightBot.x = area->x + area->w - 1;
      rectNext.rightBot.y = area->y + area->h - 1;

      // Only draw section headers in vertical mode (for now)
      if ((aAppInfo[thisApp].flags & kFlagSectionHeader) && (menuOrientation == kMenuHorizontal)) {
	continue;
      }

      // Draw the line, quitting if it doesn't fit in the area
      if (!drawMenuLine(thisApp, &rectNext, kTextFlagLeft)) {
	break;
      }

      if (menuOrientation == kMenuVertical) {
	rectNext.leftTop.y = rectNext.rightBot.y + 1 + aSpacing[kSpaceBelowMenuEntry];
	rectNext.rightBot.y = area->y + area->h - 1;
      } else {
	rectNext.leftTop.x = rectNext.rightBot.x + 1 + aSpacing[kSpaceBelowMenuEntry];
	rectNext.rightBot.x = area->x + area->w - 1;
      }

      botApp = thisApp;
    }
  }
  
  // Draw the "next" arrow
  color = (frameCount & 0x10) ? aColors[kColorArrow1] : aColors[kColorArrow2];
  mmlSetTextProperties(fc, aFont[aFontIndices[kStrFontIndexArrow]],
		       aFontSizes[kStrFontIndexArrow], color, DC, eBlend, 0, UNUSED);
  rectArrow.leftTop.x  = 0;
  rectArrow.leftTop.y  = 0;
  rectArrow.rightBot.x = SCREEN_W;
  rectArrow.rightBot.y = SCREEN_H;
  mmlGetTextBox(fc, aStrings[kStrArrowNext], 0, 0, &rectArrow);

  if (aPos[kPosArrowNext].x == -1) {
    if (aTextHAlign[kTextHAlignMenu] == kHAlignCenter) {
      rectArrow.leftTop.x  += rectNext.leftTop.x;
      rectArrow.leftTop.y  += rectNext.leftTop.y;
    } else {
      if (menuOrientation == kMenuVertical) {
	rectArrow.leftTop.x = area->x + (area->w>>1) - (rectArrow.rightBot.x>>1);
	rectArrow.leftTop.y = rectNext.leftTop.y;
      } else {
	rectArrow.leftTop.x = rectNext.leftTop.x;
	rectArrow.leftTop.y = area->y;
      }
    }
  } else {
    rectArrow.leftTop.x = aPos[kPosArrowNext].x;
    rectArrow.leftTop.y = aPos[kPosArrowNext].y;
  }
  rectArrow.rightBot.x += rectArrow.leftTop.x;
  rectArrow.rightBot.y += rectArrow.leftTop.y;
  
  if ((botApp < (numApps - 1)) ||
      ((aTextHAlign[kTextHAlignMenu] == kHAlignCenter) && (botApp != curApp)) ||
      menuWrap) {
    drawBackground(&rectArrow, FALSE);
    mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], aStrings[kStrArrowNext], 1, &rectArrow);
  } else {
    if (checkRedraw(kRedrawMenu)) {
      drawBackground(&rectArrow, FALSE);
    }
  }

#ifdef TRACE
  printf("%s() end\n", __func__);
#endif
}

//==================================================================================================
// drawImages()
//
// Draws the app images, but only when using absolute positioning.  This is done here rather than
// in drawAppStuff() because the images must be drawn before *all* text - including the title and
// menu - just in case the images serve as a background.
//
// Inputs:
//   rectIn - clipping rectangle of the area to draw (NULL for full screen)
// Outputs: none
// Globals modified: none
//
//==================================================================================================

void drawImages(m2dRect *rectIn)
{
  int sr;
  m2dRect rect, clearRect;
  m2dPoint point;
  sApp *app;
  sAppImgGroup *imgGroup;
  sAppImg *img;
  int curImgGroup;
  int draw, drawFull;
  sRect redrawRect;
  long sec, usec;
  int debug = 0;
  int debugTime = 0;

  // Don't do anything if not using absolute positioning
  if (!absolutePositioning) {
    return;
  }

  // If the images aren't displayed then don't bother drawing
  if ((state != kStateConsoleAndStuff) && (state != kStateMenuAndStuff)) {
    return;
  }

  // If no apps, then return
  if (!aAppInfo) {
    return;
  }
  if (curApp >= numApps) {
    return;
  }
  if (!aAppInfo[curApp].flags & kFlagAppValid) {
    return;
  }

#ifdef TRACE
  if (rectIn) {
    printf("%s() begin (%d, {%d,%d,%d,%d})\n", __func__, curApp, rectIn->leftTop.x, rectIn->leftTop.y,
	   rectIn->rightBot.x, rectIn->rightBot.y);
  } else {
    printf("%s() begin (%d, NULL)\n", __func__, curApp);
  }
#endif

  app = &aAppInfo[curApp];

  // Cycle through all the image groups
  for (curImgGroup = 0; curImgGroup < app->numImgGroups; curImgGroup++) {

    imgGroup = &app->aImgGroups[curImgGroup];

    // Check timer to see if we need to rotate this image
    if (imgGroup->time == -1) {
      _TimeElapsed(&sec, &usec);
      imgGroup->time = (sec * 1000000) + usec;
      if (debugTime) {
	printf("%s() - Initializing time for image group %d: %d\n",
	       __func__, curImgGroup, imgGroup->time);
      }
    } else if ((aRotateImg[kRotateImg1 + curImgGroup] > 0) && (imgGroup->numImg > 1)) {
      _TimeElapsed(&sec, &usec);
      if ((((sec * 1000000) + usec) - imgGroup->time) >
	  (aRotateImg[kRotateImg1 + curImgGroup] * 1000000)) {

	imgGroup->time = (sec * 1000000) + usec;
	imgGroup->curImg++;
	if (debugTime) {
	  printf("%s() - Image %d rotate (%ld > %d)\n", __func__,
		 curImgGroup, (sec * 1000000) + usec, imgGroup->time);
	}
	// Check if we wrapped
	if (imgGroup->curImg >= imgGroup->numImg) {
	  if (debugTime) {
	    printf("%s() - Image %d wrapped (%d images)\n", __func__,
		   curImgGroup, imgGroup->numImg);
	  }
	  imgGroup->curImg = 0;
	}
	// Force a redraw if the image is already ready
	if (imgGroup->aAppImg[imgGroup->curImg].status == kImgStatusDrawn) {
	  imgGroup->draw = 2;
	}
	if (debugTime) {
	  printf("%s() - New image %d\n", __func__, imgGroup->curImg);
	}
      }
    }

    drawFull = FALSE;
    if ((imgGroup->draw > 0) && !rectIn) {
      drawFull = TRUE;

      // Mask interrupts so that the TGA handler doesn't return right as we're updating things
      DISABLE_INTERRUPTS(sr);
      imgGroup->draw--;
      RESTORE_INTERRUPTS(sr);

      clearRect.leftTop.x = aPos[kPosImg1 + curImgGroup].x;
      clearRect.leftTop.y = aPos[kPosImg1 + curImgGroup].y;
      clearRect.rightBot.x = clearRect.leftTop.x + aPos[kPosImg1 + curImgGroup].w - 1;
      clearRect.rightBot.y = clearRect.leftTop.y + aPos[kPosImg1 + curImgGroup].h - 1;
      drawBackground(&clearRect, TRUE);  // Clear background, but don't draw images again
#ifdef TRACE
      printf("%s() - Erasing img %d (%d,%d,%d,%d)\n", __func__, curImgGroup, clearRect.leftTop.x,
	     clearRect.leftTop.y, clearRect.rightBot.x, clearRect.rightBot.y);
#endif
    }
      
    if ((imgGroup->numImg > 0) && (imgGroup->aAppImg[imgGroup->curImg].status == kImgStatusDrawn)) {

      img = &imgGroup->aAppImg[imgGroup->curImg];

      point.x = aPos[kPosImg1 + curImgGroup].x;
      point.y = aPos[kPosImg1 + curImgGroup].y;

      rect.leftTop.x = 0;
      rect.leftTop.y = 0;
      rect.rightBot.x = img->w - 1;
      rect.rightBot.y = img->h - 1;

      if ((point.x + img->w) > SCREEN_W) {
	rect.rightBot.x -= (point.x + img->w) - SCREEN_W;
      }
      if ((point.y + img->h) > SCREEN_H) {
	rect.rightBot.y -= (point.y + img->h) - SCREEN_H;
      }
      // Clamp to the size of the region
      if (aPos[kPosImg1 + curImgGroup].w <= rect.rightBot.x) {
	rect.rightBot.x = aPos[kPosImg1 + curImgGroup].w - 1;
      }
      if (aPos[kPosImg1 + curImgGroup].h <= rect.rightBot.y) {
	rect.rightBot.y = aPos[kPosImg1 + curImgGroup].h - 1;
      }

      // See if we only draw part of an image, but only if we're not drawing the full thing
      draw = (img->status == kImgStatusDrawn) ? TRUE : FALSE;
      if (rectIn) {
	// Check for clipping on left/right side
	if (rectIn->leftTop.x > point.x) {
	  rect.leftTop.x += rectIn->leftTop.x - point.x;
	  point.x += rectIn->leftTop.x - point.x;
	  if (rect.leftTop.x >= rect.rightBot.x) {
	    draw = FALSE;
	  }
	} else {
	  if (rectIn->rightBot.x < point.x) {
	    draw = FALSE;
	  } else {
	    if (rectIn->rightBot.x < (point.x + rect.rightBot.x)) {
	      rect.rightBot.x -= (point.x + rect.rightBot.x) - rectIn->rightBot.x;
	    }
	  }	      
	}

	// Check for clipping on top/bottom
	if (rectIn->leftTop.y > point.y) {
	  rect.leftTop.y += rectIn->leftTop.y - point.y;
	  point.y += rectIn->leftTop.y - point.y;
	  if (rect.leftTop.y >= rect.rightBot.y) {
	    draw = FALSE;
	  }
	} else {
	  if (rectIn->rightBot.y < point.y) {
	    draw = FALSE;
	  } else {
	    if (rectIn->rightBot.y < (point.y + rect.rightBot.y)) {
	      rect.rightBot.y -= (point.y + rect.rightBot.y) - rectIn->rightBot.y;
	    }
	  }	      
	}
      } else {
	if (!drawFull) {
	  draw = FALSE;
	}
      }

      if (draw) {
	m2dCopyRectDis(&gc, img->pixmap, &frameBuffer[drawBuffer], &rect, point);

#ifdef TRACE
	printf("%s() - Drawing img %d (%d,%d,%d,%d %d,%d)\n", __func__, curImgGroup,
	       rect.leftTop.x, rect.leftTop.y, rect.rightBot.x, rect.rightBot.y,
	       point.x, point.y);
#endif
	
	if (drawFull) {
	  // Check if this image overlapped any text.  If so, set flag to redraw it.
	  redrawRect.x = point.x;
	  redrawRect.y = point.y;
	  redrawRect.w = rect.rightBot.x - rect.leftTop.x + 1;
	  redrawRect.h = rect.rightBot.y - rect.leftTop.y + 1;
	  if (checkOverlap(&redrawRect, &aPos[kPosTitle]) ||
	      checkOverlap(&redrawRect, &aPos[kPosInstr])) {
	    redraw |= kRedrawTitle<<1;  // force redraw this frame
	    if (debug) {
	      printf("  Overlap title\n");
	    }
	  }
	  if (checkOverlap(&redrawRect, &aPos[kPosMenu])) {
	    redraw |= kRedrawMenu<<1;;  // force redraw this frame
	    if (debug) {
	      printf("  Overlap menu\n");
	    }
	  }
	  if (checkOverlap(&redrawRect, &aPos[kPosDesc])) {
	    redraw |= kRedrawAppStuff<<1;;  // force redraw this frame
	    if (debug) {
	      printf("  Overlap description (%ld,%ld,%ld,%ld) (%ld,%ld,%ld,%ld)\n",
		     redrawRect.x, redrawRect.y, redrawRect.w, redrawRect.h,
		     aPos[kPosDesc].x, aPos[kPosDesc].y, aPos[kPosDesc].w, aPos[kPosDesc].h);
	    }
	  }
	  redraw |= kRedrawConsole<<1;
	}
      }
    }
  }



#ifdef TRACE
  printf("%s() end\n", __func__);
#endif
}

//==================================================================================================
// checkOverlap()
//
// Returns TRUE or FALSE depending on whether the two input rectangles overlap.
//
// Inputs:
//   rect1 & rect2 - the rectangles to check for overlap
// Outputs: TRUE or FALSE
// Globals modified: none
//
//==================================================================================================

int checkOverlap(sRect *rect1, sRect *rect2)
{
  if (rect1->x >= (rect2->x + rect2->w)) {
    return FALSE;
  }
  if (rect1->y >= (rect2->y + rect2->h)) {
    return FALSE;
  }
  if (rect2->x >= (rect1->x + rect1->w)) {
    return FALSE;
  }
  if (rect2->y >= (rect1->y + rect1->h)) {
    return FALSE;
  }

  return TRUE;
}

//==================================================================================================
// drawAppStuff()
//
// Draws the images and description text associated with the currently selected app.
//
// Inputs:
//   areaIn - pointer to a rectangle defining the region in which to draw
//   appNum - the index of the currently selected app
// Outputs: none
// Globals modified: none
//
//==================================================================================================

void drawAppStuff(sRect *areaIn, int appNum)
{
  m2dRect rect;
  m2dPoint point;
  sRect *area, rectRend;
  sApp *app;
  sAppImgGroup *imgGroup;
  sAppImg *img;

  // Quit if the app isn't valid
  if (!aAppInfo) {
    return;
  }
  if (!aAppInfo[appNum].flags & kFlagAppValid) {
    return;
  }
  app = &aAppInfo[appNum];

  // Check if new images are ready to be drawn
  if (!absolutePositioning) {
    imgGroup = &app->aImgGroups[0];
    if (imgGroup->draw == 2) {
      imgGroup->draw = 0;
      redraw |= kRedrawAppStuff << 1;
    }
    imgGroup = &app->aImgGroups[1];
    if (imgGroup->draw == 2) {
      imgGroup->draw = 0;
      redraw |= kRedrawAppStuff << 1;
    }
  }

  // Only redraw if necessary
  if (!checkRedraw(kRedrawAppStuff)) {
    return;
  }
  
  // If the stuff isn't displayed then don't bother drawing
  if ((state != kStateConsoleAndStuff) && (state != kStateMenuAndStuff)) {
    return;
  }

  // Only redraw if everything is ready to be drawn
  //  if (!imagesReady(appNum)) {
  //    return;
  //  }

  // When using absolute positioning, images are drawn in drawImages() rather than here.
  // So just draw the description and be done with it.
  if (absolutePositioning) {
    drawText(app->txtDesc, &aPos[kPosDesc], kStrFontIndexDesc, aColors[kColorDesc], topDesc,
	     aTextHAlign[kTextHAlignDesc]);
    return;
  }

  rect.leftTop.x = areaIn->x;
  rect.leftTop.y = areaIn->y;
  rect.rightBot.x = areaIn->x + areaIn->w - 1;
  rect.rightBot.y = areaIn->y + areaIn->h - 1;
  drawBackground(&rect, FALSE);

#ifdef TRACE
  printf("%s() begin\n", __func__);
#endif
  
  area = areaIn ? areaIn : &rectRight;

  coordImg1.x = area->x;
  coordImg1.y = area->y;

  //
  // Image 1
  //
  
  // Set up image 2 starting coordinates.  Assume image 1 doesn't exist.
  coordImg2.x = coordImg1.x;
  coordImg2.y = coordImg1.y;

  // Get a pointer to the current image in group 0
  img = (app->numImgGroups >= 1) ? &app->aImgGroups[0].aAppImg[app->aImgGroups[0].curImg] : NULL;
  if (areaIn && img && img->pixmap) {
    rect.leftTop.x = 0;
    rect.leftTop.y = 0;
    rect.rightBot.x = img->w - 1;
    rect.rightBot.y = img->h - 1;
    if (img->w > area->w) {
      rect.rightBot.x = area->w - 1;
    }
    if (img->h > (area->h - (coordImg1.y - area->y))) {
      rect.rightBot.y = area->h - (coordImg1.y - area->y) - 1;
    }
    
    point.x = coordImg1.x + ((area->w - (rect.rightBot.x - rect.leftTop.x + 1)) >> 1);
    point.y = coordImg1.y;
    
    m2dCopyRectDis(&gc, img->pixmap, &frameBuffer[drawBuffer], &rect, point);

    // Image 1 did exist so push down image 2 by its height
    coordImg2.y += (rect.rightBot.y - rect.leftTop.y) + 1 + aSpacing[kSpaceBelowImg1];
  }
  
  //
  // Image 2
  //
  
  // Set up description starting coordinates2.  Assume image 2 doesn't exist.
  coordDesc.x = coordImg2.x;
  coordDesc.y = coordImg2.y;
  
  // Get a pointer to the current image in group 1
  img = (app->numImgGroups >= 2) ? &app->aImgGroups[1].aAppImg[app->aImgGroups[1].curImg] : NULL;
  if (areaIn && img && img->pixmap) {
    rect.leftTop.x = 0;
    rect.leftTop.y = 0;
    rect.rightBot.x = img->w - 1;
    rect.rightBot.y = img->h - 1;
    if (img->w > area->w) {
      rect.rightBot.x = area->w - 1;
    }
    if (img->h > (area->h - (coordImg2.y - area->y))) {
      rect.rightBot.y = area->h - (coordImg2.y - area->y) - 1;
    }
    
    point.x = coordImg2.x + ((area->w - (rect.rightBot.x - rect.leftTop.x + 1)) >> 1);
    point.y = coordImg2.y;
    
    // Only draw if there's room left
    if (point.y < (area->y + area->h)) {
      m2dCopyRectDis(&gc, img->pixmap, &frameBuffer[drawBuffer], &rect, point);
    }
    
    // Image 2 did exist so push down the description by its height
    coordDesc.y += (rect.rightBot.y - rect.leftTop.y) + 1 + aSpacing[kSpaceBelowImg2];
  }
  
  //
  // Description
  //

  rectRend.x = area->x;
  rectRend.y = coordDesc.y;
  rectRend.w = area->w;
  rectRend.h = area->h - (coordDesc.y - area->y);
 
  // Now actually draw the text (if there's room)
  if (areaIn && (coordDesc.y < (area->y + area->h))) {
    drawText(app->txtDesc, &rectRend, kStrFontIndexDesc, aColors[kColorDesc], topDesc,
	     aTextHAlign[kTextHAlignDesc]);
  }

#ifdef TRACE
  printf("%s() end\n", __func__);
#endif
}

//==================================================================================================
// parseText()
//
// Parses a string to be drawn such that the line breaks are in the correct position given the
// size of the rendering area and size of the font.
//
// Inputs:
//   str  - string to render
//   rect - rectangle in which to render
//   fontIndex - index into font arrays
// Outputs: none
// Globals modified: none
//
//==================================================================================================

void parseText(char *str, sRect *rectIn, int fontIndex)
{
  int charIndex;
  m2dRect rect;
  char *ptr, *dest;
  int done;
  int autoW = FALSE;
  int autoH = FALSE;
  int maxW = -1;
  int foundSpace;
  int debug = 0;

  // Parse the string by inserting line breaks accordingly.  This changes depending on the
  // font size and the width of the area in which to draw.
    
#ifdef TRACE
  printf("%s(\"%s\", {%ld,%ld,%ld,%ld}, %d) begin\n", __func__, str, rectIn->x, rectIn->y, rectIn->w, rectIn->h, fontIndex);
#endif
  
  // Skip initial formatting
  while ((str[0] == '%') && (str[1] != '%')) {
    str++;
    while (str[0] && (str[0] != '%')) {
      str++;
    }
    if (str[0] == '%') {
      str++;
    }
  }

  // The user can pass in -1 for width or height and the program will use the minimum box that
  // bounds the text.
  if (rectIn->w == -1) {
    if (debug) printf("%s() - Auto computing width for string '%s'\n", __func__, str);
    autoW = TRUE;
    rectIn->w = SCREEN_W - rectIn->x;
  }
  if (rectIn->h == -1) {
    if (debug) printf("%s() - Auto computing height for string '%s'\n", __func__, str);
    autoH = TRUE;
    rectIn->h = 0;
  }

  // Add the new lines where appropriate given the font size
  mmlSetTextProperties(fc, aFont[aFontIndices[fontIndex]], aFontSizes[fontIndex],
		       aColors[kColorDesc], DC, eBlend, 0, UNUSED); // color is a don't care
    
  if (debug) printf("Parsing string \"%s\"\n", str);

  // Go through and remove all new lines, replacing them with a space as necessary
  ptr = dest = str;
  foundSpace = -1;  // haven't found a space or a letter yet
  while (ptr && *ptr) {
    *dest = *ptr;  // copy the character speculatively

    if ((*ptr == 0x0D) || (*ptr == 0x0A)) {
      // Got a new-line.  If there wasn't a space *before* the new-line then we need to check for
      // one afterwards.
      if (foundSpace == 0) {

	// Only keep the last of the two new-line characters CR & LF
	if ((ptr[1] != 0x0D) && (ptr[1] != 0x0A)) {
	  
	  // Only keep it if there's no space *after* the new-line
	  if ((ptr[1] != ' ') || (ptr[1] != '\t')) {

	    // Okay, there was no space before or after the new-line.  Add one.
	    *dest = ' ';
	    dest++;
	    foundSpace = 1;
	  }
	}
      }

    } else if ((*ptr == ' ') || (*ptr == '\t')) {
      if (foundSpace == 0) {
	foundSpace = 1;
      }
      dest++;  // character is okay - keep it

    } else if (*ptr > ' ') {
      foundSpace = 0;
      dest++;  // character is okay - keep it

    } else {
      // remove any non-printable characters that aren't handled above
    }

    ptr++;
  }
  *dest = 0;

  if (debug) printf("Removed new lines \"%s\"\n", str);

  // Traverse the whole string
  ptr = str;
  while (ptr && *ptr) {
      
    // Look for word breaks until the string up to that point cannot fit
    charIndex = 0;
    do {
      // If there's a new line then skip it and start over.
      while ((ptr[charIndex] == '\\') && (ptr[charIndex+1] == 'n')) {
	if (debug) printf("--Skipping new line\n");
	ptr[charIndex] = ' ';
	ptr[charIndex+1] = '\n';
	ptr += charIndex + 2;
	charIndex = 0;
      }
      
      done = FALSE;
      do {
	charIndex++;
	if (debug) printf("--%c\n", ptr[charIndex]);

	// Look for new line
	if ((ptr[charIndex] == '\\') && (ptr[charIndex+1] != '\\')) {
	  done = TRUE;
	  if (debug) printf("--Found code\n");
	}
	// Look for space
	if (ptr[charIndex] == ' ') {
	  done = TRUE;
	  if (debug) printf("--Found space\n");
	}
	// Look for tab
	if (ptr[charIndex] == '\t') {
	  done = TRUE;
	  if (debug) printf("--Found tab\n");
	}
	// Look for end of string
	if (ptr[charIndex] == 0) {
	  done = TRUE;
	  if (debug) printf("--Found end\n");
	}
      } while (!done);

      rect.leftTop.x = 0;
      rect.leftTop.y = 0;
      rect.rightBot.x = rectIn->w - 1;
      rect.rightBot.y = SCREEN_H - 1;
      mmlGetTextBox(fc, ptr, 0, charIndex-1, &rect);
      if (debug) printf("--Line width=%d (vs. %ld) from '%c' to '%c'\n",
			rect.rightBot.x, rectIn->w - 1, ptr[0], ptr[charIndex-1]);

    } while (ptr[charIndex] && (rect.rightBot.x < (rectIn->w - 1)));
      
    if (debug) printf("--End of line\n");

    // Go back to the previous word break
    if (rect.rightBot.x >= (rectIn->w - 1)) {
      do {
	charIndex--;
      } while ((charIndex > 0) && (ptr[charIndex] != ' '));
    }
      
    // Special case: Single word too long to fit so just display it all.
    // It will be truncated automatically.
    if (charIndex <= 0) {
      // Skip over the word to get to the space at the end
      do {
	charIndex++;
      } while (ptr[charIndex] && (ptr[charIndex] != ' '));

      // Point to last space in case there is more than one.
      do {
	charIndex++;
      } while (ptr[charIndex] && (ptr[charIndex] == ' '));
      charIndex--;
    }

    // When computing auto-width we must use the longest of the lines
    rect.leftTop.x = 0;
    rect.leftTop.y = 0;
    rect.rightBot.x = SCREEN_W - 1;
    rect.rightBot.y = SCREEN_H - 1;
    mmlGetTextBox(fc, ptr, 0, charIndex-1, &rect);
    if (autoW && (rect.rightBot.x > maxW)) {
      maxW = rect.rightBot.x + 1;
      if (debug) printf("%s() - Max line width: %d - %s\n", __func__, maxW, ptr);
    }
    if (autoH) {
      rectIn->h += rect.rightBot.y + 1;
      if (debug) printf("%s() - Max line height: %ld\n", __func__, rectIn->h);
    }
      
    // Update the pointer to the character after the break we just inserted
    if (ptr[charIndex] == 0) {
      ptr = NULL;
    } else {
      ptr[charIndex] = '\n';
      ptr += charIndex + 1;
    }
  }
  if (debug) printf("Done parsing.\n");

  // Set the auto values when appropriate
  if (autoW) {
    rectIn->w = maxW;
    if (debug) printf("%s() - Auto computed width:  %ld\n", __func__, rectIn->w);
  }
  if (autoH) {
    if (debug) printf("%s() - Auto computed height: %ld\n", __func__, rectIn->h);
  }

#ifdef TRACE
  printf("%s() end\n", __func__);
#endif
  
}


//==================================================================================================
// drawText()
//
// Renders text to the given bounding box.  Lines will wrap as appropriate to stay within the box
// when possible.
//
// Inputs:
//   str  - string to render
//   rect - rectangle in which to render
//   fontIndex - index into font arrays
//   color - color in which to draw the text
//   topLine - the line at which to start drawing (for scrolling)
//   flags - flags that modify text drawing
// Outputs: none
// Globals modified: none
//
//==================================================================================================

void drawText(char *str, sRect *rectIn, int fontIndex, unsigned long color, int topLine, int align)
{
  int charIndex;
  m2dRect rect;
  char *ptr;
  int line;
  int flags = 0;
  int debug = 0;

  if (str == NULL) {
    return;
  }

#ifdef TRACE
  printf("%s(\"%s\", {%ld,%ld,%ld,%ld}, ...) begin\n", __func__, str, rectIn->x, rectIn->y, rectIn->w, rectIn->h);
#endif
  
  // Parse the string if it has not already been parsed.
  if (str[-1] == 0) {
    parseText(str, rectIn, fontIndex);
    str[-1] = 1;
  }

  switch (align) {
  case kHAlignLeft:   flags |= kTextFlagLeft;  break;
  case kHAlignCenter: flags |= kTextFlagCenter; break;
  case kHAlignRight:  flags |= kTextFlagRight;  break;
  }

  // Grab any format flags from the beginning of the string
  while ((str[0] == '%') && (str[1] != '%')) {
    // Left justify the text
    if (strncmp(str, "%LEFT%", strlen("%LEFT%")) == 0) {
      str += strlen("%LEFT%");
      flags &= ~kTextFlagCenter;
    }
    // Center the text
    if (strncmp(str, "%CENTER%", strlen("%CENTER%")) == 0) {
      str += strlen("%CENTER%");
      flags |= kTextFlagCenter;
    }
  }

  // First clear the area
  rect.leftTop.x = rectIn->x;
  if (flags & kTextFlagCenter) {
    rect.leftTop.x -= rectIn->w >> 1;
  }
  rect.leftTop.y = rectIn->y;
  rect.rightBot.x = rect.leftTop.x + rectIn->w - 1;
  rect.rightBot.y = rect.leftTop.y + rectIn->h - 1;
  if (!(flags & kTextFlagNoDraw)) {
    drawBackground(&rect, FALSE);
  }

  mmlSetTextProperties(fc, aFont[aFontIndices[fontIndex]], aFontSizes[fontIndex],
		       color, DC, eBlend, 0, UNUSED);
    
  line = 0;
  ptr = str;
  while (ptr && *ptr) {
    // Find the end of this line
    for (charIndex = 0; ptr[charIndex] && (ptr[charIndex] != '\n'); charIndex++);
      
    // Draw just this line.
    if ((line >= topLine) || (ptr[charIndex] == 0)) {
      rect.leftTop.x = 0;
      rect.rightBot.x = SCREEN_W - 1;
      rect.rightBot.y = SCREEN_H - 1;
      if (debug) printf("%s() - rect: %d,%d %d,%d\n", __func__,
			rect.leftTop.x, rect.leftTop.y, rect.rightBot.x, rect.rightBot.y);
      mmlGetTextBox(fc, ptr, 0, charIndex-1, &rect);
      if (debug) printf("%s() - rect: %d,%d %d,%d\n", __func__,
			rect.leftTop.x, rect.leftTop.y, rect.rightBot.x, rect.rightBot.y);

      // Only draw if there's room left
      if (rect.rightBot.y < (rectIn->y + rectIn->h)) {

	// Set the starting coordinates
	rect.leftTop.x  = rectIn->x - ((flags & kTextFlagCenter) ? ((rect.rightBot.x + 1)>>1) : 0);
	rect.rightBot.x = rectIn->x + rectIn->w - 1;
	rect.rightBot.y = rectIn->y + rectIn->h - 1;

	if (rect.rightBot.x >= SCREEN_W) {
	  rect.rightBot.x = SCREEN_W - 1;
	}
	if (rect.rightBot.y >= SCREEN_H) {
	  rect.rightBot.y = SCREEN_H - 1;
	}

	if (debug) printf("%s() - rect: %d,%d %d,%d\n", __func__,
			  rect.leftTop.x, rect.leftTop.y, rect.rightBot.x, rect.rightBot.y);

	if (!(flags & kTextFlagNoDraw)) {
	  if (debug) printf("%s() - rect: %d,%d %d,%d (DRAW)\n", __func__,
			    rect.leftTop.x, rect.leftTop.y, rect.rightBot.x, rect.rightBot.y);
	  mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], ptr, charIndex, &rect);
	}
      } else {
	ptr = NULL;
      }

      // Update the location so the next line is drawn below this one
      rect.leftTop.y = rect.rightBot.y + 1;
    }
      
    line++;
      
    if (ptr && ptr[charIndex]) {
      ptr += charIndex + 1;
    } else {
      ptr = NULL;
    }
  }

#ifdef TRACE
  printf("%s() end\n", __func__);
#endif
}


//==================================================================================================
// drawIntro()
//
// Draws the intro image for the given app.
//
// Inputs:  none
// Outputs: none
// Globals modified: none
//
//==================================================================================================
void drawIntro()
{
  m2dRect rect;
  m2dPoint point;
  sApp *app;
  
  if (!aAppInfo) {
    return;
  }
  
#ifdef TRACE
  printf("%s() begin\n", __func__);
#endif

  app = &aAppInfo[curApp];

  if ((app->imgIntro.status != kImgStatusDrawn) || (app->imgIntro.pixmap == NULL)) {
    return;
  }
  
  // Copy the intro image to both frame buffers
  rect.leftTop.x = 0;
  rect.leftTop.y = 0;
  rect.rightBot.x = app->imgIntro.w - 1;
  rect.rightBot.y = app->imgIntro.h - 1;
  point.x = 0;
  point.y = 0;
  m2dCopyRectDis(&gc, app->imgIntro.pixmap, &frameBuffer[drawBuffer], &rect, point);

#ifdef TRACE
  printf("%s() end\n", __func__);
#endif
}

//==================================================================================================
// drawLoading()
//
// Draws the "Loading..." screen when the user has selected an app.
//
// Inputs:  none
// Outputs: none
// Globals modified: none
//
//==================================================================================================

void drawLoading()
{
  mmlSetTextProperties(fc,aFont[aFontIndices[kStrFontIndexLoading]],aFontSizes[kStrFontIndexLoading],
		       aColors[kColorLoading], DC, eBlend, 0, UNUSED);
  
  drawStringCentered(SCREEN_W>>1, SCREEN_H>>1, aStrings[kStrLoading]);
}

//==================================================================================================
// drawStringCentered()
//
// Draws a string centered on a point.  Assumes the font size and color have been set already.
//
// Inputs:
//   x,y - coordinates of the center of the string
//   str - pointer to the string
// Outputs:
//   m2dRect - returns the rectangle returned by mmlSimpleDrawText()
// Globals modified: none
//
//==================================================================================================

// Text properties must be set up before calling this
m2dRect drawStringCentered(int x, int y, char *str)
{
  m2dRect rect;

  // Set up clipping region to be the whole screen
  rect.leftTop.x = 0;
  rect.leftTop.y = 0;
  rect.rightBot.x = SCREEN_W;
  rect.rightBot.y = SCREEN_H;
  mmlGetTextBox(fc, str, 0, strlen(str)-1, &rect);

  // Rect now has text extent in rightBot member
  rect.leftTop.x = x - (rect.rightBot.x >> 1);
  rect.leftTop.y = y;
  rect.rightBot.x = SCREEN_W;
  rect.rightBot.y = SCREEN_H;
  mmlSimpleDrawText(fc, &frameBuffer[drawBuffer], str, strlen(str), &rect);

  return rect;
}

//==================================================================================================
// insertConsoleString()
//
// Adds a string to be displayed on the console.  If a string is too long to be displayed it will
// be truncated and have an ellipses at the end.  When an error is added to the console the
// display will automatically switch to show it.
//
// Inputs:
//   str - the string to insert
// Outputs: none
// Globals modified:
//   consoleStart - an index into the console array pointing to the most recently inserted string
//   aConsole[]   - the array of console strings
//
//==================================================================================================

void insertConsoleString(char *str)
{
  m2dRect rect;
  int lastChar;
  int truncated = FALSE;
  int i;

#ifndef DEBUG
  if (!strncmp(str, "**DBG**", strlen("**DBG**"))) { forbidAppend = TRUE; return; }
#endif

  redraw |= kRedrawConsole;
  
  forbidAppend = FALSE;
  
  consoleStart++;
  if (consoleStart >= CONSOLE_LINES_MAX) consoleStart = 0;

  // Set up the font size
  mmlSetTextProperties(fc, aFont[aFontIndices[kStrFontIndexConsole]], aFontSizes[kStrFontIndexConsole],
		       aColors[kColorConsoleNormal], DC, eBlend, 0, UNUSED);

  // Remove any trailing newlines
  for (i = 0; i < strlen(str); i++) {
    if (str[i] == '\n') str[i] = 0;
    if (str[i] == '\r') str[i] = 0;
  }
  
#ifdef CONSOLE_PRINT
  printf("CONSOLE: %s\n", str);  fflush(stdout);
#endif
  
  rect.leftTop.x  = 0;
  rect.leftTop.y  = 0;
  rect.rightBot.y = rectLeft.h - 1;

  // Find how much of the string will fit in the console display.
  // Also trucate to our max string length if necessary.
  lastChar = strlen(str) - 1;
  if (lastChar >= CONSOLE_LINE_SIZE) {
    lastChar = CONSOLE_LINE_SIZE - 1;
    truncated = TRUE;
  }

  // Remove characters until it will fit in the allotted space
  for (; lastChar > 0; lastChar--) {
    rect.leftTop.x  = 0;
    rect.rightBot.x = rectLeft.w - 1;
    mmlGetTextBox(fc, str, 0, lastChar, &rect);
    if (rect.rightBot.x < (rectLeft.w - 1)) break;
    truncated = TRUE;
  }

  // Add an ellipsis if the string was truncated
  if (truncated && (lastChar >= 2)) {
    str[lastChar+1] = 0;
    str[lastChar-0] = '.';
    str[lastChar-1] = '.';
    str[lastChar-2] = '.';
  }

  // For errors, automatically switch to console mode so the user can see them
  if (!strncmp(str, "**ERR**", strlen("**ERR**")) &&
      ((state == kStateMenuAndStuff) || (state == kStateInit))) {
    nextState = kStateConsoleAndMenu;
    redraw |= kRedrawAll;
  }
  
  // *Finally* copy the string into our array
  strcpy(aConsole[consoleStart], str);
  
  updateScreen();
}

//==================================================================================================
// appendConsoleString()
//
// Appends a string to the end of the previously inserted console string.
//
// Inputs:
//   str - the string to append on the previous console message
// Outputs: none
// Globals modified:
//   consoleStart - decrements this since insertConsoleString() will increment it
//
//==================================================================================================

void appendConsoleString(char *str)
{
  char catbuf[100];

#ifndef DEBUG
  if (forbidAppend) return;
#endif
  
  strcpy(catbuf, aConsole[consoleStart]);
  strcat(catbuf, str);
  consoleStart--;

  insertConsoleString(catbuf);
}

//==================================================================================================
// checkDataFile()
//
// Looks for and attempts to load the data file.  There are many checks along the way and if any
// fail the process will be repeated.  It will perform three attempts before giving up.
//
// Inputs:  none
// Outputs: none
// Globals modified:
//   fpDataFile  - file pointer to the data file (NULL if failed)
//   bufDataFile - pointer to the data buffer read (NULL if failed)
//
//==================================================================================================

void checkDataFile()
{
  long buf[4];
  int actualFileSize;
  int bytesRead;
  int reopenAttempts = REOPEN_ATTEMPTS;
  int tryAgain = TRUE;
  int totalFileSize;
  long compare;
  long actualChecksum;

  if (DataFile) {
    bufDataFile = DataFile;
    headerSize   = ((long*) bufDataFile)[3];
    numDataFiles = ((long*) bufDataFile)[5];
    return;
  }

  // Give a couple chances in case the CD library barfs once or twice
  while (tryAgain && reopenAttempts) {
    tryAgain = FALSE;
    reopenAttempts--;

    // Try to open the data file
    insertConsoleString("Looking for dat file");

    fpDataFile = myfopen(fileDat);
    if (!fpDataFile) {
      appendConsoleString(" - not found");
      msprintf(tmpStr, "cdError=%d", cdError);
      insertConsoleString(tmpStr);
      return;
    }

    msprintf(strInit3, "Data file");
    //    if (reopenAttempts == (REOPEN_ATTEMPTS-1)) msprintf(strInit3, "Data file");
    //    else                                       msprintf(strInit3, "Data file (retry)");
    redraw |= kRedrawInit3;
    updateScreen();
    
    appendConsoleString(" - found");
  
    insertConsoleString("Reading data file header");

    // Read the file header.  This is sixteen bytes long and includes:
    // o the text "BOOTLOAD"
    // o the size of the initial section (everything but the executables)
    // o the size of the full file
    bytesRead = myfread(buf, 16, fpDataFile);

    if ((bytesRead != 16) || cdError) {
      if (reopenAttempts) {
	insertConsoleString("**WARN**");
	tryAgain = TRUE;
      } else {
	insertConsoleString("**ERR**");
      }
      if (cdError)
	appendConsoleString(" Error reading header");
      else {
	appendConsoleString(" Didn't get all bytes");
	msprintf(tmpStr, "   Exp=%d", 16);
	insertConsoleString(tmpStr);
	msprintf(tmpStr, "   Act=%d", bytesRead);
	insertConsoleString(tmpStr);
      }
      
    } else {
      // Read the first 16 bytes

      insertConsoleString("   Checking signature");
      
      // Now check BOOT
      compare = ('B' << 24) | ('O' << 16) | ('O' << 8) | ('T' << 0);
      if (buf[0] != compare) {
	if (reopenAttempts) {
	  insertConsoleString("**WARN**");
	  tryAgain = TRUE;
	} else {
	  insertConsoleString("**ERR**");
	}
	appendConsoleString(" Signature mismatch");
	msprintf(tmpStr, "   BOOT != %04lx", compare);
	insertConsoleString(tmpStr);
	
      } else {
	// BOOT matched
	// Now check LOAD
	
	compare = ('L' << 24) | ('O' << 16) | ('A' << 8) | ('D' << 0);
	if (buf[1] != compare) {
	  if (reopenAttempts) {
	    insertConsoleString("**WARN**");
	    tryAgain = TRUE;
	  } else {
	    insertConsoleString("**ERR**");
	  }
	  appendConsoleString(" Signature mismatch");
	  msprintf(tmpStr, "   LOAD != %04lx", compare);
	  insertConsoleString(tmpStr);
	  
	} else {
	  // BOOT matched
	  // LOAD matched
	  // Now check the file size
	  
	  insertConsoleString("   Checking file size");
	  
	  totalFileSize = buf[2];
	  headerSize    = buf[3];

	  // Make sure the total file size matches what we expect
	  actualFileSize = getFileSize(fpDataFile);
	  if (actualFileSize != totalFileSize) {
	    if (reopenAttempts) {
	      insertConsoleString("**WARN**");
	      tryAgain = TRUE;
	    } else {
	      insertConsoleString("**ERR**");
	    }
	    appendConsoleString(" File size mismatch");
	    msprintf(tmpStr, "   Exp=%d", totalFileSize);
	    insertConsoleString(tmpStr);
	    msprintf(tmpStr, "   Act=%d", actualFileSize);
	    insertConsoleString(tmpStr);
	  } else {
	    // BOOT matched
	    // LOAD matched
	    // File size matched
	    // Now read the full header

	    insertConsoleString("   Reading full header");
	    
	    // Allocate space for the full header, leaving room to pad to a dword boundary
	    bufDataFile = (char *) malloc(sizeof(char) * (headerSize + 3));

	    if (!bufDataFile) {
	      if (reopenAttempts) {
		insertConsoleString("**WARN**");
		tryAgain = TRUE;
	      } else {
		insertConsoleString("**ERR**");
	      }
	      msprintf(tmpStr, "Couldn't allocate %d bytes", headerSize+3);
	      appendConsoleString(tmpStr);
	      
	    } else {
	      // Read the initial part of the data file.  This includes the applist.txt, background,
	      // app images, and app descriptions.  Executables are not read since they are too
	      // large to fit them all in memory at the same time.
	      myfseek(fpDataFile, 0, SEEK_SET);
	      bytesRead = myfread(bufDataFile, headerSize, fpDataFile);

	      if ((bytesRead != headerSize) || cdError) {
		if (reopenAttempts) {
		  insertConsoleString("**WARN**");
		  tryAgain = TRUE;
		} else {
		  insertConsoleString("**ERR**");
		}
		if (cdError)
		  appendConsoleString(" Error reading header");
		else {
		  appendConsoleString(" Didn't read full header");
		  msprintf(tmpStr, "   Exp=%d", headerSize);
		  insertConsoleString(tmpStr);
		  msprintf(tmpStr, "   Act=%d", bytesRead);
		  insertConsoleString(tmpStr);
		}
	      
	      } else {
		// BOOT matched
		// LOAD matched
		// File size matched
		// Successfully read full header
		// Now check checksum

		insertConsoleString("   Verifying checksum");
	      
		// pad zeros to a dword boundary
		while (bytesRead & 0x3) {
		  bufDataFile[bytesRead++] = 0;
		}

		// Verify the checksum of the initial portion
		actualChecksum = getChecksum(bufDataFile, bytesRead);

		if (actualChecksum) {
		  if (reopenAttempts) {
		    insertConsoleString("**WARN**");
		    tryAgain = TRUE;
		  } else {
		    insertConsoleString("**ERR**");
		  }
		  appendConsoleString(" Checksum non-zero: ");
		  msprintf(tmpStr, "0x%08lx", actualChecksum);
		  appendConsoleString(tmpStr);
		
		} else {
		  // BOOT matched
		  // LOAD matched
		  // File size matched
		  // Successfully read full header
		  // Checksum matched
		
		  insertConsoleString("   VERIFIED!");

		  numDataFiles = ((long*) bufDataFile)[5];
		
		}  // if (headerChecksum != actualChecksum)
	      }  // if (bytesRead != headerSize)
	    }  // if (!bufDataFile)
	  }  // if (actualFileSize != totalFileSize)
	}  // if (buf[1] != compare)
      }  // if (buf[0] != compare)
    }  // if ((bytesRead != 16) || cdError)

    // If something failed along the way free everything and try again from scratch
    if (tryAgain) {
      if (fpDataFile ) { myfclose(fpDataFile ); fpDataFile  = NULL; }
      if (bufDataFile) { free    (bufDataFile); bufDataFile = NULL; }
      if (reopenAttempts) insertConsoleString("Trying again...");
    }
    
  }  // while (tryAgain)
}

//==================================================================================================
// getChecksum()
// 
// Generates a checksum of a given buffer.  This is done dword-by-dword.  Returns the final sum.
//
// Inputs:
//   buffer - a pointer to the buffer
//   bytes  - the number of bytes to verify (should be a multiple of 4)
// Outputs: none
// Globals modified: none
//
//==================================================================================================

long getChecksum(char *buffer, int bytes)
{
  unsigned char *ptr;
  long checksum = 0;
  long val;
  
  for (ptr = buffer; bytes; bytes -= 4, ptr += 4) {
    val = (ptr[0] << 24) | (ptr[1] << 16) | (ptr[2] << 8) | (ptr[3] << 0);
    checksum += val;
  }

  return checksum;
}

//==================================================================================================
// getAppList()
//
// Reads the applist.txt file.  The file contains a list of application directories in which to
// look for executables and related files.  The list also has parameters that change the look-and-
// feel of the boot loader.
//
// After this completes the app info array will be initialized with each app's directory name.
//
// Inputs:  none
// Outputs:
//   int - returns either RET_NORMAL or RET_ERROR
// Globals modified:
//   strInit3   - initialization string displaying what is currently happening
//   bgImg      - pointer to the background image (NULL if none found)
//   redraw     - flags to indicate what should be redrawn
//   aAppInfo[] - array of structures, one for each app
//
//==================================================================================================

int getAppList()
{
  int i;
  int r, g, b;
  int lineNum;
  int appNum;
  int returnValue = RET_NORMAL;
  int charsLeft;
  char *fileBuffer, *bufPtr;
  char *tmpPtr;
  int print;
  char buf[80];
  int matched;
#ifndef OLD_TGA
  long sec, prevSec;
#endif

#ifdef TRACE
  printf("%s() begin\n", __func__);
#endif
  
  // Read background image if it exists
  msprintf(strInit3, "Background");
  redraw |= kRedrawInit3;
  insertConsoleString("Checking for background image");
#ifndef OLD_TGA
  g_img = &bgImg;
#endif
  BG_IMG_PTR = checkAndReadImg(NULL, fileImgBG);
#ifndef OLD_TGA
  if (BG_IMG_PTR) {
    bgImg.w = g_imgW;  // grab the dimensions from the globals
    bgImg.h = g_imgH;
  }
#endif

#ifndef OLD_TGA
  // Wait for the helper MPE to be done rendering the image
  _TimeElapsed(&prevSec, NULL);
  while (bgImg.status == kImgStatusSubmitted) {
    // Time out after a bit
    _TimeElapsed(&sec, NULL);
    if ((sec - prevSec) > 5) {
      insertConsoleString("**ERR** Background image render timeout");
      break;
    }
  }
#endif

  // Ensure the background is the correct size
#ifdef OLD_TGA
  if (bgImg && ((bgImg->wide != 720) || (bgImg->high != 480))) {
#else
  if (BG_IMG_PTR && ((bgImg.w != SCREEN_W) || (bgImg.h != SCREEN_H))) {
#endif
    insertConsoleString("**ERR** Background must be 720x480");
    mmlReleasePixmaps((mmlPixmap *) BG_IMG_PTR, &sysRes, 1);
    BG_IMG_PTR = NULL;
  }

  // Read sounds if they exist
  msprintf(strInit3, "Sounds");
  redraw |= kRedrawInit3;
  insertConsoleString("Checking for sounds");
  wavMove   = checkAndReadFile(NULL, fileWavMove,   kFileTypeWAV);
  wavSelect = checkAndReadFile(NULL, fileWavSelect, kFileTypeWAV);

  // Now read the app list file
  msprintf(strInit3, "App list");
  redraw |= kRedrawInit3;
  
  insertConsoleString("Getting application list data from data file");

  strcpy(tmpStr, HOST_PATH);
  strcat(tmpStr, fileAppList);
  fileBuffer = getDataFilePtr(tmpStr, &curFileSize);
			    
  if (!fileBuffer || (curFileSize == -1)) {
    insertConsoleString("**ERR** In getAppList, myReadFile failed!");
    return(RET_ERROR);
  }

  // Got the app list file, now parse it
  
  lineNum = 1;
  numApps = 0;

  charsLeft = curFileSize;
  bufPtr = fileBuffer;

  // The first pass parses parameters and counts apps
  while (charsLeft) {
    mygets(tmpStr, TMP_STR_SIZE, &bufPtr, &charsLeft);
    
#ifdef APPLIST_PRINT
    printf("\n%s", tmpStr);
#endif
    
    print = TRUE;
    switch (tmpStr[0]) {
      case '/':  numApps++; break;
      case '+':
	//
	// Parameter
	//
	matched = FALSE;
	redraw |= kRedrawAll;

	if (!strncmp(&tmpStr[0], strSection, strlen(strSection))) {
	  numApps++;
	}

	// Color parameter?
	for (i = 0; (i < kColorMax) && !matched; i++) {
	  if (!strncmp(aParamColor[i], &tmpStr[1], strlen(aParamColor[i])) &&
	      (tmpStr[strlen(aParamColor[i])+1] == '=')) {
	    
	    matched = TRUE;
	    tmpPtr = &tmpStr[strlen(aParamColor[i])+2];
	    r = strtol(tmpPtr, &tmpPtr, 10);
	    g = strtol(tmpPtr, &tmpPtr, 10);
	    b = strtol(tmpPtr, &tmpPtr, 10);

	    if ((r < 0) || (r > 255)) {
	      msprintf(buf, "**ERR** Red value out of range: %d", r);
	      insertConsoleString(buf);
	    } else if ((g < 0) || (g > 255)) {
	      msprintf(buf, "**ERR** Green value out of range: %d", g);
	      insertConsoleString(buf);
	    } else if ((b < 0) || (b > 255)) {
	      msprintf(buf, "**ERR** Blue value out of range: %d", b);
	      insertConsoleString(buf);
	    } else {
	      aColors[i] = mmlColorFromRGB(r, g, b);
	    }
	  }
	}

	// String parameter?
	for (i = 0; (i < kStrMax) && !matched; i++) {
	  if (!strncmp(aParamString[i], &tmpStr[1], strlen(aParamString[i])) &&
	      (tmpStr[strlen(aParamString[i])+1] == '=')) {
	    
	    matched = TRUE;

	    if (aStrings[i]) {
	      aStrings[i]--;  // grab the parse flag back
	      free(aStrings[i]);
	    }
	      
	    tmpPtr = &tmpStr[strlen(aParamString[i])+2];
	    aStrings[i] = (char *) malloc (sizeof(char) * (strlen(tmpPtr) + 1));
	    if (!aStrings[i]) {
	      returnValue = RET_ERROR;
	      insertConsoleString("**ERR** Could not allocate string space");
	    } else {
	      aStrings[i]++;  // allow one character for the parse flag
	      memcpy(aStrings[i], tmpPtr, strlen(tmpPtr) + 1);
	      while (aStrings[i][strlen(aStrings[i])-1] < ' ') {
		aStrings[i][strlen(aStrings[i])-1] = 0;
	      }
	      aStrings[i][-1] = 0;  // set flag indicating this string needs to be parsed
	      //	      printf("%s() - String override %d, '%s'\n", __func__, i, aStrings[i]);
	    }
	  }
	}
	// Re-select the instruction string in case it changed
	instrStr = (controlType == kControlRemote)
	  ? aStrings[kStrInstr1Remote] : aStrings[kStrInstr1Normal];

 	// User font name parameter?
	if (!strncmp(strParamUserFont, &tmpStr[1], strlen(strParamUserFont)) &&
	    (tmpStr[strlen(strParamUserFont)+1] >= '1') &&
	    (tmpStr[strlen(strParamUserFont)+1] <= '9') &&
	    (tmpStr[strlen(strParamUserFont)+2] == '=')) {
	  matched = TRUE;

	  // Get font index
	  r = tmpStr[strlen(strParamUserFont)+1] - '1';

	  if ((r >= 0) && (r < MAX_USER_FONTS)) {
	    // Get pointer to font name
	    aUserFontNames[r] =
	      (char *) malloc(sizeof(char) * (strlen(&tmpStr[strlen(strParamUserFont)+3]) + 1));
	    if (!aUserFontNames[r]) {
	      msprintf(buf, "**ERR** Could not allocate memory for font name '%s'",
		       &tmpStr[strlen(strParamUserFont)+3]);
	      insertConsoleString(buf);
	    }

	    // Get the font name
	    strcpy(aUserFontNames[r], &tmpStr[strlen(strParamUserFont)+3]);

	    // Remove trailing new lines or other white space
	    while (isspace((int) aUserFontNames[r][strlen(aUserFontNames[r])-1])) {
	      aUserFontNames[r][strlen(aUserFontNames[r])-1] = 0;
	    }

	    // Read font if it exists
	    tmpPtr = checkAndReadFile(NULL, aUserFontNames[r], kFileTypeTTF);
	    if (tmpPtr) {
	      initUserFont(r, tmpPtr, curFileSize);
	    } else {
	      msprintf(buf, "**ERR** Could not find font '%s'", aUserFontNames[r]);
	      insertConsoleString(buf);
	    }

	  } else {
	    msprintf(buf, "**ERR** User font index out of range: %d", r);
	    insertConsoleString(buf);
	  }
	}

 	// Font size parameter?
	for (i = 0; (i < kStrFontIndexMax) && !matched; i++) {
	  if (!strncmp(aParamFontSize[i], &tmpStr[1], strlen(aParamFontSize[i])) &&
	      (tmpStr[strlen(aParamFontSize[i])+1] == '=')) {
	    matched = TRUE;
	    tmpPtr = &tmpStr[strlen(aParamFontSize[i])+2];

	    // Make sure the font size is within the valid range
	    r = strtol(tmpPtr, NULL, 10);

	    if ((r >= MIN_FONT_SIZE) && (r <= MAX_FONT_SIZE)) {
	      aFontSizes[i] = r;
	    } else {
	      msprintf(buf, "**ERR** Font size out of range: %d", r);
	      insertConsoleString(buf);
	    }
	  }
	}

	// Font index parameter?
	for (i = 0; (i < kStrFontIndexMax) && !matched; i++) {
	  if (!strncmp(aParamFontIndices[i], &tmpStr[1], strlen(aParamFontIndices[i])) &&
	      (tmpStr[strlen(aParamFontIndices[i])+1] == '=')) {
	    matched = TRUE;
	    tmpPtr = &tmpStr[strlen(aParamFontIndices[i])+2];

	    // Make sure the font index is within the valid range
	    r = strtol(tmpPtr, NULL, 10);

	    if ((r >= MIN_FONT_INDEX) && (r <= MAX_FONT_INDEX)) {
	      if (r < 0) {
		r = -r + MAX_USER_FONTS;
	      }
	      aFontIndices[i] = r;
	    } else {
	      msprintf(buf, "**ERR** Font index out of range: %d", r);
	      insertConsoleString(buf);
	    }
	  }
	}

 	// Font alignment parameter?
	for (i = 0; (i < kTextHAlignMax) && !matched; i++) {
	  if (!strncmp(aParamHAlign[i], &tmpStr[1], strlen(aParamHAlign[i])) &&
	      (tmpStr[strlen(aParamHAlign[i])+1] == '=')) {
	    matched = TRUE;
	    tmpPtr = &tmpStr[strlen(aParamHAlign[i])+2];

	    if (strncmp(tmpPtr, textAlignLeft, strlen(textAlignLeft)) == 0) {
	      aTextHAlign[i] = kHAlignLeft;
	    } else if (strncmp(tmpPtr, textAlignCenter, strlen(textAlignCenter)) == 0) {
	      aTextHAlign[i] = kHAlignCenter;
	    } else if (strncmp(tmpPtr, textAlignRight, strlen(textAlignRight)) == 0) {
	      aTextHAlign[i] = kHAlignRight;
	    } else {
	      msprintf(buf, "**ERR** Unrecognized parameter for %s", aParamHAlign[i]);
	      insertConsoleString(buf);
	      insertConsoleString("Expecting one of:");
	      msprintf(buf, "   %s", textAlignLeft);
	      insertConsoleString(buf);
	      msprintf(buf, "   %s", textAlignCenter);
	      insertConsoleString(buf);
	      msprintf(buf, "   %s", textAlignRight);
	      insertConsoleString(buf);
	    }
	  }
	}

	// Spacing parameter?
	for (i = 0; (i < kSpaceMax) && !matched; i++) {
	  fflush(stdout);
	  if (!strncmp(aParamSpacing[i], &tmpStr[1], strlen(aParamSpacing[i])) &&
	      (tmpStr[strlen(aParamSpacing[i])+1] == '=')) {
	    matched = TRUE;
	    tmpPtr = &tmpStr[strlen(aParamSpacing[i])+2];

	    // Make sure the spacing is within the valid range
	    r = strtol(tmpPtr, NULL, 10);

	    if ((r >= MIN_SPACING) && (r <= MAX_SPACING)) {
	      aSpacing[i] = r;
	    } else {
	      msprintf(buf, "**ERR** Spacing value out of range: %d", r);
	      insertConsoleString(buf);
	    }
	  }
	}

	// Border parameter?
	for (i = 0; (i < kBorderMax) && !matched; i++) {
	  if (!strncmp(aParamBorder[i], &tmpStr[1], strlen(aParamBorder[i])) &&
	      (tmpStr[strlen(aParamBorder[i])+1] == '=')) {
	    matched = TRUE;
	    tmpPtr = &tmpStr[strlen(aParamBorder[i])+2];

	    // Make sure the font size is within the valid range
	    r = strtol(tmpPtr, NULL, 10);

	    if ((r >= MIN_BORDER) && (r <= MAX_BORDER)) {
	      aBorders[i] = r;
	    } else {
	      msprintf(buf, "**ERR** Border value out of range: %d", r);
	      insertConsoleString(buf);
	    }
	  }
	}

	// +ShowConsole?
	if (!matched && !strncmp(strParamShowConsole, &tmpStr[1], strlen(strParamShowConsole))) {
	  matched = TRUE;
	  nextState = kStateConsoleAndMenu;
	  redraw |= kRedrawAll;
	}

	// +AbsolutePositioning?
	if (!matched && !strncmp(strParamAbsolutePos, &tmpStr[1], strlen(strParamAbsolutePos))) {
	  matched = TRUE;
	  absolutePositioning = TRUE;
	  insertConsoleString("Using absolute positioning");
	}

	// Element position?
	for (i = 0; (i < kPosMax) && !matched; i++) {
	  if (!strncmp(aParamPos[i], &tmpStr[1], strlen(aParamPos[i])) &&
	      (tmpStr[strlen(aParamPos[i])+1] == '=')) {
	    matched = TRUE;
	    tmpPtr = &tmpStr[strlen(aParamPos[i])+2];

	    aPos[i].x = strtol(tmpPtr, &tmpPtr, 10);
	    aPos[i].y = strtol(tmpPtr, &tmpPtr, 10);
	    aPos[i].x2 = strtol(tmpPtr, &tmpPtr, 10);
	    aPos[i].y2 = strtol(tmpPtr, &tmpPtr, 10);
	    aPos[i].w = (aPos[i].x2 == -1) ? -1 : (aPos[i].x2 - aPos[i].x + 1);
	    aPos[i].h = (aPos[i].y2 == -1) ? -1 : (aPos[i].y2 - aPos[i].y + 1);
	  }
	}

	// Image rotation
	for (i = 0; (i < kRotateImgMax) && !matched; i++) {
	  if (!strncmp(aParamRotateImg[i], &tmpStr[1], strlen(aParamRotateImg[i])) &&
	      (tmpStr[strlen(aParamRotateImg[i])+1] == '=')) {
	    matched = TRUE;
	    tmpPtr = &tmpStr[strlen(aParamRotateImg[i])+2];

	    if (strncmp(tmpPtr, rotateImgNever, strlen(rotateImgNever)) == 0) {
	      aRotateImg[i] = 0;
	    } else if (strncmp(tmpPtr, rotateImgOnHighlight, strlen(rotateImgOnHighlight)) == 0) {
	      aRotateImg[i] = -1;
	    } else {
	      r = strtol(tmpPtr, NULL, 10);
	      if ((r >= MIN_ROTATE_IMG_DELAY) && (r <= MAX_ROTATE_IMG_DELAY)) {
		aRotateImg[i] = r;
	      } else {
		aRotateImg[i] = 0;
		msprintf(buf, "**ERR** Unrecognized parameter for %s", aParamRotateImg[i]);
		insertConsoleString(buf);
		insertConsoleString("Expecting one of:");
		msprintf(buf, "   %s", rotateImgNever);
		insertConsoleString(buf);
		msprintf(buf, "   %s", rotateImgOnHighlight);
		insertConsoleString(buf);
		msprintf(buf, "   Integer from %d to %d",MIN_ROTATE_IMG_DELAY,MAX_ROTATE_IMG_DELAY);
		insertConsoleString(buf);
	      }
	    }
	  }
	}

	// +MenuHorizontal?
	if (!matched && !strncmp(strParamMenuHorizontal, &tmpStr[1], 
				 strlen(strParamMenuHorizontal))) {
	  matched = TRUE;
	  menuOrientation = kMenuHorizontal;
	  if (strcmp(aStrings[kStrArrowPrev], "^") == 0) {
	    free(aStrings[kStrArrowPrev]);
	    aStrings[kStrArrowPrev] = (char *) malloc (sizeof(char) * (strlen("<")+1));
	    if (!aStrings[kStrArrowPrev]) {
	      returnValue = RET_ERROR;
	      insertConsoleString("**ERR** Could not allocate string space");
	    } else {
	      aStrings[kStrArrowPrev]++;  // allow one character for the parse flag
	      memcpy(aStrings[kStrArrowPrev], "<", 1);
	      aStrings[i][-1] = 0;  // set flag indicating this string needs to be parsed
	    }
	  }
	  if (strcmp(aStrings[kStrArrowNext], "v") == 0) {
	    free(aStrings[kStrArrowNext]);
	    aStrings[kStrArrowNext] = (char *) malloc (sizeof(char) * (strlen(">")+1));
	    if (!aStrings[kStrArrowNext]) {
	      returnValue = RET_ERROR;
	      insertConsoleString("**ERR** Could not allocate string space");
	    } else {
	      aStrings[kStrArrowNext]++;  // allow one character for the parse flag
	      memcpy(aStrings[kStrArrowNext], ">", 1);
	      aStrings[i][-1] = 0;  // set flag indicating this string needs to be parsed
	    }
	  }
	}

	// +MenuMaxEntries?
	if (!matched &&!strncmp(strParamMenuMaxEntries,&tmpStr[1],strlen(strParamMenuMaxEntries))) {

	  matched = TRUE;
	  tmpPtr = &tmpStr[strlen(strParamMenuMaxEntries)+2];

	  // Make sure the number is greater than zero
	  r = strtol(tmpPtr, NULL, 10);

	  if ((r >= MENU_MAX_ENTRIES_MIN) && (r <= MENU_MAX_ENTRIES_MAX)) {
	    menuMaxEntries = r;
	  } else {
	    msprintf(buf, "**ERR** Max menu entries must %d-%d",
		     MENU_MAX_ENTRIES_MIN, MENU_MAX_ENTRIES_MAX);
	    insertConsoleString(buf);
	    menuMaxEntries = -1;
	  }
	}

	// +MenuWrap?
	if (!matched && !strncmp(strParamMenuWrap, &tmpStr[1], strlen(strParamMenuWrap))) {
	  matched = TRUE;
	  menuWrap = TRUE;
	}

	// +MenuNoHighlight?
	if (!matched && !strncmp(strParamMenuNoHighlight, &tmpStr[1],
				 strlen(strParamMenuNoHighlight))) {
	  matched = TRUE;
	  menuNoHighlight = TRUE;
	}

	if (!matched) {
	  insertConsoleString("Unrecognized parameter");
	}

	break;
      default:
	print = FALSE;
	break;
    }
    print = TRUE;
    if (print) {
      msprintf(buf, "   Line %d: %s", lineNum, tmpStr);
      insertConsoleString(buf);
    }
    lineNum++;
  }
  
  insertConsoleString("   EOF");
  
  msprintf(tmpStr, "Found %d apps", numApps);
  insertConsoleString(tmpStr);

  // Allocate space for the given number of apps
  aAppInfo = (sApp *) malloc (sizeof(sApp) * numApps);

  if (!aAppInfo) {
    insertConsoleString("**ERR** Could not allocate aAppInfo");
    return RET_ERROR;
  }

  // Initialize the app elements
  for (appNum = 0; appNum < numApps; appNum++) {
    aAppInfo[appNum].flags        = 0;
    aAppInfo[appNum].dirName      = NULL;
    aAppInfo[appNum].txtTitle     = NULL;
    aAppInfo[appNum].txtDesc      = NULL;
    aAppInfo[appNum].numImgGroups = 0;
    aAppInfo[appNum].aImgGroups   = NULL;
    aAppInfo[appNum].imgIntro.status = kImgStatusInit;
    aAppInfo[appNum].imgIntro.pixmap = NULL;
    aAppInfo[appNum].imgIntro.parent = NULL;
  }

  //
  // Now fill in the app names
  //
  insertConsoleString("Getting individual app info");

  appNum = 0;
  charsLeft = curFileSize;
  bufPtr = fileBuffer;
  
  while (charsLeft && (appNum < numApps)) {
    
    mygets(tmpStr, TMP_STR_SIZE, &bufPtr, &charsLeft);

    if (tmpStr[0] == '/') {
      if (curApp == -1) {
	curApp = appNum;
      }
      
      // Remove trailing white space
      while (tmpStr[strlen(tmpStr)-1] <= ' ') {
	tmpStr[strlen(tmpStr)-1] = 0;
      }
      // Add a slash at end if not already there
      i = strlen(tmpStr);
      if (tmpStr[i-1] != '/') {
	tmpStr[i+0] = '/';
	tmpStr[i+1] = 0;
      }

      aAppInfo[appNum].dirName  = (char *) malloc (sizeof(char *) * (strlen(tmpStr) + 1));
      strcpy(aAppInfo[appNum].dirName, tmpStr);
      appNum++;
    }

    if (!strncmp(&tmpStr[0], "+Section", strlen("+Section"))) {
      // Skip over "+Section=" to get to the name
      tmpPtr = &tmpStr[strlen("+Section=")];

      // Remove trailing white space
      while (tmpPtr[strlen(tmpPtr)-1] < ' ') {
	tmpPtr[strlen(tmpPtr)-1] = 0;
      }
      
      aAppInfo[appNum].dirName  = (char *) malloc (sizeof(char) * (strlen(tmpPtr) + 1));
      strcpy(aAppInfo[appNum].dirName, tmpPtr);
      aAppInfo[appNum].flags |= kFlagSectionHeader;
      appNum++;
    }
  }

#ifdef TRACE
  printf("%s() end\n", __func__);
#endif
  
  return(returnValue);
}

//==================================================================================================
// getAppData()
//
// This function performs two actions.  It verifies the existence of an app's executable.  It also
// reads in the app's title/description file and inserts that into the app info array.
//
// This function must be called continuously until it returns non-zero.  It only operates on one 
// app each time it is called.  This was done so the screen can be painted in between apps.
//
// Inputs:  none
// Outputs:
//   int - status can be RET_NORMAL or RET_DONE (done reading apps)
// Globals modified:
//   aAppInfo[] - the array of structures containing information about each app
//
//==================================================================================================

int getAppData()
{
  int i, j, k;
  static int appNum = 0;
  int charIndex;
  char *ptr;
  int imgGroupNum, imgNum;
  int curImg;
  sApp *app;
  sAppImgGroup *imgGroup;

  if (appNum >= numApps) return RET_DONE;

  app = &aAppInfo[appNum];

#ifdef TRACE
  printf("%s() begin\n", __func__);
#endif

  // Section headers are skipped
  if (app->flags & kFlagSectionHeader) {
    insertConsoleString("   Section: ");
    appendConsoleString(app->dirName);
    msprintf(strInit3, "Section: %s", app->dirName);
    redraw |= kRedrawInit3;
    appNum++;

#ifdef TRACE
  printf("%s() end\n", __func__);
#endif

    return RET_NORMAL;
  }

  //  msprintf(strInit3, "App #%d: %s", appNum + 1, app->dirName);
  msprintf(strInit3, "%s", app->dirName);
  redraw |= kRedrawInit3;

  insertConsoleString("   ");
  appendConsoleString(app->dirName);

  // Make sure the executable file exists for this app.
  if (checkAndReadFile(app->dirName, fileAppExe1, kFileTypeExe)) {
    app->flags |= kFlagAppValid;
  }

  if (!(app->flags & kFlagAppValid)) {
    if (checkAndReadFile(app->dirName, fileAppExe2, kFileTypeExe)) {
      app->flags |= kFlagAppValid;
    }
  }

  redraw |= kRedrawMenu;

  // If the executable was not found, print an error
  if (!(app->flags & kFlagAppValid)) {
    insertConsoleString("**ERR** Could not find executable for ");
    appendConsoleString(app->dirName);

    appNum++;

#ifdef TRACE
  printf("%s() end\n", __func__);
#endif

    return RET_NORMAL;
  }

  //
  // Read app title and description
  //
  app->txtDesc =
    (char *) checkAndReadFile(app->dirName, fileAppText, kFileTypeText);
    
  // If we found a description file, separate it into the title and description
  if (app->txtDesc) {

    // Find the newline
    for (charIndex = 0; app->txtDesc[charIndex]; charIndex++) {
      if (app->txtDesc[charIndex] == '\n') {
	break;
      }
    }
      
    if (app->txtDesc[charIndex] == 0) {
      // Only found one line so use it as the title
	
      // Copy the correct number of bytes.
      app->txtTitle = (char *) malloc ((sizeof(char) * charIndex) + 1);
      app->txtTitle++;  // allow one character for the parse flag 
      strncpy(app->txtTitle, app->txtDesc, charIndex);
      app->txtTitle[charIndex] = 0;
      app->txtTitle[-1] = 0;  // set flag indicating this string needs to be parsed

      app->txtDesc  = NULL;

    } else {
      // Found multiple lines.

      //
      // First is the title
      //

      // Copy the correct number of bytes.
      app->txtTitle = (char *) malloc ((sizeof(char) * charIndex) + 1);
      app->txtTitle++;  // allow one character for the parse flag 
      strncpy(app->txtTitle, app->txtDesc, charIndex);
      app->txtTitle[charIndex] = 0;
      app->txtTitle[-1] = 0;  // set flag indicating this string needs to be parsed

      // Remove non-printable characters from the end
      while (strlen(app->txtTitle) &&
	     (app->txtTitle[strlen(app->txtTitle)-1] < ' ')) {
	app->txtTitle[strlen(app->txtTitle)-1] = 0;
      }

      // If no more characters are left then free the space
      if (!strlen(app->txtTitle)) {
	app->txtTitle--;  // grab the flag back
	free(app->txtTitle);
	app->txtTitle = NULL;
      }

      //
      // The rest is the description
      //
	
      // First remove characters until we reach the first printable one
      do {
	charIndex++;
      } while ((app->txtDesc[charIndex] > 0  ) &&
	       (app->txtDesc[charIndex] < ' '));

      // Save the old src pointer so we can free it later
      ptr = app->txtDesc;
	
      // Check if there's no printable text
      if (app->txtDesc[charIndex] != 0) {

	// Allocate the correct number of bytes.
	app->txtDesc =
	  (char *) malloc (sizeof(char) * (strlen(ptr) - charIndex + 1));
	app->txtDesc++;  // allow one character for the parse flag 
	strncpy(app->txtDesc, &ptr[charIndex], strlen(ptr) - charIndex);
	app->txtDesc[strlen(ptr) - charIndex] = 0;
	app->txtDesc[-1] = 0;  // set flag indicating this string needs to be parsed

	// Remove all non-printable characters from the end
	while (strlen(app->txtDesc) &&
	       (app->txtDesc[strlen(app->txtDesc)-1] < ' ')) {
	  app->txtDesc[strlen(app->txtDesc)-1] = 0;
	}

      } else {
	app->txtDesc = NULL;
      }
/*
      // Remove non-printable characters from the string
      ptr = dst = app->txtDesc;
      while (ptr[0] != 0) {
	// Only keep if it is printable
	if (ptr[0] >= ' ') {
	  *dst = ptr[0];
	  dst++;
	}
	ptr++;
      }
      *dst = 0;
*/
    }
      
    //      printf("\nTitle: *%s*", app->txtTitle);
    //      printf("\nDesc: *%s*", app->txtDesc);
    //      fflush(stdout);
      
    redraw |= kRedrawMenu;
    redraw |= kRedrawAppStuff;
  }  // if (app->txtDesc)

  //
  // Get the number of image groups
  //
  /*
  insertConsoleString("      Counting number of groups");

  for (imgGroupNum = 0; imgGroupNum < 9; imgGroupNum++) {
    fileAppImgA[3] = '1' + imgGroupNum;
    fileAppImgB[3] = '1' + imgGroupNum;

    // First look for the standard file name
    if (checkImg(app->dirName, fileAppImgA)) {
      msprintf(tmpStr, "         Group %c found", imgGroupNum + '1');
      insertConsoleString(tmpStr);
      app->numImgGroups++;
    } else {
      // Couldn't find the standard file name - check the new ones
      for (imgNum = 0; imgNum < 26; imgNum++) {
	fileAppImgB[4] = 'a' + imgNum;
	if (checkImg(app->dirName, fileAppImgB)) {
	  msprintf(tmpStr, "         Group %c found", imgGroupNum + '1');
	  insertConsoleString(tmpStr);
	  app->numImgGroups++;
	  // Only need to find one per group this time through
	  break;
	}
      }
    }
  }
  */
  /*
  if (app->numImgGroups < 2) {
    app->numImgGroups = 2;
  }
  */
  app->numImgGroups = 9;

  // Allocate space for the groups
  app->aImgGroups =
    (sAppImgGroup*) malloc(sizeof(sAppImgGroup) * app->numImgGroups);

  if (app->aImgGroups == NULL) {
    msprintf(tmpStr, "**ERR** Could not allocate room for %d image groups",
	     app->numImgGroups);
    insertConsoleString(tmpStr);
    app->numImgGroups = 0;

    appNum++;

#ifdef TRACE
  printf("%s() end\n", __func__);
#endif

    return RET_NORMAL;
  }

  // Clear out the image groups we just allocated
  for (imgGroupNum = 0; imgGroupNum < app->numImgGroups; imgGroupNum++) {
    app->aImgGroups[imgGroupNum].aAppImg    = NULL;
    app->aImgGroups[imgGroupNum].numImg     = 0;
    app->aImgGroups[imgGroupNum].curImg     = 0;
    app->aImgGroups[imgGroupNum].speed      = 0;
    app->aImgGroups[imgGroupNum].order      = NULL;
    app->aImgGroups[imgGroupNum].draw       = 0;
    app->aImgGroups[imgGroupNum].time       = -1;
  }

  insertConsoleString("      Counting number of images per group");

  //
  // Now get the number of images in each group
  //
  
  for (imgGroupNum = 0; imgGroupNum < 9; imgGroupNum++) {

    // Get a pointer to the image group since we'll be doing a lot with it
    imgGroup = &app->aImgGroups[imgGroupNum];

    fileAppImgA[3] = '1' + imgGroupNum;
    fileAppImgB[3] = '1' + imgGroupNum;

    // First look for the standard file name
    if (checkImg(app->dirName, fileAppImgA)) {
      imgGroup->numImg++;
    }

    // Now check the new image file names
    for (imgNum = 0; imgNum < 26; imgNum++) {
      fileAppImgB[4] = 'a' + imgNum;
      if (checkImg(app->dirName, fileAppImgB)) {
	imgGroup->numImg++;
      }
    }

    // If this group had images then allocate space for them
    if (imgGroup->numImg > 0) {

      msprintf(tmpStr,"         Group %c has %d images", imgGroupNum + '1', imgGroup->numImg);
      insertConsoleString(tmpStr);

      // Initialize the image order array if there's more than one and it's not static
      if ((imgGroup->numImg > 1) && (aRotateImg[kRotateImg1 + imgGroupNum] != 0)) {
	imgGroup->order = (int*) malloc (sizeof(int) * imgGroup->numImg);
	if (imgGroup->order == NULL) {
	  msprintf(tmpStr, "**ERR** Could not allocate order array for group %d", imgGroupNum);
	  insertConsoleString(tmpStr);
	} else {
	  // First initialize the array to be sequential
	  for (i = 0; i < imgGroup->numImg; i++) {
	    imgGroup->order[i] = i;
	  }
	  // Now shuffle the entries
	  for (i = 0; i < imgGroup->numImg; i++) {
	    j = rand() % imgGroup->numImg;
	    k = imgGroup->order[i];
	    imgGroup->order[i] = imgGroup->order[j];
	    imgGroup->order[j] = k;
	  }
	  /*
	  for (i = 0; i < imgGroup->numImg; i++) {
	    printf("%d ", imgGroup->order[i]);
	  }
	  printf("\n");
	  */
	}
      }

      imgGroup->aAppImg =
	(sAppImg*) malloc(sizeof(sAppImg) * imgGroup->numImg);

      // Quit if we didn't allocate the room
      if (imgGroup->aAppImg == NULL) {
	msprintf(tmpStr, "**ERR** Could not allocate room for %d images in group %d",
		 imgGroup->numImg, imgGroupNum);
	insertConsoleString(tmpStr);
	imgGroup->numImg = 0;
	
	appNum++;

#ifdef TRACE
  printf("%s() end\n", __func__);
#endif

	return RET_NORMAL;
      }

      // Got the space - fill in default values
      for (imgNum = 0; imgNum < imgGroup->numImg; imgNum++) {
	imgGroup->aAppImg[imgNum].filename = NULL;
	imgGroup->aAppImg[imgNum].pixmap   = NULL;
	imgGroup->aAppImg[imgNum].status   = kImgStatusInit;
	imgGroup->aAppImg[imgNum].parent   = imgGroup;
      }

      // Get the image file names
      curImg = 0;

      // First look for the standard file name
      if (checkImg(app->dirName, fileAppImgA)) {
	insertConsoleString("            ");
	appendConsoleString(fileAppImgA);
	imgGroup->aAppImg[curImg].filename =
	  (char*) malloc(sizeof(char) * (strlen(fileAppImgA) + 1));
	if (!imgGroup->aAppImg[curImg].filename) {
	  msprintf(tmpStr, "**ERR** Could not allocate room for image filename %s", fileAppImgA);
	  insertConsoleString(tmpStr);
	} else {
	  strcpy(imgGroup->aAppImg[curImg].filename, fileAppImgA);
	  curImg++;
	}
      }

      // Now check the new image file names
      for (imgNum = 0; imgNum < 26; imgNum++) {
	fileAppImgB[4] = 'a' + imgNum;
	if (checkImg(app->dirName, fileAppImgB)) {
	  insertConsoleString("            ");
	  appendConsoleString(fileAppImgB);
	  imgGroup->aAppImg[curImg].filename =
	    (char*) malloc(sizeof(char) * (strlen(fileAppImgB) + 1));
	  if (!imgGroup->aAppImg[curImg].filename) {
	    msprintf(tmpStr, "**ERR** Could not allocate room for image filename %s", fileAppImgB);
	    insertConsoleString(tmpStr);
	  } else {
	    strcpy(imgGroup->aAppImg[curImg].filename, fileAppImgB);
	    curImg++;
	  }
	}
      }

      if (curImg != imgGroup->numImg) {
	insertConsoleString("**ERR** Image count verification error.");
      }
    }
  }

  appNum++;

#ifdef TRACE
  printf("%s() end\n", __func__);
#endif

  return RET_NORMAL;
}

//==================================================================================================
// getAppImages()
//
// Reads the two images associated with an app.
//
// Inputs:
//   appNum - the index of the app to get images
// Outputs: none
// Globals modified:
//   aAppInfo[] - the array of structures that contain app info
//
//==================================================================================================

void getAppImages(int appNum)
{
  int imgGroupNum;
  sApp *app;
  sAppImgGroup *imgGroup;
  sAppImg *img;

  if (!aAppInfo || (appNum >= numApps)) return;
  if (!aAppInfo[appNum].flags & kFlagAppValid) return;

  app = &aAppInfo[appNum];

  // If the user is holding down the pad to scroll then don't load images yet.
  if (JOY_BUTTONS) return;
  
  // Loop through all the image groups
  for (imgGroupNum = 0; imgGroupNum < app->numImgGroups; imgGroupNum++) {

    // Get the group as a pointer
    imgGroup = &app->aImgGroups[imgGroupNum];

    if (imgGroup->numImg > 0) {
      // Get a pointer to the current image
      img = &imgGroup->aAppImg[imgGroup->curImg];

      // If we haven't already loaded the current image then start doing so now
      if (img->status == kImgStatusInit) {
	g_img = img;
	img->pixmap = (mmlDisplayPixmap *) checkAndReadImg(app->dirName, img->filename);
	if (img->pixmap) {
	  img->w = g_imgW;  // grab the dimensions from the globals
	  img->h = g_imgH;

	  // JPGs are drawn immediately
	  if (img->status == kImgStatusDrawn) {
	    redraw |= kRedrawAppStuff;
	  }
	}

	return;
      }
    }
  }

/*
  //
  // Load the rest of the images in the background
  //

  // Loop through all the image groups
  for (imgGroupNum = 0; imgGroupNum < app->numImgGroups; imgGroupNum++) {

    // Get the group as a pointer
    imgGroup = &app->aImgGroups[imgGroupNum];

    // Loop through all the images in the group
    first = TRUE;
    for (imgNum = imgGroup->curImg;
	 (imgNum == imgGroup->curImg) && !first;
	 imgNum = (imgNum >= imgGroup->numImg) ? 0 : (imgNum+1)) {

      // Get the image as a pointer
      img = &imgGroup[imgNum];
      
      first = FALSE;
    }
  }
*/
}
		  
//==================================================================================================
// getUserInput()
//
// Deals with the controller.
//
// Inputs:  none
// Outputs: none
// Globals modified:
//   curApp    - the index of the currently highlighted app
//   topApp    - the index of the app at the top of the menu
//   redraw    - use input causes things to get redrawn
//   topDesc   - the line the description at which to start drawing
//   instrStr  - a pointer to the string to drawn beneath the title
//   fieldsAccum - total number of fields that have passed - used for holding down buttons
//
//==================================================================================================

void getUserInput()
{
  static long prevButtons = 0;
  long buttons;
  long buttonsChanged;
  int load;
  char *str;
  int val;
  static int prevControlType = kControlJoy1;
  int nextApp;
  int imgGroupNum;
  sAppImgGroup *imgGroup;
  int debug = 0;

#ifdef TRACE
  printf("%s() begin\n", __func__);
#endif
  
  buttons = JOY_BUTTONS;
  buttonsChanged = buttons ^ prevButtons;

  // Change some text depending on what controller the user is using
  if (Buttons(_Controller[0])) controlType = kControlRemote;
  else if (Buttons(_Controller[1])) controlType = kControlJoy1;
  else if (Buttons(_Controller[2])) controlType = kControlJoy2;

  if (controlType != prevControlType) redraw |= kRedrawTitle;
  prevControlType = controlType;
  
  // Start the app
  if ((buttons & JOY_A) && ((buttonsChanged & JOY_A) || (fieldsAccum >= BUTTON_DELAY))) {
    if (aAppInfo[curApp].flags & kFlagAppValid) {
      if (wavSelect) {
	PlayWAV(wavSelect, -1, &Pan, 0x40000000, 0x40000000);
      }
      load = showIntroImage(curApp);
      if (load) {
	loadApp(curApp);
      }
#ifdef TRACE
      printf("%s() - Select app %d\n", __func__, curApp);
#endif
    }
  }

  // Switch to/from console
  if ((buttons & JOY_NUON) && (buttonsChanged & JOY_NUON)) {
    if (wavSelect) PlayWAV(wavSelect, -1, &Pan, 0x40000000, 0x40000000);
    switch (state) {
      case kStateConsoleAndStuff:  nextState = kStateMenuAndStuff;     break;
      case kStateConsoleAndMenu:   nextState = kStateConsoleAndStuff;  break;
      case kStateMenuAndStuff:     nextState = kStateConsoleAndMenu;   break;
    }
    redraw |= kRedrawAll;
#ifdef TRACE
    printf("%s() - State change (%d)\n", __func__, state);
#endif
  }
  
  //
  // Navigate the menu
  //

  // Previous app
  if (((menuOrientation == kMenuVertical) &&
       ((buttons & JOY_UP) && ((buttonsChanged & JOY_UP) || (fieldsAccum >= BUTTON_DELAY)))) ||
      ((menuOrientation == kMenuHorizontal) &&
       ((buttons & JOY_LEFT) && ((buttonsChanged & JOY_LEFT) || (fieldsAccum >= BUTTON_DELAY))))) {

    if ((curApp > 0) || menuWrap) {
      for (nextApp = curApp - 1, nextApp += (nextApp < 0) ? numApps : 0;
	   (nextApp != curApp) && (aAppInfo[nextApp].flags & kFlagSectionHeader);
	   nextApp--, nextApp += (nextApp < 0) ? numApps : 0) {
	// Quit if we wrapped when we weren't supposed to
	if ((nextApp > curApp) && !menuWrap) {
	  nextApp = curApp;
	  break;
	}
      }
      
      if ((nextApp >= 0) && (nextApp != curApp)) {
	redrawImages(curApp, nextApp);
	curApp = nextApp;
	while (curApp < topApp) {
	  topApp--;
	}
	if (curApp > botApp) {
	  topApp = curApp;
	}
	if ((curApp == topApp) && (curApp > 0) && (aAppInfo[curApp-1].flags & kFlagSectionHeader)) {
	  topApp--;
	}
	topDesc = 0;
	redraw |= kRedrawAppStuff;
	redraw |= kRedrawMenu;
	if (wavMove) {
	  PlayWAV(wavMove, -1, &Pan, 0x40000000, 0x40000000);
	}
	// Rotate the image if it is OnHighlight
	for (imgGroupNum = 0; imgGroupNum < aAppInfo[curApp].numImgGroups; imgGroupNum++) {
	  imgGroup = &aAppInfo[curApp].aImgGroups[imgGroupNum];
	  if ((imgGroup->numImg > 1) &&
	      (aRotateImg[kRotateImg1 + imgGroupNum] == kRotateImgOnHighlight)) {
	    imgGroup->curImg++;
	    if (debug) printf("%s() - Image %d rotate on highlight\n", __func__, imgGroupNum);
	    // Check if we wrapped
	    if (imgGroup->curImg >= imgGroup->numImg) {
	      if (debug) printf("%s() - Image %d wrapped (%d images)\n", __func__, imgGroupNum,
				imgGroup->numImg);
	      imgGroup->curImg = 0;
	    }
	  }
	}

#ifdef TRACE
	printf("%s() - PREV APP\n", __func__);
#endif
      }
    }
  }
  
  // Next app
  if (((menuOrientation == kMenuVertical) &&
       ((buttons & JOY_DOWN) && ((buttonsChanged & JOY_DOWN) || (fieldsAccum >= BUTTON_DELAY)))) ||
      ((menuOrientation == kMenuHorizontal) &&
       ((buttons & JOY_RIGHT) && ((buttonsChanged & JOY_RIGHT) || (fieldsAccum >= BUTTON_DELAY))))){

    if ((curApp < (numApps-1)) || menuWrap) {
      for (nextApp = curApp + 1, nextApp -= (nextApp >= numApps) ? numApps : 0;
	   (nextApp != curApp) && (aAppInfo[nextApp].flags & kFlagSectionHeader);
	   nextApp++, nextApp -= (nextApp >= numApps) ? numApps : 0) {
	// Quit if we wrapped when we weren't supposed to
	if ((nextApp < curApp) && !menuWrap) {
	  nextApp = curApp;
	  break;
	}
      }

      if ((nextApp < numApps) && (nextApp != curApp)) {
	if (nextApp > botApp) {
	  topApp += nextApp - curApp;
	}
	if (nextApp < topApp) {
	  topApp = nextApp;
	}
	redrawImages(curApp, nextApp);
	curApp = nextApp;
	topDesc = 0;
	redraw |= kRedrawAppStuff;
	redraw |= kRedrawMenu;
	if (wavMove) {
	  PlayWAV(wavMove, -1, &Pan, 0x40000000, 0x40000000);
	}
	// Rotate the image if it is OnHighlight
	for (imgGroupNum = 0; imgGroupNum < aAppInfo[curApp].numImgGroups; imgGroupNum++) {
	  imgGroup = &aAppInfo[curApp].aImgGroups[imgGroupNum];
	  if ((imgGroup->numImg > 1) &&
	      (aRotateImg[kRotateImg1 + imgGroupNum] == kRotateImgOnHighlight)) {
	    imgGroup->curImg++;
	    printf("%s() - Image %d rotate on highlight\n", __func__, imgGroupNum);
	    // Check if we wrapped
	    if (imgGroup->curImg >= imgGroup->numImg) {
	      printf("%s() - Image %d wrapped (%d images)\n", __func__, imgGroupNum,
		     imgGroup->numImg);
	      imgGroup->curImg = 0;
	    }
	  }
	}

#ifdef TRACE
	printf("%s() - NEXT APP\n", __func__);
#endif
      }
    }
  }

  // Go to previous section
  if (((menuOrientation == kMenuVertical) &&
       ((buttons & JOY_LEFT) && ((buttonsChanged & JOY_LEFT) || (fieldsAccum >= BUTTON_DELAY)))) ||
      ((menuOrientation == kMenuHorizontal) &&
       ((buttons & JOY_UP) && ((buttonsChanged & JOY_UP) || (fieldsAccum >= BUTTON_DELAY))))) {
    
    for (nextApp = curApp - 2, nextApp += (nextApp < 0) ? numApps : 0;
	 nextApp != curApp;
	 nextApp--, nextApp += (nextApp < 0) ? numApps : 0) {

      // Wrap around if necessary
      if (nextApp < 0) {
	if (menuWrap) {
	  nextApp += numApps;
	} else {
	  break;
	}
      }

      // Look for the first section header
      if (aAppInfo[nextApp].flags & kFlagSectionHeader) {

	// Found a header.  Now look for the first app after the header.
	for (; nextApp < numApps; nextApp++) {
	  if (!(aAppInfo[nextApp].flags & kFlagSectionHeader)) {
	    break;
	  }
	}
	break;
      }
    }

    // Check if we found the header
    if (nextApp < 0) {
      // Didn't find the header.  Choose the first app in the list.
      for (nextApp = 0; nextApp < (numApps - 1); nextApp++) {
	if (!(aAppInfo[nextApp].flags & kFlagSectionHeader)) {
	  break;
	}
      }
    }

    if (nextApp != curApp) {
      redrawImages(curApp, nextApp);
      curApp = nextApp;
      topApp = nextApp;
      if ((curApp > 0) && (aAppInfo[curApp-1].flags & kFlagSectionHeader)) {
	topApp--;
      }
      topDesc = 0;
      redraw |= kRedrawAppStuff;
      redraw |= kRedrawMenu;
      if (wavMove) {
	PlayWAV(wavMove, -1, &Pan, 0x40000000, 0x40000000);
      }
#ifdef TRACE
      printf("%s() - PREV SECTION\n", __func__);
#endif
    }
  }
  
  // Go to next section
  if (((menuOrientation == kMenuVertical) &&
       ((buttons & JOY_RIGHT) && ((buttonsChanged & JOY_RIGHT) || (fieldsAccum >= BUTTON_DELAY))))||
      ((menuOrientation == kMenuHorizontal) &&
       ((buttons & JOY_DOWN) && ((buttonsChanged & JOY_DOWN) || (fieldsAccum >= BUTTON_DELAY))))) {
    
    for (nextApp = curApp + 1;
	 nextApp != curApp;
	 nextApp++, nextApp -= (nextApp >= numApps) ? numApps : 0) {

      // Quit if we wrapped when we weren't supposed to
      if ((nextApp < curApp) && !menuWrap) {
	break;
      }

      // Look for the first section header
      if (aAppInfo[nextApp].flags & kFlagSectionHeader) {

	// Found a header.  Now look for the first app after the header.
	for (; nextApp < numApps; nextApp++) {
	  if (!(aAppInfo[nextApp].flags & kFlagSectionHeader)) {
	    break;
	  }
	}
	break;
      }
    }

    // Check if we found the header
    if (nextApp >= numApps) {
      // Didn't find the header.  Choose the last app in the list.
      for (nextApp = numApps - 1; nextApp > curApp; nextApp--) {
	if (!(aAppInfo[nextApp].flags & kFlagSectionHeader)) {
	  break;
	}
      }
    }
    
    if (nextApp != curApp) {
      redrawImages(curApp, nextApp);
      curApp = nextApp;
      topApp = nextApp;
      if ((curApp > 0) && (aAppInfo[curApp-1].flags & kFlagSectionHeader)) {
	topApp--;
      }
      topDesc = 0;
      redraw |= kRedrawAppStuff;
      redraw |= kRedrawMenu;
      if (wavMove) {
	PlayWAV(wavMove, -1, &Pan, 0x40000000, 0x40000000);
      }
#ifdef TRACE
      printf("%s() - NEXT SECTION\n", __func__);
#endif
    }
  }

  // Scroll the description
#ifdef NO_SCROLL_DESC
  if (0) {
#else
  if ((buttons & JOY_C_DOWN) && ((buttonsChanged & JOY_C_DOWN) || (fieldsAccum >= BUTTON_DELAY))) {
#endif
    redraw |= kRedrawAppStuff;
    topDesc++;
    if (wavMove) {
      PlayWAV(wavMove, -1, &Pan, 0x40000000, 0x40000000);
    }
#ifdef TRACE
    printf("%s() - SCROLL DESC DOWN\n", __func__);
#endif
  }
  
#ifdef NO_SCROLL_DESC
  if (0) {
#else
  if ((buttons & JOY_C_UP) && ((buttonsChanged & JOY_C_UP) || (fieldsAccum >= BUTTON_DELAY))) {
#endif
    redraw |= kRedrawAppStuff;
    if (topDesc > 0) topDesc--;
    if (wavMove) {
      PlayWAV(wavMove, -1, &Pan, 0x40000000, 0x40000000);
    }
#ifdef TRACE
    printf("%s() - SCROLL DESC UP\n", __func__);
#endif
  }

  // Toggle fast mode
  if ((buttons & JOY_C_RIGHT) && (buttonsChanged & JOY_C_RIGHT)) {
    
    // Read the configa register from a stopped MPE to get the chip revision
    if (ariesVer != 4) {
      fastMode = -1;
      insertConsoleString("**WARN** Not Aries3 - cannot enter fast mode");
    } else {
      if (wavSelect) PlayWAV(wavSelect, -1, &Pan, 0x40000000, 0x40000000);
      redraw |= kRedrawTitle;
      fastMode = (fastMode == 1) ? 0 : 1;
      if (fastMode == 1) {
	val = 0x200;
	insertConsoleString("Fast mode enabled");
      } else {
	val = 0x100;
	insertConsoleString("Fast mode disabled");
      }
      _MPEWriteRegister(0, (void*)0x20500000, val);  // enable fast mode
      _MPEWriteRegister(1, (void*)0x20500000, val);
      _MPEWriteRegister(2, (void*)0x20500000, val);
	
      __asm__ volatile( 
		       "	nop         \n"
		       "	nop         \n"
		       "	st_s	%0,0x20500000\n"
		       "	nop         \n"
		       :
		       : "r"(val)
		       );
    }
  }
  
  // Draw instructions
  str = instrStr;
  if ((JOY_BUTTONS & JOY_L) && (JOY_BUTTONS & JOY_R)) {
    instrStr = strAbout;
#ifdef NO_INSTR2
  } else if (0) {
#else
  } else if (JOY_BUTTONS & JOY_B) {
#endif
    instrStr = (ariesVer == 4) ? aStrings[kStrInstr2Aries3] : aStrings[kStrInstr2Normal];
  } else {
    instrStr = (controlType == kControlRemote)
      ? aStrings[kStrInstr1Remote] : aStrings[kStrInstr1Normal];
  }

  if (str != instrStr) redraw |= kRedrawTitle;

  
  if (fieldsAccum >= BUTTON_DELAY) {
    fieldsAccum = 0;
  }
  
  if (buttonsChanged) {
    fieldsAccum = 0;
  }

  prevButtons = buttons;
  
#ifdef TRACE
  printf("%s() end\n", __func__);
#endif
}

//==================================================================================================
// showIntroImage()
//
// Displays the intro image for the given app.  If the user presses START then the app will be
// loaded.  Pressing B will cancel back to the menu.
//
// Inputs:
//   appNum - the index of the app to load
// Outputs:
//   int - returns TRUE or FALSE to instruct the calling function whether the user confirmed the
//         load (TRUE) or canceled it (FALSE)
// Globals modified: none
//
//==================================================================================================

int showIntroImage(int appNum)
{
  sApp *app;
  int ret = -1;
  int prevState;
#ifndef OLD_TGA
  long sec, prevSec;
#endif
  
  if (!aAppInfo) {
    return(FALSE);
  }

#ifdef TRACE
  printf("%s() begin\n", __func__);
#endif

  app = &aAppInfo[appNum];

  // First free all non-essential video memory
  freeAppImages();

  // Allocate this every time since it's unlikely we'll get here more than once
  if (app->imgIntro.status == kImgStatusInit) {
    g_img = &app->imgIntro;
    app->imgIntro.pixmap = (mmlDisplayPixmap *) checkAndReadImg(app->dirName, fileImgIntro);
    if (app->imgIntro.pixmap) {
      app->imgIntro.w = g_imgW;  // grab the dimensions from the globals
      app->imgIntro.h = g_imgH;
    }
  }

  if (error) {
    redraw |= kRedrawAll;
    error = FALSE;
#ifdef TRACE
  printf("%s() end early 1\n", __func__);
#endif
  
    return FALSE;
  }

#ifndef OLD_TGA
  // Wait for the helper MPE to be done rendering the image
  _TimeElapsed(&prevSec, NULL);
  while (app->imgIntro.status == kImgStatusSubmitted) {
    // Time out after a bit
    _TimeElapsed(&sec, NULL);
    if ((sec - prevSec) > 5) {
      insertConsoleString("**ERR** Intro image render timeout");
      break;
    }
  }
#endif

  // Intro screen must be the same size as the screen
  if (app->imgIntro.pixmap && ((app->imgIntro.w != SCREEN_W) || (app->imgIntro.h != SCREEN_H))) {
    msprintf(tmpStr, "**ERR** Intro image must be %dx%d", SCREEN_W, SCREEN_H);
    insertConsoleString(tmpStr);
    msprintf(tmpStr, "        Image is %dx%d", app->imgIntro.w, app->imgIntro.h);
    insertConsoleString(tmpStr);
    
    mmlReleasePixmaps((mmlPixmap *) app->imgIntro.pixmap, &sysRes, 1);
    app->imgIntro.pixmap = NULL;
    app->imgIntro.status = kImgStatusInit;
    redraw |= kRedrawAll;

#ifdef TRACE
  printf("%s() end early 2\n", __func__);
#endif
  
    return FALSE;
  }

  // No image so just load the app immediately
  if (!app->imgIntro.pixmap) {
#ifdef TRACE
  printf("%s() end early 3\n", __func__);
#endif
  
    return(TRUE);
  }

  prevState = state;
  nextState = kStateIntro;
  redraw |= kRedrawAll;
      
  // Wait for user to release all buttons
  while (JOY_BUTTONS) {
    updateScreen();
  }

  // Wait for user to press B or START
  while (ret == -1) {
    if (JOY_BUTTONS & JOY_B) {
      nextState = prevState;
      redraw |= kRedrawAll;
      ret = FALSE;
    }
    if (JOY_BUTTONS & JOY_START) {
      if (wavSelect) PlayWAV(wavSelect, -1, &Pan, 0x40000000, 0x40000000);
      redraw |= kRedrawAll;
      ret = TRUE;
    }
    updateScreen();
  }

  // Release the memory
  mmlReleasePixmaps((mmlPixmap *) app->imgIntro.pixmap, &sysRes, 1);
  app->imgIntro.pixmap = NULL;
  app->imgIntro.status = kImgStatusInit;

#ifdef TRACE
  printf("%s() end\n", __func__);
#endif
  
  return ret;
}

//==================================================================================================
// loadApp()
//
// Loads the executable associated with an app.
//
// Inputs:
//   appNum - the index of the app to load
// Outputs: none
// Globals modified: none
//
//==================================================================================================

void loadApp(int appNum)
{
  myFileType *fp = NULL;
  int fileSize;
  int offset;

  insertConsoleString("Loading coff");
  
  // Clear the screen
  // DISABLED: why did I put this here?
  //  drawBackground(NULL, FALSE);
  //  VideoSync();

  if (appNum < numApps) {

    // Construct the full path to the executable
    strcpy(tmpStr, HOST_PATH);
    strcat(tmpStr, &(aAppInfo[appNum].dirName[1]));
    strcat(tmpStr, fileAppExe1);

    // First check for 'cd_app.cof'
    appendConsoleString(" from data file");
      
    offset = (int) getDataFilePtr(tmpStr, &fileSize);
    if (offset) {
      myfseek(fpDataFile, offset, SEEK_SET);
      fp = fpDataFile;
      fileSize += 4;  // add 4 bytes for checksum
    }
      
    if (fp) {
      loadAppFromFP(fp, fileSize, 0);
      if (fp != fpDataFile) myfclose(fp);
    } else {

      strcpy(tmpStr, HOST_PATH);
      strcat(tmpStr, &(aAppInfo[appNum].dirName[1]));      
      strcat(tmpStr, fileAppExe2);

      // Then check for NUON.CD
      offset = (int) getDataFilePtr(tmpStr, &fileSize);
      if (offset) {
	myfseek(fpDataFile, offset, SEEK_SET);
	fp = fpDataFile;
	fileSize += 4;  // add 4 bytes for checksum
      }
      
      if (fp) {
	loadAppFromFP(fp, fileSize, 0x14F4C); // 0x570);  // don't know where I got 0x570 from
	if (fp != fpDataFile) myfclose(fp);
      } else {
	insertConsoleString("**ERR** Could not open file ");
	appendConsoleString(tmpStr);
      }
    }
  }  // (appnum < numApps)

  // Function should not return
  insertConsoleString("**ERR** loadApp failed!");
}

//==================================================================================================
// loadAppFromFP()
//
// Actually loads the reads the file from the disc and runs it.  It first reads the executable
// into main memory (video memory) so as to not step on the boot loader code.  Then it loads some
// assembly onto MPE2 that will do the loading.
//
// It starts MPE2 running by passing it a pointer to the executable.  Then it jumps into a small
// routine in scratch space that invalidates all the caches and waits for a comm bus packet.
//
// MPE2 does its thing, transfering the coff file into system ram.  When MPE2 completes it sends
// MPE3 a comm bus packet and halts.  MPE3 detects this and jumps to the start of the new program.
//
// Inputs:
//   fp - file pointer from which to load the executable
//   fileSize - the size of the executable
//   exeHeaderSize - the size of the executable's header
// Outputs: none
// Globals modified: none
//
//==================================================================================================
void loadAppFromFP(myFileType *fp, int fileSize, int exeHeaderSize)
{
  int i;
  void *ptr = NULL, *coffDataBuf = NULL;
  long *longPtr = NULL;
  int bytesRead;
  long checksum;
  int prevState = nextState;
  int retries;
  int appLoc;

  // Do this before deallocating the background so that the background image will show on the
  // loading screen.
  nextState = kStateLoading;
  redraw |= kRedrawAll;
  updateScreen();
  updateScreen();

  //
  // Free all video ram.
  // No need to free system ram since the new app will reinitialize the heap.
  //
  //  for (i = 0; i < kStrMax; i++) if (aStrings[i]) free(aStrings[i]);
  //  if (screenshotBuffer) free(screenshotBuffer);
  freeAppImages();

  if (BG_IMG_PTR) {
    mmlReleasePixmaps((mmlPixmap *) BG_IMG_PTR, &sysRes, 1);
    BG_IMG_PTR = NULL;
  }
  //  if (wavMove) {
  //    wavMove = NULL;
  //  }
  //  if (wavSelect) {
  //    wavSelect = NULL;
  //  }
  //  if (bufDataFile) free(bufDataFile);

  // Free the audio MPE after all sounds are done
#ifndef NO_AUDIO
  while (PCMGetUsedVoices()) updateScreen();
  AUDIOExit();
  //  _AudioReset();
#endif
  
  _MediaShutdownMPE();
  _MediaInitMPE();
  
#ifndef OLD_TGA
  _IntSetVector(kIntrCommRecv, cbMagicNum);
#endif

  if (fileSize != -1) {
    // Allocate room in main ram for the executable
    coffDataBuf = _MemAlloc(fileSize, 4, kMemSDRAM);

    if (!coffDataBuf) {
      msprintf(tmpStr, "**ERR** Could not allocate %d bytes", fileSize);
      insertConsoleString(tmpStr);
    } else {

      appLoc = myftell(fp);
		       
      retries = REOPEN_ATTEMPTS;
      while (retries > 0) {
	checksum = 0;
      
	// Read the file off the disc
	myfseek(fp, appLoc, SEEK_SET);
	bytesRead = myfread(coffDataBuf, fileSize, fp);

	// Do checksum verification
	longPtr = (long *) coffDataBuf;
	for (i = 0; i < fileSize; i += 4) {
	  checksum += *longPtr;
	  longPtr++;
	}

	retries--;

	if (checksum) {
	  msprintf(tmpStr, "Try #%d EXE checksum non-zero: %08lx",
		   REOPEN_ATTEMPTS - retries, checksum);
	  insertConsoleString(tmpStr);
	} else {
	  // Force the loop to exit if the checksum passed
	  retries = 0;
	}
      }
      
      if (checksum) {
	msprintf(tmpStr, "**ERR** Checksum failed %d times", REOPEN_ATTEMPTS);
	insertConsoleString(tmpStr);
	
      } else {

	ptr = coffDataBuf + exeHeaderSize;  // skip over the header if there is one
	ptr += 4;                           // skip the checksum

	if (cdError) {
	  insertConsoleString("**ERR** iso_read failed!");
	} else if (bytesRead != fileSize) {
	  msprintf(tmpStr, "**ERR** bytesRead != fileSize (%d != %d)", bytesRead, fileSize);
	  insertConsoleString(tmpStr);
	} else {
	  myfclose(fp);
	  if (cdError) {
	    insertConsoleString("**ERR** iso_close failed!");
	  } else {

	    // Shut down the CD library
	    CDExit();

	    // Load the wait routine into scratch space.  We'll jump there while MPE2 is loading
	    // the coff file.
	    _DMALinear(DMA_READ_BIT | (((WaitEnd-Wait+3)>>2)<<16), Wait, localScratch);

	    // Load the coff loader into MPE2 (stop the TGA helper first)
	    _MPEStop(MPE2);
	    _MPELoad(MPE2, (void *)0x20300000, coff_c_start, (long) coff_c_size);
	    _MPERun (MPE2, (void *)0x20300000);
	
	    mmlReleasePixmaps((mmlPixmap *) &frameBuffer[0], &sysRes, 1);
	    mmlReleasePixmaps((mmlPixmap *) &frameBuffer[1], &sysRes, 1);

	    _MemFree(coffDataBuf);

	    // We should always have all 8MB of main ram free at this point
	    if (getFreeMem(kMemSDRAM) != 0x800000) {
#ifdef DSI_DEBUG
	      printf("getFreeMem() did not return 8MB: %0ld bytes\n", getFreeMem(kMemSDRAM));
#endif
	    }

	    _DCacheFlush();

	    // Send a pointer to the data to MPE2
	    _CommSendDirect(3, (int) ((void *) ptr), DC, DC, MPE2, DC);

	    // Jump to our routine in scratch space
	    {
	      register int ptr = (int) localScratch;
	      __asm__ volatile( 
			       "	nop         \n"
			       "	jsr	(%0)\n"
			       "	nop         \n"
			       "	nop         \n"
			       :
			       : "r"(ptr)
			       );
	    }
	  }  // (cdError)
	}  // (bytesRead != fileSize)
      }  // (checksum)
    }  // !coffDataBuf
  }  // fileSize != -1

  //
  // Shouldn't get here unless error
  //
  
#ifndef NO_AUDIO
  AUDIOInit();
#endif
  
#ifndef OLD_TGA
  // Reinstall comm bus handler
  cbMagicNum = _IntSetVector(kIntrCommRecv, TGADoneCommHandler);

  _MPEStop(MPE2);
  _MPELoad(MPE2, (void *)0x20300000, tgac_start, (long) tgac_size);
  _MPELoad(MPE2, (void *)0x20100000, tgad_start, (long) tgad_size);
  _MPERun (MPE2, (void *)0x20300000);
#endif

  nextState = prevState;
  redraw |= kRedrawAll;
  updateScreen();
}

//==================================================================================================
// checkAndReadImg()
//
// This is a special case routine that calls checkAndReadFile for images.  It checks for the
// filename in both .tga and .jpg formats.
//
// Inputs:
//   dir - the directory for the image file
//   filename - the name of the image file
// Outputs:
//   void* - a pointer to the image data
// Globals modified: none
//
//==================================================================================================

void* checkAndReadImg(char *dir, char *filename)
{
  char buf[80];
  void *outImg;

  // Generate the full path to the file
  if (dir) strcpy(buf, dir);
  if (filename) strcat(buf, filename);

  // Looking for a TGA file
  filename[strlen(filename)-3] = 't';
  filename[strlen(filename)-2] = 'g';
  filename[strlen(filename)-1] = 'a';
  outImg = checkAndReadFile(dir, filename, kFileTypeTGA);
  if (outImg) return(outImg);

  // Didn't find it
  // Now looking for a JPG file
  filename[strlen(filename)-3] = 'j';
  filename[strlen(filename)-2] = 'p';
  filename[strlen(filename)-1] = 'g';
  outImg = checkAndReadFile(dir, filename, kFileTypeJPG);
  return(outImg);
}

//==================================================================================================
// checkImg()
//
// Checks to see if an image file exists in the data file.
//
// Inputs:
//   filename - full path to the image to find
// Outputs: none
// Globals modified: none
//
//==================================================================================================

int checkImg(char *dir, char *filename)
{
  char buf[80];

  // Construct the full filename path
  strcpy(buf, HOST_PATH);
  if (dir) strcat(buf, &dir[1]);
  if (filename) strcat(buf, filename);

  // Looking for a TGA file
  buf[strlen(buf)-3] = 't';
  buf[strlen(buf)-2] = 'g';
  buf[strlen(buf)-1] = 'a';
  if (getDataFilePtr(buf, NULL)) {
    return TRUE;
  }

  // Didn't find it
  // Now looking for a JPG file
  buf[strlen(buf)-3] = 'j';
  buf[strlen(buf)-2] = 'p';
  buf[strlen(buf)-1] = 'g';
  if (getDataFilePtr(buf, NULL)) {
    return TRUE;
  }

  return FALSE;
}

//==================================================================================================
// checkAndReadFile()
//
// Checks for the existence of a given file.  If found, returns the file in the appropriate format.
// For executables we only check for existence.  For images, we convert the image into a
// mmlDisplayPixmap for use by the program.  For text or wavs we just return the buffer.
//
// Inputs:
//   dir      - the directory name where the file is located
//   filename - the name of the file to read
//   fileType - the type of file to be read (see eFileType)
//
// Outputs:
//   void* - a pointer to the file data
//
// Globals modified:
//   fileSize - size of the file
//   imgW - the width of an image file
//   imgH - the height of an image file
//
//==================================================================================================

void* checkAndReadFile(char *dir, char *filename, eFileType fileType)
{
  char *fileBuffer = NULL;
  mmlDisplayPixmap *imgPixmap = NULL;
  long dmaFlags;
  char buf[80];
  int pixmapW, pixmapH;
  long *jpgImg = NULL;
  mmlAppPixmap appPixmap;
  m2dRect rect;
  m2dPoint point;
  mmlStatus status;

  // Construct the full filename path
  strcpy(buf, HOST_PATH);
  if (dir) strcat(buf, &dir[1]);
  if (filename) strcat(buf, filename);

  insertConsoleString("      ");
  appendConsoleString(filename);
  appendConsoleString(" - ");
  
  fileBuffer = getDataFilePtr(buf, &curFileSize);
  if (fileBuffer) appendConsoleString("found");
  else            appendConsoleString("not found");

  // Now that we have the data, what do we do with it?
  if (fileBuffer) {
    switch (fileType) {
      case kFileTypeExe:
      case kFileTypeText:
      case kFileTypeWAV:
      case kFileTypeTTF:
	// do nothing but return the buffer pointer
	break;
	    
      case kFileTypeTGA:
	// Decode the TGA

	// First get the image size
	//	GetTGASize(fileBuffer, &g_imgW, &g_imgH);
	g_imgW = (unsigned char) fileBuffer[12] | ((unsigned char) fileBuffer[13] << 8);
	g_imgH = (unsigned char) fileBuffer[14] | ((unsigned char) fileBuffer[15] << 8);

	// Allocate the pixmap that will point to the image
	imgPixmap = (mmlDisplayPixmap *) malloc(sizeof(mmlDisplayPixmap));
	if (!imgPixmap) {
	  // Oops, the alloc failed.
	  insertConsoleString("**ERR** Could not allocate pixmap for ");
	  appendConsoleString(buf);
	  fileBuffer = NULL;
	  error = TRUE;
	} else {
	  // Got the pixmap.

	  // Round the dimensions up so width is a multiple of 8 and height a multiple of 16
	  pixmapW = (g_imgW +  7) & ~0x7;
	  pixmapH = (g_imgH + 15) & ~0xF;

	  // Initialize space for the pixmap
	  status = mmlInitDisplayPixmaps(imgPixmap, &sysRes, pixmapW, pixmapH, e888Alpha, 1, NULL);
	  
	  // Alloc failed once.  Try freeing all other images.
	  if (status != eOK) {

	    // Free all the app images
	    freeAppImages();
	    
	    // Try again
	    status = mmlInitDisplayPixmaps(imgPixmap, &sysRes, pixmapW, pixmapH, e888Alpha, 1,NULL);
	  }

	  // Alloc failed twice.  Barf.
	  if (status != eOK) {
	    insertConsoleString("**ERR** Didn't get pixmap for ");
	    appendConsoleString(buf);
	    fileBuffer = NULL;
	    free(imgPixmap);
	    error = TRUE;
	  } else {
	    // Got the image space we need, now decode the TGA into that memory.
	    dmaFlags = DMA_PIXEL_WRITE | ((pixmapW/8)<<16) | (4<<4) | (1<<11);

#ifdef OLD_TGA
	    if (!ReadTGA(fileBuffer, dmaFlags, imgPixmap->memP, 0, 0, NULL, 0)) {
	      // Error while decoding TGA
	      insertConsoleString("**ERR** Error decoding TGA ");
	      appendConsoleString(buf);
	      if (!bufDataFile) free(fileBuffer);
	      fileBuffer = NULL;
	      mmlReleasePixmaps((mmlPixmap *) imgPixmap, &sysRes, 1);
	      free(imgPixmap);
	      error = TRUE;
	    } else {
	      // Return a pointer to the pixmap
	      fileBuffer = (void *) imgPixmap;

	      if (g_img) {
		g_img->status = kImgStatusDrawn;
	      }
	    }
#else

#ifdef TRACE
	    printf("Decoding TGA %08lX %08lX %08lX %08X\n",
		   (long) fileBuffer, (long) imgPixmap->memP, dmaFlags, (int) g_img);
#endif

	    if (g_img) {
	      g_img->status = kImgStatusSubmitted;
	    }
	    _CommSendDirect((long) fileBuffer, (long) imgPixmap->memP, dmaFlags, (int)g_img,MPE2,0);
	    /*
	    {
	      long buffer[4];
	      int info;
	      _CommRecvInfo(&info, buffer);
	      printf("%08X %08X %08X %08X (%08X)\n", buffer[0], buffer[1], buffer[2], buffer[3], g_img);
	    }
	    */
	    fileBuffer = (void *) imgPixmap;
#endif
	  }
	}
	break;

      case kFileTypeJPG:
	// Decode the JPEG, returning a pointer to the image
	jpgImg = decompressJPEG(fileBuffer, curFileSize, &g_imgW, &g_imgH);

	if (!jpgImg) {
	  // Error during JPG decoding
	  insertConsoleString("**ERR** Could not decompress jpeg");
	  fileBuffer = NULL;
	  error = TRUE;
	} else {
	  // make sure jpeg is written to memory
	  _DCacheFlush();

	  // The JPG image is returned in raster order in main memory so assign it to an App pixmap
	  mmlInitAppPixmaps(&appPixmap, &sysRes, g_imgW, g_imgH, e888Alpha, 1, jpgImg);

	  // Now create an image pixmap to hold the final image
	  imgPixmap = (mmlDisplayPixmap *) malloc(sizeof(mmlDisplayPixmap));
	  if (!imgPixmap) {
	    // Didn't get the pixmap - bail
	    insertConsoleString("**ERR** Could not allocate pixmap for ");
	    appendConsoleString(buf);
	    free(jpgImg);
	    fileBuffer = NULL;
	    error = TRUE;
	  } else {
	    // Got the pixmap

	    // Round the image width up to a multiple of 8 and the height to a multiple of 16
	    pixmapW = (g_imgW +  7) & ~0x7;
	    pixmapH = (g_imgH + 15) & ~0xF;

	    // Allocate sdram space for the image
	    status = mmlInitDisplayPixmaps(imgPixmap, &sysRes, pixmapW, pixmapH, e888Alpha, 1,NULL);

	    // Alloc failed once.  Try freeing all other images.
	    if (status != eOK) {

	      // Free all the app images
	      freeAppImages();
	    
	      // Try again
	      status = mmlInitDisplayPixmaps(imgPixmap, &sysRes, pixmapW, pixmapH,e888Alpha,1,NULL);
	    }

	    // Alloc failed twice.  Barf.
	    if (status != eOK) {
	      insertConsoleString("**ERR** Didn't get pixmap for ");
	      appendConsoleString(buf);
	      free(jpgImg);
	      fileBuffer = NULL;
	      free(imgPixmap);
	      error = TRUE;	
	    } else {
	      // Got the memory for the pixmap.

	      // Now transfer the app pixmap (system ram) to display pixmap (main ram).
	      rect.leftTop.x = 0;
	      rect.leftTop.y = 0;
	      rect.rightBot.x = g_imgW - 1;
	      rect.rightBot.y = g_imgH - 1;
	      point.x = 0;
	      point.y = 0;
	      m2dCopyRect(&gc, &appPixmap, imgPixmap, &rect, point);  // do the copy

	      // Free the App pixmap memory
	      free(jpgImg);
	    
	      // Return a pointer to the pixmap
	      fileBuffer = (void *) imgPixmap;

	      if (g_img) {
		g_img->status = kImgStatusDrawn;
	      }
	    }
	  }
	}
	break;
	  
      default:
	insertConsoleString("**ERR** Bad default case");
	fileBuffer = NULL;
	break;
    } // switch (fileType)
      
  } // fileBuffer

  return fileBuffer;
}

//==================================================================================================
// getDataFilePtr()
//
// Searches for the given file within the data file.  Returns a pointer to the start of the file
// data if it exists.
//
// If the file is beyond the part that's been read then just return the byte offset.  The calling
// function will need to seek to and read the file itself.
//
// Inputs:
//   filename - the name of the file to get
//   filesize - gets filled in with the file's size
// Outputs:
//   void* - a pointer in the data file buffer where the file starts
// Globals modified:
//   none
//
//==================================================================================================

void* getDataFilePtr(char *filename, int *fileSize)
{
  char *bufPtr;
  int numChars;
  int offset;
  int i;
  
  if (fileSize) {
    *fileSize = 0;
  }

  // Get a pointer to the filename list in the data file
  bufPtr = &bufDataFile[DATA_FILE_HEADER_SIZE];

  // Loop through all the files
  for (i = 0; i < numDataFiles; i++) {

    // Does this filename match?
    if (!strcmp(filename, bufPtr)) {

      //
      // Found a match
      //
      
      // The file's offset and size are located after the filename
      numChars = 0;
      while (*bufPtr) { bufPtr++; numChars++; }             // skip file name
      do { bufPtr++; numChars++; } while (numChars & 0x3);  // round up to long boundary

      if (fileSize) {
	*fileSize = ((long*) bufPtr)[1];                    // fill in file size
      }
      offset = ((long*) bufPtr)[0];                         // get the offset to this file
      
      // If the file is within the header that we read then return a pointer to the
      // actual data.  Otherwise just return the byte offset.
      return (offset < headerSize) ? &bufDataFile[offset] : ((void*) offset);
    }

    numChars = 0;
    while (*bufPtr) { bufPtr++; numChars++; }             // skip file name
    do { bufPtr++; numChars++; } while (numChars & 0x3);  // round up to long boundary
    bufPtr += 8;                                          // skip offset and size
  }

  // Didn't find the file - return NULL
  return NULL;
}

//==================================================================================================
// getFileSize()
//
// Computes the size of a file.  It does this by seeking to the end of the file, getting the offset,
// and seeking back.
//
// Inputs:
//   fp - the file to compute the size
// Outputs:
//   int - returns the file size
// Globals modified: none
//
//==================================================================================================

int getFileSize(myFileType *fp)
{
  int fileSize;
  int fileloc;

  // Get the current file position
  fileloc = myftell(fp);
  if (cdError) {
    insertConsoleString("**ERR** iso_lseek failed! (1)");
    return(-1);
  }

  // Seek to the end
  myfseek(fp, 0, SEEK_END);

  // Get the position (file size)
  fileSize = myftell(fp);
  if (cdError) {
    insertConsoleString("**ERR** iso_lseek failed! (2)");
    return(-1);
  }

  // Return back from whence we came
  myfseek(fp, fileloc, SEEK_SET);
  if (cdError) {
    insertConsoleString("**ERR** iso_lseek failed! (3)");
    return(-1);
  }
  
  /*
  iso_fstat(cdContextPtr, fp, &fileStat, &cdError);
  if (cdError) {
    insertConsoleString("**ERR** iso_fstat failed!");
    return(-1);
  }
  fileSize = (unsigned long) fileStat.st_rdev | ((unsigned long) fileStat.st_gid << 16);
  */

  return fileSize;
}

//==================================================================================================
// myReadFile()
//
// Reads the given file into a buffer that gets allocated here.  Some checking is done to ensure
// the number of bytes read is the same as the file size.
//
// Inputs:
//   fp - pointer to the file to read
//   fileSize - gets filled in with the file size
// Outputs:
//   char* - a pointer to a newly allocated buffer where the file was read
// Globals modified: none
//
//==================================================================================================

char* myReadFile(myFileType *fp, int *fileSize)
{
  char *buf = NULL;
  int bytesRead;

  // First get file's size
  *fileSize = getFileSize(fp);
  if (*fileSize == -1) {
    insertConsoleString("**ERR** In myReadFile, getFileSize failed!");
    myfclose(fp);
    return(NULL);
  }

  // Allocate enough room for it plus two.  One of the extra characters is because we will try to
  // read one more than the filesize to see if the filesize was computed incorrectly.  The other is
  // to always terminate with a NULL character.
  buf = (char *) malloc(sizeof(char) * (*fileSize + 2));
  buf[*fileSize  ] = 0;
  buf[*fileSize+1] = 0;
  
  if (!buf) {
    insertConsoleString("**ERR** Could not allocate file buffer");
    myfclose(fp);
    if (cdError) insertConsoleString("**ERR** iso_close failed!");
    return(NULL);
  }

  // Try to read one byte more than the filesize.  It shouldn't do it, but if the filesize was
  // computed incorrectly it might.
  bytesRead = myfread(buf, *fileSize + 1, fp);
  
  if (cdError) {
    insertConsoleString("**ERR** iso_read failed! ");
    myfclose(fp);
    if (cdError) insertConsoleString("**ERR** iso_close failed!");
    free(buf);
    return(NULL);
  }
  myfclose(fp);
  if (cdError) insertConsoleString("**ERR** iso_close failed!");
  
  if (bytesRead != *fileSize) {
    insertConsoleString("**ERR** bytesRead != fileSize!");
    free(buf);
    return(NULL);
  }
   
  return(buf);
}

//==================================================================================================
// myfopen()
//
// Opens a file differently depending on whether it's on a dev system (using the file server) or
// real hardware (using the CD lib).
//
// Inputs:
//   filename - name of the file to open
// Outputs:
//   myFileType* - a pointer to the newly opened file (or NULL if it wasn't)
// Globals modified: none
//
//==================================================================================================

myFileType* myfopen(char *filename)
{
  myFileType *fp;
  
#ifdef USE_LOCAL_HOST
  fp = fopen(filename, "r");
#else
  
  fp = iso_open(cdContextPtr, filename, 0, 0, &cdError);
  if (cdError) {
    // cdError is set for file not found
    return(NULL);
  }
#endif

  return(fp);
}
  
//==================================================================================================
// mygets()
//
// Operates like fgets, but uses an internal buffer rather than a file pointer.  Copies at most
// maxDest characters from src to dest, stopping when a newline is found.  The dest is then null
// terminated.
//
// Inputs:
//   dest    - destination buffer where to place the data
//   maxDest - the size of the destination buffer
//   src     - pointer to the pointer to the source buffer - gets modified
//   maxSrc  - characters left in the source buffer - gets modified
// Outputs:
//   int    - returns the number of characters that were read
//   src    - pointer gets updated to just after the characters read
//   maxSrc - gets decremented as characters are read
// Globals modified: none
//
//==================================================================================================

int mygets(char *dest, int maxDest, char **src, int *maxSrc)
{
  int numChars = 0;
  char thisChar;
  int done = FALSE;

  while (!done && *maxSrc && (numChars < (maxDest - 1))) {

    thisChar = **src;
    *dest = thisChar;

    numChars++;
    (*maxSrc)--;
    (*src)++;

    if (thisChar == 0x0D) {
      done = TRUE;
      while ((**src == 0x0D) || (**src == 0x0A)) {
	(*maxSrc)--;
	(*src)++;
      }
    } else if (thisChar == 0x0A) {
      done = TRUE;
      while ((**src == 0x0D) || (**src == 0x0A)) {
	(*maxSrc)--;
	(*src)++;
      }
    } else {
      dest++;
      numChars++;
    }
  }

  // null terminate the string
  *dest = 0;

  return numChars;
}

//==================================================================================================
// decompressJPEG()
//
// I snagged this code from the jpeg slideshow sample.
//
// Inputs:
//   jpegdata  - a pointer to the JPG data
//   jpeg_size - the size of the JPG data
//   *imgW     - gets filled in with the image width
//   *imgH     - gets filled in with the image height
// Outputs: none
//   void* - a pointer to the raster image data in system memory
//   *imgW - filled in with image width
//   *imgH - filled in with image height
// Globals modified: none
//
//==================================================================================================

void* decompressJPEG(void *jpegdata, int jpeg_size, long *imgW, long *imgH)
{
  struct jpeg_decompress_struct cinfo;
  struct jpeg_error_mgr jerr;
  JOCTET *s;
  JSAMPARRAY buffer;
  int i, n, row_stride;
  int x, y;
  int scale = 1;
  long *outImg, *ptr;

  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_decompress(&cinfo);
  jpeg_mem_src(&cinfo, (JOCTET *)jpegdata, jpeg_size);
  jpeg_read_header(&cinfo, TRUE);

  cinfo.out_color_space = JCS_YCbCr;
  cinfo.scale_num = 1;
  cinfo.scale_denom = scale;
  jpeg_start_decompress(&cinfo);

  *imgW = cinfo.output_width;
  *imgH = cinfo.output_height;
  outImg = (long *) malloc(cinfo.output_width * cinfo.output_height * 4);
  if (!outImg) {
    jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);
    return(NULL);
  }
  
  row_stride = cinfo.output_components * cinfo.output_width;
  buffer = (*cinfo.mem->alloc_sarray) ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride,
				       cinfo.rec_outbuf_height);

  y = 0;
  ptr = outImg;
  while (cinfo.output_scanline < cinfo.output_height) {
    
    n = jpeg_read_scanlines(&cinfo, buffer, cinfo.rec_outbuf_height);

    for (i = 0; i < n; i++) {

      s = buffer[i];
      for (x = 0; x < cinfo.output_width; x++) {

	// Unfortunately, the buffer output by the JPEG routines is not ready-to-use
	// We gotta reshuffle the pixel component to what we want.
	*ptr++ = (s[0] << 24) | (s[2] << 16) | (s[1] << 8);
    
	s += 3;
      }
      y++;
    }
  }
  
  // free resources
  jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);

  return(outImg);
}

//==================================================================================================
// getFreeMem()
//
// Computes the amount of free memory in a particular memory space.
//
// Inputs:
//   memType - which memory to count - either kMemSDRAM or kMemSysRam (same as bios)
// Outputs:
//   unsigned long - the amount of free memory
// Globals modified: none
//
//==================================================================================================

unsigned long getFreeMem(int memType)
{
  void *ptr[50];
  int index = 0;
  unsigned long curAmt;
  unsigned long totalAmt = 0;

  if (memType == kMemSDRAM) {
    
    for (curAmt = 0x800000; curAmt; ) {
      ptr[index] = _MemAlloc(curAmt, 1, kMemSDRAM);
      if (ptr[index]) {
	//	printf("\nAllocated: %ld bytes", curAmt);
	//      fflush(stdout);
	index++;
	totalAmt += curAmt;
      } else {
	curAmt >>= 1;
      }
    }

    //    printf("\nTotal video memory allocated: %ld bytes", totalAmt);
    //    fflush(stdout);
  
    while(index) {
      index--;
      _MemFree(ptr[index]);
    }

  } else {
    for (curAmt = 0x800000; curAmt; ) {
      ptr[index] = malloc(curAmt);
      if (ptr[index]) {
	//	printf("\nAllocated: %ld bytes", curAmt);
	//      fflush(stdout);
	index++;
	totalAmt += curAmt;
      } else {
	curAmt >>= 1;
      }
    }

    //    printf("\nTotal system memory allocated: %ld bytes", totalAmt);
    //    fflush(stdout);

    while(index) {
      index--;
      free(ptr[index]);
    }
  }
  
  return(totalAmt);
}

//==================================================================================================
// freeAppImages()
//
// Frees the memory used by the app images (img1 & img2).
//
// Inputs:   none
// Outputs:  none
// Globals modified: all app images are freed and set to null
//
//==================================================================================================

void freeAppImages()
{
  int appNum;
  int imgGroupNum;
  int imgNum;
  sApp *app;
  sAppImgGroup *imgGroup;
  sAppImg *img;

  for (appNum = 0; appNum < numApps; appNum++) {
    app = &aAppInfo[appNum];

    for (imgGroupNum = 0; imgGroupNum < app->numImgGroups; imgGroupNum++) {
      imgGroup = &app->aImgGroups[imgGroupNum];

      for (imgNum = 0; imgNum < imgGroup->numImg; imgNum++) {
	img = &imgGroup->aAppImg[imgNum];
	
	if (img->pixmap) {
	  mmlReleasePixmaps((mmlPixmap *) img->pixmap, &sysRes, 1);
	}
	img->status = kImgStatusInit;
	img->pixmap = NULL;
      }
    }
  }
}


//==================================================================================================
// imagesReady()
//
// Returns TRUE or FALSE indicating if all app images are ready for drawing
//
// Inputs:   none
// Outputs:  boolean
// Globals modified: none
//
//==================================================================================================

int imagesReady(int appNum)
{
  int imgGroupNum;
  sApp *app;
  sAppImgGroup *imgGroup;

  if (!aAppInfo) {
    return FALSE;
  }

  app = &aAppInfo[appNum];

  for (imgGroupNum = 0; imgGroupNum < app->numImgGroups; imgGroupNum++) {
    imgGroup = &app->aImgGroups[imgGroupNum];

    if (imgGroup->aAppImg && (imgGroup->numImg > 0)) {
      if (imgGroup->aAppImg[imgGroup->curImg].status == kImgStatusInit) {
	return FALSE;
      }
    }
  }

  return TRUE;
}


//==================================================================================================
// redrawImages()
//
// Forces images to be redrawn given a state change from one app to another.
//
// Inputs:   
//   thisApp - the current app
//   nextApp - the app to which we are switching
// Outputs:  boolean
// Globals modified: none
//
//==================================================================================================

void redrawImages(int thisApp, int nextApp)
{
  int imgGroupNum;
  sApp *app1, *app2;

  if (!aAppInfo) {
    return;
  }

#ifdef TRACE
  printf("%s(%d, %d) begin\n", __func__, thisApp, nextApp);
#endif

  app1 = &aAppInfo[thisApp];
  app2 = &aAppInfo[nextApp];
  for (imgGroupNum = 0; imgGroupNum < app1->numImgGroups; imgGroupNum++) {
    // Only clear the images that were there before
    if ((app1->aImgGroups[imgGroupNum].numImg > 0) || (app2->aImgGroups[imgGroupNum].numImg > 0)) {
#ifdef TRACE
      printf("%s() - Clearing img %d\n", __func__, imgGroupNum);
#endif

      app2->aImgGroups[imgGroupNum].draw = 2;
    }
  }

#ifdef TRACE
  printf("%s() end\n", __func__);
#endif
}

//==================================================================================================
// Future enhancements
// o load images in the background to speed navigation
// o allow auto-W/H for instructions (must take worst case of all instruction text)
// o Vertical alignment parameters
// o String formatting to allow changing of font sizes/colors/styles
// o Create pixmaps for text to allow smooth scrolling.
// o Put text in overlay channel and images on main channel so they no longer conflict.
//==================================================================================================
